/* $Id: fa_utils.c,v 1.89 2001/09/27 17:33:54 jm Exp $
 * Foreign Agent - independent help routines
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifndef _GNU_SOURCE
#define _GNU_SOURCE /* for snprintf() etc. */
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>
#include <unistd.h>
#include <assert.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>

#include "tunnel.h"
#include "debug.h"
#include "rsa_dyn.h"
#include "dyn_ip.h"
#include "fa.h"
#include "fa_mn_addr.h"
#include "fa_hash.h"
#include "util.h"
#include "proxyarp.h"


#ifndef IP_PKTINFO
#define IP_PKTINFO 8
#endif
#ifndef IP_RECVTTL
#define IP_RECVTTL      12
#endif

#define PURGE_BINDING_TUNNELS 2

extern struct interface_entry *up_interface;
extern int fa_info_sock;
extern struct sockaddr_un fa_info_addr;
extern struct sockaddr_in upper_fa_addr;
extern struct fa_config *config;


/**
 * send_address_ok:
 * @to_addr: 
 * 
 * try to avoid looping messages to own addresses
 *
 * Returns: TRUE if the address is okay or FALSE if it is not 
 */
int
send_address_ok(struct in_addr to_addr)
{
#ifndef DEBUG_SIMULATED_TEST
	char dev[IFNAMSIZ];

	/* check that destionation interface != lo */
	if (dyn_ip_route_get(to_addr, dev, IFNAMSIZ) == 0 &&
	    strcmp(dev, "lo") == 0) {
		LOG2(LOG_ALERT,
		     "send_address_ok: destination to 'lo' "
		     "interface - possible loop => refusing\n");
		return FALSE;
	}
#endif

	return TRUE;
}


/**
 * own_sendto:
 * @sock: 
 * @to_addr: 
 * @to_port: 
 * @data: 
 * @len: 
 * 
 * 
 * 
 * Returns: 
 */
int
own_sendto(int sock, struct in_addr to_addr, unsigned int to_port, void *data,
	   int len)
{
	struct sockaddr_in sa;

	if (!send_address_ok(to_addr))
		return -2;
	memset(&sa, 0, sizeof(sa));
	sa.sin_family = AF_INET;
	sa.sin_addr = to_addr;
	sa.sin_port = to_port;
	return sendto(sock, data, len, 0,
		     (struct sockaddr *) &sa, sizeof(sa));
}


/**
 * close_socket:
 * @fd: 
 * 
 * 
 */
void
close_socket(int fd)
{
	if (fd < 0) return;
	if (close(fd) == -1)
		DEBUG(DEBUG_FLAG, "close_socket(%i) - %s\n", fd,
		      strerror(errno));
}


/**
 * free_unconfirmed_data:
 * @unc: 
 * 
 * 
 */
void
free_unconfirmed_data(struct unconfirmed_request *unc)
{
	struct unconfirmed_request *prev;

	while (unc != NULL) {
		prev = unc;
		unc = unc->next;
		if (prev->fa_pubkey != NULL)
			free(prev->fa_pubkey);
		if (prev->challenge != NULL)
			free(prev->challenge);
		free(prev);
	}
}

/**
 * free_binding:
 * @binding: the binding to be freed
 * @bcounters: binding counters (confirmed bindings, pending registrations)
 * 
 * Frees the memory used by a binding and updates the binding and pending
 * registration counts.
 */
static void
free_binding(struct bindingentry *binding, struct binding_counters *bcounters)
{
	struct tunnel_data *t_data;

	if (binding == NULL || binding->data == NULL)
		return;

	t_data = (struct tunnel_data *) binding->data;

	if (t_data->used_challenges != NULL) {
		int i;
		for (i = 0; i < 2 * config->challenge_window; i++)
			if (t_data->used_challenges[i] != NULL)
				free(t_data->used_challenges[i]);
		free(t_data->used_challenges);
	}
	if (t_data->fa_pubkey != NULL)
		free(t_data->fa_pubkey);
	if (t_data->last_sent_fa_pubkeyrep != NULL)
		free(t_data->last_sent_fa_pubkeyrep);
	if (t_data->last_recv_fa_pubkeyrep != NULL)
		free(t_data->last_recv_fa_pubkeyrep);
	if (t_data->last_challenge != NULL)
		free(t_data->last_challenge);

	free_unconfirmed_data(t_data->unc_req);

	if (bcounters != NULL) {
		if (t_data->confirmed)
			bcounters->bindingcount--;
		else
			bcounters->pendingcount--;
	}

	free(binding->data);
	free(binding);
}


static int
elim_binding(struct node *node, void *data, int force)
{
	struct bindingentry *binding;
	int type;
	struct tunnel_data *t_data;

	binding = (struct bindingentry *) node;
	t_data = (struct tunnel_data *) binding->data;

	if (t_data->arpentry) {
		DEBUG(DEBUG_FLAG, "Removing the ARP entry of the MN\n");
		if (arp_del_permanent_item(t_data->arp_ipaddr, t_data->arp_dev)
		    < 0) {
			LOG2(LOG_WARNING, "Could not remove permanent ARP "
			     "entry (%s, %s)\n",
			     inet_ntoa(t_data->arp_ipaddr), t_data->arp_dev);
		} else
			t_data->arpentry = 0;
	}

	if (t_data->confirmed) {
#ifdef DYNAMICS_SET_MAIN_TABLE_MN_ROUTE
		if (t_data->fa_decapsulation &&
		    dyn_ip_route_del(binding->mn_addr, t_data->info.iface->dev)
		    != 0) {
			LOG2(LOG_WARNING,
			     "dyn_ip_route_del(%s,%s) failed\n",
			     inet_ntoa(binding->mn_addr),
			     t_data->info.iface->dev);
		}
#endif /* DYNAMICS_SET_MAIN_TABLE_MN_ROUTE */

		if (t_data->up_tunl &&
		    dyn_ip_rule_del_table(NULL, &binding->mn_addr,
					  t_data->up_tunl->to_mn_table_id,
					  "lo") != 0) {
			LOG2(LOG_WARNING, "Removal of routing rule for "
			     "locally generated packets failed (mn_addr=%s, "
			     "table=%i)\n", inet_ntoa(binding->mn_addr),
			     t_data->up_tunl->to_mn_table_id);
		}

		if (tunnel_unconnect(t_data->up_tunl, t_data->down_tunl,
				     binding->mn_addr, t_data->tunnel_mode ==
				     TUNNEL_MODE_REVERSE ? 1 : 0, force,
				     t_data->force_route_dev,
				     t_data->force_reverse_dev) < 0) {
			LOG2(LOG_WARNING, "tunnel unconnect failed "
			     "(tun_dev=%s, dev=%s, encaps_delivery=%i)\n",
			     binding->tun_dev, t_data->info.iface->dev,
			     t_data->encaps_delivery);
		}

		if (tunnel_delete_ptr(t_data->down_tunl, force) == -1) {
			LOG2(LOG_WARNING,
			     "tunnel down - tunnel_delete failed\n");
		}
		if (tunnel_delete_ptr(t_data->up_tunl, force) == -1) {
			LOG2(LOG_WARNING,
			     "tunnel up - tunnel_delete failed\n");
		}
	}

	if (t_data->up_type == TUNNEL_IPIP)
		type = -1;
	else
		type = t_data->up_key;
	if (t_data->mn_addr_added) {
		if (mn_addr_remove(&binding->mn_addr, type) < 0) {
			LOG2(LOG_WARNING, "mn_addr_remove(%s,%i) failed\n",
			     inet_ntoa(binding->mn_addr), type);
		} else {
			DEBUG(DEBUG_FLAG, "mn_addr - released addr=%s, "
			      "type=%i\n", inet_ntoa(binding->mn_addr), type);
		}
	}

	binding_remove(binding);
	free_binding(binding, data);

	return TRUE;
}


/**
 * eliminate_binding_entry:
 * @node: 
 * @data: 
 * 
 * 
 * 
 * Returns: 
 */
int
eliminate_binding_entry(struct node *node, void *data)
{
	return elim_binding(node, data, 0);
}


/**
 * eliminate_binding_entry_force:
 * @node: 
 * @data: 
 * 
 * 
 * 
 * Returns: 
 */
int
eliminate_binding_entry_force(struct node *node, void *data)
{
	return elim_binding(node, data, 1);
}


/**
 * update_binding:
 * @hash: pointer to the binding hashtable
 * @binding: the binding to be updated
 * @count: confirmed bindings count
 * @pcount: pending registration count
 * 
 * Updates the binding entry by removing the entry from the binding table and
 * adding it back. This needs to be done when the lifetime of the binding has
 * changed.
 * 
 * Returns: 1 if successful, else 0
 */
int
update_binding(struct bindingtable *hash, struct bindingentry *binding,
	       struct binding_counters *bcounters)
{
	if (!binding_update(hash, binding)) {
		LOG2(LOG_WARNING, "update_binding: binding_update failed\n");
		free_binding(binding, bcounters);
		return 1;
	}
	return 0;
}


/**
 * make_new_binding:
 * @bhash: Table of all bindings.
 * @ext: Registation request extensions.
 * @max_lifetime: Maximum lifetime allowed for the binding.
 * 
 * Create a binding and add it to the binding table @bhash. The
 * lifetime for the binding will the smaller value of @max_lifetime
 * and the requested lifetime in @ext.
 * 
 * Returns: A pointer to the binding, or NULL on failure.
 */
struct bindingentry *
make_new_binding(struct bindingtable *bhash, struct msg_extensions *ext,
		 int max_lifetime, struct binding_counters *bcounters)
{
	struct bindingentry *binding;
	struct tunnel_data *t_data;
	int i;

	binding = (struct bindingentry *) malloc(sizeof(struct bindingentry));
	if (binding == NULL)
		return NULL;
	t_data = (struct tunnel_data *) malloc(sizeof(struct tunnel_data));
	if (t_data == NULL) {
		free(binding);
		return NULL;
	}

	memset(binding, 0, sizeof(struct bindingentry));
	memset(t_data, 0, sizeof(struct tunnel_data));

	t_data->used_challenges = (struct challenge_ext **)
		malloc(2 * config->challenge_window *
		       sizeof(struct challenge_ext *));
	if (t_data->used_challenges == NULL) {
		free(binding);
		free(t_data);
		return NULL;
	}
	for (i = 0; i < 2 * config->challenge_window; i++)
		t_data->used_challenges[i] = NULL;

	binding->data = t_data;

	time(&binding->create_time);
	time(&binding->mod_time);
	binding->mn_addr.s_addr = ext->req->home_addr.s_addr;
	binding->ha_addr.s_addr = ext->req->ha_addr.s_addr;
	if (ext->priv_ha != NULL)
		binding->priv_ha = ntohl(ext->priv_ha->priv_ha);
	else
		binding->priv_ha = 0;
	t_data->confirmed = FALSE;

#ifdef INCLUDE_IPAY
	if (ext->mn_nai) {
		t_data->mn_nai.len = GET_MN_NAI_LEN(ext->mn_nai);
		t_data->mn_nai.nai = malloc(t_data->mn_nai.len);
		if (t_data->mn_nai.nai == NULL) {
			free(t_data);
			free(binding);
			return NULL;
		}
		memcpy(t_data->mn_nai.nai, MSG_MN_NAI_DATA(ext->mn_nai),
		       t_data->mn_nai.len);
	}
#endif

	/* add default values for upper tunnel - changed if needed */
	t_data->up_type = TUNNEL_IPIP;
	t_data->up_key = 0;

	if (ntohs(ext->req->lifetime) < max_lifetime)
		binding->timeout = ntohs(ext->req->lifetime);
	else
		binding->timeout = max_lifetime;
	binding->exp_time = time(NULL) + binding->timeout;
	if (!binding_add(bhash, binding)) {
		free_binding(binding, NULL);
		return NULL;
	}

	bcounters->pendingcount++;
	return binding;
}





/**
 * equal_pubkey:
 * @key1: 
 * @key2: 
 * 
 * Returns: non-zero if key1 and key2 are identical or zero if they differ 
 */
int
equal_pubkey(struct msg_key *key1, struct msg_key *key2)
{
	if (key1 == NULL && key2 == NULL)
		return 1;
	if (key1 == NULL || key2 == NULL)
		return 0;

	if (key1->type != key2->type || key1->length != key2->length ||
	    key1->spi != key2->spi)
		return 0;

	return memcmp(MSG_KEY_DATA(key1), MSG_KEY_DATA(key2),
		      GET_KEY_LEN(key1)) == 0;
}


/**
 * do_rsa_decrypt:
 * @keyrep: 
 * @sk: 
 * @sk_len: 
 * 
 * 
 * 
 * Returns: 
 */
int
do_rsa_decrypt(struct msg_key *keyrep, unsigned char *sk, int sk_len)
{
#ifndef NODEBUGMODE
	int i;
#endif

	if (rsa_decrypt_session_key(MSG_KEY_DATA(keyrep), keyrep->length,
				    sk, sk_len) != 0) {
		LOG2(LOG_WARNING, "Session key decryption failed!\n");
		return 1;
	}

#ifndef NODEBUGMODE
	DEBUG(DEBUG_FLAG, "Session key (RSA): ");
	for (i = 0; i < sk_len; i++)
		DEBUG(DEBUG_FLAG, "%02x", sk[i]);
	DEBUG(DEBUG_FLAG, "\n");
#endif

	return 0;
}


static int
create_tunnel_upwards(struct bindingentry *binding, struct hashtable *thash,
		      int highest_FA)
{
	struct tunnel_data *t_data;

	t_data = binding->data;

	DEBUG(DEBUG_FLAG, "create_tunnel_upwards: highest_FA=%i, "
	      "t_data->up_type=%i, t_data->down_type=%i\n", highest_FA,
	      t_data->up_type, t_data->down_type);

	assert(up_interface != NULL);
	t_data->up_tunl = tunnel_add(thash, highest_FA ? binding->ha_addr :
				     upper_fa_addr.sin_addr,
				     t_data->ha_tun_dev,
				     up_interface->addr, 1, t_data->up_type,
				     t_data->up_key);
	if (t_data->up_tunl == NULL) {
		LOG2(LOG_WARNING,
		     "create_tunnel_upwards: tunnel add failed\n");
		return -1;
	}

	if (!highest_FA) {
		/* don't remove tunnels to upper FAs */
		tunnel_set_remove(t_data->up_tunl, 1);
	}

	if (tunnel_connect(t_data->up_tunl, t_data->down_tunl,
			   binding->mn_addr, t_data->tunnel_mode ==
			   TUNNEL_MODE_REVERSE ? 1 : 0,
			   (t_data->fa_decapsulation &&
			    t_data->encaps_delivery)
			   ? t_data->info.iface->dev : NULL,
			   (!t_data->fa_decapsulation &&
			    !t_data->encaps_delivery)
			   ? t_data->info.iface->dev : NULL) < 0) {
		LOG2(LOG_WARNING, "create_tunnel_upwards: "
		     "- tunnel connection failed\n");
		return -1;
	}

	if (dyn_ip_rule_add_table(NULL, &binding->mn_addr,
				  t_data->up_tunl->to_mn_table_id,
				  "lo") != 0) {
		LOG2(LOG_WARNING, "Adding routing rule for "
		     "locally generated packets failed (mn_addr=%s, "
		     "table=%i)\n", inet_ntoa(binding->mn_addr),
		     t_data->up_tunl->to_mn_table_id);
	}

	return 0;
}


/**
 * create_tunnels:
 * @binding:
 * @thash: hash of tunnels
 * @highest_FA: is this the highest FA?
 * 
 * Called when reply received. Takes care of tunnels and routing
 * upwards. 
 * 
 * Returns: 0 if successful, else 1 or -1
 */
int
create_tunnels(struct bindingentry *binding, struct hashtable *thash,
	       int highest_FA)
{
	struct tunnel_data *t_data;
	char *dev;

	t_data = binding->data;

	/* Tunnel downwards
	 * The tunnel should be already made, but with reference count
	 * zero as the lazy tunnel deletion keeps the tunnel up long
	 * enough. Look in handle_request(). */
	assert(t_data->info.iface != NULL);
	if (t_data->fa_decapsulation && !t_data->encaps_delivery) {
		DEBUG(DEBUG_FLAG, "create_tunnels: down_key - %i=>%i\n",
		      t_data->down_key, t_data->info.iface->if_index);
		t_data->down_type = TUNNEL_NO_ENCAPS;
		/* use ifindex to make separate entries for each interface */
		t_data->down_key = t_data->info.iface->if_index;
	}

	t_data->down_tunl = tunnel_add(thash, binding->lower_addr,
				       binding->tun_dev,
				       t_data->info.iface->addr,
				       1, t_data->down_type, t_data->down_key);

	if (t_data->down_tunl == NULL) {
		LOG2(LOG_WARNING, "create_tunnels: tunnel_add failed.\n");
		return 1;
	}

	/* Lock the tunnel in use if it is to a lower FA (not to MN using
	 * MN decapsulation) */
	if (t_data->down_type != TUNNEL_NO_ENCAPS &&
	    fa_hash_check_next_hop(binding->lower_addr)) {
		/* don't remove tunnels to lower FAs */
		tunnel_set_remove(t_data->down_tunl, 1);
	}

	dev = t_data->fa_decapsulation ? t_data->info.iface->dev :
		binding->tun_dev;
	if (t_data->down_type == TUNNEL_NO_ENCAPS) {
		dynamics_strlcpy(t_data->down_tunl->device, dev, IFNAMSIZ);
	}

	/* Add an ARP entry if this FA is directly connected to the MN
	 * (i.e., LFA) */
	if (t_data->is_lowest_fa && !t_data->arpentry) {
		struct sockaddr hwaddr;

		DEBUG(DEBUG_FLAG, "Adding ARP entry\n");
		memset(&hwaddr, 0, sizeof(hwaddr));
		hwaddr.sa_family = ARPHRD_ETHER;
		memcpy(hwaddr.sa_data, t_data->info.from.sll_addr,
		       t_data->info.from.sll_halen);
		t_data->arp_ipaddr.s_addr = t_data->info.src.sin_addr.s_addr;
		memcpy(t_data->arp_dev, t_data->info.iface->dev,
		       sizeof(t_data->arp_dev));

		if (arp_add_permanent_item(t_data->arp_ipaddr,
					   t_data->arp_dev,
					   &hwaddr) < 0) {
			LOG2(LOG_WARNING,
			     "Could not add ARP entry (%s, %s, %s)\n",
			     inet_ntoa(t_data->arp_ipaddr),
			     t_data->arp_dev,
			     ether_hwtoa((unsigned char *) hwaddr.sa_data));
		} else
			t_data->arpentry = 1;
	}

#ifdef DYNAMICS_SET_MAIN_TABLE_MN_ROUTE
	/* add also a route entry to the main routing table to the LFA if
	 * MN is using FA decaps */
	/* Also this should have been done in handle request phase */
	if (t_data->fa_decapsulation &&
	    dyn_ip_route_replace(binding->mn_addr, dev) == -1) {
		LOG2(LOG_WARNING, "create_tunnels: dyn_ip_route_replace "
		     "failed.\n");
		return -1;
	}
#endif /* DYNAMICS_SET_MAIN_TABLE_MN_ROUTE */

	/* Create tunnel upwards and connect the lower tunnel */
	/* This is "new". In handle request phase only the downstream
	 * is handled but here we take care of the upstream also. */
	if (create_tunnel_upwards(binding, thash, highest_FA) < 0) {
		LOG2(LOG_WARNING,
		     "create_tunnels: create_tunnel_upwards failed.\n");
		return -1;
	}

	return 0;
}


/**
 * add_unconfirmed_data:
 * @bhash: 
 * @bcounters: 
 * @binding: 
 * @ext: 
 * @info: 
 * @config: 
 * 
 * Allocates memory to an unconfirmed_request.
 * 
 * Returns: Pointer to an unconfirmed_request, or NULL on error.
 */
struct unconfirmed_request *
add_unconfirmed_data(struct bindingtable *bhash,
		     struct binding_counters *bcounters,
		     struct bindingentry *binding, struct msg_extensions *ext,
		     struct packet_from_info *info,
		     struct fa_config *config)
{
	struct tunnel_data *t_data;
	struct unconfirmed_request *unc;
	int maxlifetime;
	time_t now;

	DEBUG(DEBUG_FLAG, "Adding unconfirmed request data\n");

	t_data = binding->data;
	time(&now);
	binding->mod_time = now;
	unc = (struct unconfirmed_request *)
		malloc(sizeof(struct unconfirmed_request));
	if (unc == NULL)
		return NULL;

	memset(unc, 0, sizeof(struct unconfirmed_request));
	memcpy(&unc->info, info, sizeof(*info));
	memcpy(&unc->req, ext->req, sizeof(struct reg_req));

	if (ext->challenge) {
		int n = GET_CHALLENGE_EXT_LEN(ext->challenge);
		unc->challenge = (struct challenge_ext *) malloc(n);
		if (unc->challenge == NULL) {
			DEBUG(DEBUG_FLAG, "\tmalloc for challenge failed\n");
			free(unc);
			return NULL;
		}
		memcpy(unc->challenge, ext->challenge, n);
	}

	unc->next = t_data->unc_req;
	unc->created = now;

	unc->id[0] = ntohl(ext->req->id[0]);
	unc->id[1] = ntohl(ext->req->id[1]);
	if (ext->mh_auth != NULL)
		unc->spi = ntohl(ext->mh_auth->spi);
	unc->ha_addr = ext->req->ha_addr;
	unc->timeout = ntohs(ext->req->lifetime);

	if ((ext->req->opts & REGREQ_REVERSE_TUNNEL) != 0 || 
	    config->force_reverse_tunneling)
		unc->tunnel_mode = TUNNEL_MODE_REVERSE;
	else
		unc->tunnel_mode = TUNNEL_MODE_TRIANGLE;

	unc->is_lowest_fa = is_sender_mobile(ext);
	if (unc->is_lowest_fa)
		DEBUG(DEBUG_FLAG,
		      "\tassuming lowest FA (i.e. request from MN)\n");
	else
		DEBUG(DEBUG_FLAG,
		      "\tassuming not lowest FA (i.e. request from FA)\n");

	/* only use FA decapsulation if this is the lowest FA and MN
	 * requests it */
	unc->fa_decapsulation = unc->is_lowest_fa &&
		(ext->req->opts & REGREQ_MN_DECAPS) == 0;
	unc->encaps_delivery = !unc->is_lowest_fa || ext->encaps_del != NULL;
	unc->old_if_index = info->iface->if_index;
	unc->mn_nai_included = ext->mn_nai != NULL;

	if (ext->fa_pubkey) {
		unc->fa_pubkey = (struct msg_key *)
			malloc(GET_KEY_EXT_LEN(ext->fa_pubkey));
		if (unc->fa_pubkey != NULL) {
			memcpy(unc->fa_pubkey, ext->fa_pubkey,
			       GET_KEY_EXT_LEN(ext->fa_pubkey));
		} else {
			LOG2(LOG_ALERT, "add_unconfirmed_data - not enough "
			     "memory\n");
			if (unc->challenge)
				free(unc->challenge);
			free(unc);
			return NULL;
		}
	} else if (ext->fa_keyreq) {
		unc->fa_spi = ntohl(ext->fa_keyreq->spi);
	}

	if (ext->gre_key != NULL) {
		unc->tunnel_type = TUNNEL_GRE;
		unc->tunnel_key = ntohl(ext->gre_key->key);
	} else 	if (unc->fa_decapsulation && !unc->encaps_delivery) {
		unc->tunnel_type = TUNNEL_NO_ENCAPS;
		unc->tunnel_key = info->iface->if_index;
	} else {
		unc->tunnel_type = TUNNEL_IPIP;
		unc->tunnel_key = 0;
	}

	t_data->unc_req = unc;

	maxlifetime = ntohs(ext->req->lifetime);
	if (config->delete_pending_after > 0 &&
	    config->delete_pending_after < maxlifetime)
		maxlifetime = config->delete_pending_after;
#ifdef SHORT_BINDING_DELAY_TIME
	/* for testing - update lifetime even if binding is confirmed */
	if (
#else
	if (!t_data->confirmed &&
#endif
	    binding->exp_time - now < maxlifetime) {
		/* update the unconfirmed binding timeout so that extra
		 * requests will not cause unwanted timeouts */
		DEBUG(DEBUG_FLAG, "\tupdated binding lifetime %i => %i\n",
		      binding->timeout, maxlifetime);
		binding->timeout = maxlifetime;
		binding->exp_time = now + binding->timeout;

		if (update_binding(bhash, binding, bcounters) != 0)
			return NULL;
	}

	return unc;
}


/**
 * unconfirmed_to_binding:
 * @binding: 
 * @unc: 
 * 
 * 
 */
void
unconfirmed_to_binding(struct bindingentry *binding,
		       struct unconfirmed_request *unc)
{
	struct tunnel_data *t_data = binding->data;
	binding->lower_addr = unc->info.src.sin_addr;
	binding->lower_port = unc->info.src.sin_port;
	memcpy(&t_data->info, &unc->info, sizeof(unc->info));
#ifdef USE_TEARDOWN
	t_data->upper_id[0] = binding->id[0] = unc->id[0];
	t_data->upper_id[1] = binding->id[1] = unc->id[1];
#endif
	binding->spi = unc->spi;
	binding->fa_spi = unc->fa_spi;
	t_data->tunnel_mode = unc->tunnel_mode;
	t_data->fa_decapsulation = unc->fa_decapsulation;
	t_data->encaps_delivery = unc->encaps_delivery;
	t_data->is_lowest_fa = unc->is_lowest_fa;
	binding->ha_addr = unc->ha_addr;
	if (t_data->fa_pubkey != NULL)
		free(t_data->fa_pubkey);
	t_data->fa_pubkey = unc->fa_pubkey;
	unc->fa_pubkey = NULL;
	if (t_data->last_sent_fa_pubkeyrep != NULL) {
		free(t_data->last_sent_fa_pubkeyrep);
		t_data->last_sent_fa_pubkeyrep = NULL;
	}
	t_data->last_used_seq_num = 0;
	t_data->req_lifetime = unc->timeout;
	t_data->old_if_index = unc->old_if_index;

	setup_device_forcing(t_data);

	t_data->down_type = unc->tunnel_type;
	DEBUG(DEBUG_FLAG, "unconfirmed_to_binding: down_key - %i=>%i\n",
	      t_data->down_key, unc->tunnel_key);
	t_data->down_key = unc->tunnel_key;
}


/**
 * check_unconfirmed_timeout:
 * @binding:
 *
 * check if given binding has unconfirmed requests that have timeouted
 * and remove them 
 */
void
check_unconfirmed_timeout(struct bindingentry *binding)
{
	time_t now;
	struct tunnel_data *t_data;
	struct unconfirmed_request *unc, *prev;

	time(&now);
	t_data = binding->data;
	unc = t_data->unc_req;
	prev = NULL;
	while (unc != NULL) {
		if (now - unc->created > UNCONFIRMED_TIMEOUT) {
			DEBUG(DEBUG_FLAG, "Freeing unconfirmed requests that "
			      "have timed out\n");
			free_unconfirmed_data(unc);
			if (prev == NULL)
				t_data->unc_req = NULL;
			else
				prev->next = NULL;
			break;
		}
		prev = unc;
		unc = unc->next;
	}
}


/**
 * remove_binding_tunnels:
 * @binding: 
 * @thash: 
 * @bcounters: 
 * 
 * 
 */
void
remove_binding_tunnels(struct bindingentry *binding, struct hashtable *thash,
		       struct binding_counters *bcounters)
{
	struct tunnel_data *t_data;
	struct in_addr dst_addr, mn_addr, ha_addr;
	int up_type;
	__u32 up_key ;

	t_data = binding->data;

	up_type = t_data->up_type;
	up_key = t_data->up_key;
	if (t_data->up_tunl)
		dst_addr = t_data->up_tunl->dst_addr;
	else
		dst_addr.s_addr = 0;

	/* Remove possible tunnel upwards to HA also */
	mn_addr.s_addr = binding->mn_addr.s_addr;
	ha_addr.s_addr = binding->ha_addr.s_addr;
	eliminate_binding_entry((struct node *) binding, bcounters);
}


/**
 * check_bindings:
 * @bhash: 
 * @thash: 
 * @bcounters: 
 * 
 * Check and remove expired bindings
 */
void
check_bindings(struct bindingtable *bhash, struct hashtable *thash,
	       struct binding_counters *bcounters)
{
	static time_t prev_check;
	static int first = 1;
	struct bindingentry *binding;
	static time_t advance_expire_time;
	int diff;
	time_t time_check;
	time_t now;

	time(&now);

	if (first) {
		time(&prev_check);
		time(&advance_expire_time);
		first = 0;
	}

	/* don't check too frequently */
	if (now - prev_check < 1)
		return;
	prev_check = now;

	time(&time_check);
	diff = time_check - advance_expire_time;
	advance_expire_time = time_check;
	if (diff < 0) {
		LOG2(LOG_WARNING,
		     "check_bindings: diff(%i) < 0 - system time changed?\n",
		     diff);
		diff = 0;
	}
	binding = binding_getexpired(bhash, diff);
	while (binding != NULL) {
		struct tunnel_data *t_data;
		struct exp_info_struct {
			struct in_addr mn_addr;
			struct in_addr ha_addr;
		} exp_info;
		t_data = (struct tunnel_data *) binding->data;
		DEBUG(DEBUG_FLAG, "check_bindings: binding expired\n");
		DEBUG(DEBUG_FLAG, "\tdiff = %d, expiretime = %d\n",
		      diff, binding->timeout);
		DEBUG(DEBUG_FLAG, "\tbinding->mn_addr = %s\n",
		      inet_ntoa(binding->mn_addr));
		DEBUG(DEBUG_FLAG, "\tbinding->lower_addr = %s\n",
		      inet_ntoa(binding->lower_addr));

		if (t_data->pending_request && t_data->unc_req != NULL) {
			struct msg_extensions ext;

			DEBUG(DEBUG_FLAG, "\tpending request timeout - "
			      "informing MN\n");
			memset(&ext, 0, sizeof(ext));
			ext.req = &t_data->unc_req->req;
			send_failure_reply(REGREP_REGISTRATION_TIMEOUT_FA,
					   &ext, &t_data->unc_req->info, NULL,
					   0);
		}

		exp_info.mn_addr = binding->mn_addr;
		exp_info.ha_addr = binding->ha_addr;

		remove_binding_tunnels(binding, thash, bcounters);
		binding = binding_getexpired(bhash, 0);

		info_send(FA_INFO_TIMEOUT, FA_INFO_OK, &exp_info,
			  sizeof(exp_info));
	}
}


/**
 * is_sender_mobile:
 * @ext: 
 * 
 * Determins if the sender of a request message is directly from a
 * mobile node based on the registration request extensions
 * 
 * Returns: 1 if the sender is a mobile, else 0
 */
int
is_sender_mobile(struct msg_extensions *ext)
{
	/* FIX: this should probably be replaced with
	 * ext->ff_auth == NULL */
	if (ext->fa_pubkey != NULL || ext->fa_keyreq != NULL)
		return 0;

	return 1;
}


/**
 * info_send:
 * @type: 
 * @status: 
 * @data: 
 * @len: 
 * 
 * 
 * 
 * Returns: 
 */
int
info_send(int type, int status, void *data, int len)
{
	void *buf;
	int n;

	if (fa_info_sock < 0) return -1;
	if (len > 0 && data == NULL) return -1;

	buf = malloc(sizeof(type) + sizeof(status) + len);
	if (buf == NULL) return -1;

	memcpy(buf, &type, sizeof(type));
	memcpy((char *)buf + sizeof(type), &status, sizeof(status));
	if (len > 0)
		memcpy((char *)buf + sizeof(type) + sizeof(status), data, len);
	n = sendto(fa_info_sock, buf, sizeof(type) + sizeof(status) + len, 0,
		   (struct sockaddr *) &fa_info_addr,
		   sizeof(fa_info_addr.sun_family) +
		   strlen(fa_info_addr.sun_path));
	free(buf);
	if (n < 0)
		DEBUG(DEBUG_FLAG, "info_send: %s\n", strerror(errno));
	return n;
}


/**
 * create_challenge_ext:
 * @config: pointer to FA configuration data
 * @type: extension type (AGENT_ADV_CHALLENGE_EXT / MN_FA_CHALLENGE_EXT)
 *
 * Create a new challenge extension. This function allocates memory for the
 * challenge and it is up to the caller to release the memory.
 * 
 * Returns:
 *   pointer to the created challenge or %NULL on failure
 */
struct challenge_ext *
create_challenge_ext(struct fa_config *config, int type)
{
	int left;
	struct challenge_ext *challenge;
	__u32 *pos, tmp;

	challenge = (struct challenge_ext *)
		malloc(sizeof(struct challenge_ext) +
		       config->challenge_length);
	if (challenge == NULL) {
		DEBUG(DEBUG_FLAG, "malloc for challenge_ext failed\n");
		return NULL;
	}

	left = config->challenge_length;
	challenge->type = type;
	challenge->length = config->challenge_length;
	pos = (__u32 *) MSG_CHALLENGE_EXT_DATA(challenge);
	while (left >= 4) {
		*pos++ = get_rand32();
		left -= 4;
	}
	if (left > 0) {
		tmp = get_rand32();
		memcpy(pos, &tmp, left);
	}

	return challenge;
}


/**
 * equal_challenge:
 * @c1: pointer to challenge extension 1 (or %NULL)
 * @c2: pointer to challenge extension 2 (or %NULL)
 *
 * Compare two challenges and return 1 only if both exist and are equal in
 * length and contents.
 *
 * Returns:
 *   1 if @c1 and @c2 are equal challenges or 0 if not
 */
int equal_challenge(struct challenge_ext *c1, struct challenge_ext *c2)
{
	int len1, len2;

	if (c1 == NULL || c2 == NULL)
		return 0;
	len1 = GET_CHALLENGE_LEN(c1);
	len2 = GET_CHALLENGE_LEN(c2);
	if (len1 != len2)
		return 0;
	if (memcmp(MSG_CHALLENGE_EXT_DATA(c1), MSG_CHALLENGE_EXT_DATA(c2),
		   len1) == 0)
		return 1;
	return 0;
}


/**
 * setup_device_forcing:
 * @t_data: FA's binding data
 *
 * Setup device forcing for reverse tunneling.
 */
void setup_device_forcing(struct tunnel_data *t_data)
{
	/* When FA decapsulation is used with encapsulating delivery,
	 * the down connection uses non-capsulated delivery to the MN,
	 * but encapsulated delivery from the MN. */
	if (t_data->fa_decapsulation && t_data->encaps_delivery)
		memcpy(t_data->force_route_dev, t_data->info.iface->dev,
		       IFNAMSIZ);
	else
		memset(t_data->force_route_dev, 0, IFNAMSIZ);

	/* When MN decapsulation is used without encapsulating delivery,
	 * the down connection uses encapsulated delivery to the MN,
	 * but non-capsulated delivery from the MN. */
	if (!t_data->fa_decapsulation && !t_data->encaps_delivery)
		memcpy(t_data->force_reverse_dev, t_data->info.iface->dev,
		       IFNAMSIZ);
	else
		memset(t_data->force_reverse_dev, 0, IFNAMSIZ);
}


/**
 * packet_socket_send:
 * @dst: L2 & L3 destination (from the original registration request)
 * @msg: message to be sent (UDP payload)
 * @len: length of the buffer (UDP payload) in bytes
 *
 * Send a registration reply using a packet and the destination learned from
 * the corresponding registration request. This function generates the IP and
 * UDP headers and instructs kernel to attach a suitable L2 header.
 *
 * Returns:
 * -1 on failure or number of sent bytes when successful
 */
int packet_socket_send(struct packet_from_info *dst, const unsigned char *msg,
		       int len)
{
	unsigned char *buf;
	struct iphdr *ip;
	struct udphdr *udp;
	int total_len, res;

	total_len = len + sizeof(*ip) + sizeof(*udp);
	buf = (unsigned char *) malloc(total_len);
	if (buf == NULL) {
		DEBUG(DEBUG_FLAG, "packet_socket_send: malloc failed\n");
		return -1;
	}

	/* FIX: if packet is longer than MTU, it should be fragmented; since
	 * the packet is send through a packet socket, the fragmentation is
	 * left to us.. */

	ip = (struct iphdr *) buf;
	memset(ip, 0, sizeof(*ip));
	ip->ihl = 5;
	ip->version = 4;
	ip->ttl = 255;
	ip->protocol = IPPROTO_UDP;
	ip->daddr = dst->src.sin_addr.s_addr;
	ip->saddr = dst->dst_addr.s_addr;
	ip->tot_len = htons(total_len);
	ip->check = 0;
	ip->check = ip_checksum(buf, sizeof(*ip));

	udp = (struct udphdr *) (ip + 1);
	udp->source = htons(434);
	udp->dest = dst->src.sin_port;
	udp->len = htons(sizeof(*udp) + len);

	memcpy(udp + 1, msg, len);

	udp->check = 0;
	udp->check = udp_checksum(ip, udp);

	res = sendto(dst->iface->udp_packet, buf, total_len, 0,
		     (struct sockaddr *) &dst->from, sizeof(dst->from));
	free(buf);

	if (res < 0)
		return res;

	return res - sizeof(*ip) - sizeof(*udp);
}
