/* $Id: mn_reg.c,v 1.64 2001/09/29 14:56:23 jm Exp $
 * Mobile Node - registration
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 * 
 * Additional modifications Copyright 2010 Juho Paaso, Aalto 
 * University / Department of Communication and Networking"
 */

#include "config.h"
#include <stdlib.h>
#include <assert.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#ifdef DYN_TARGET_LINUX
#include <net/if_arp.h>
#endif
#include <string.h>
#include <time.h>

#include <linux/icmp.h> 
#include <linux/udp.h>

#include "agentapi.h"
#include "auth.h"
#include "debug.h"
#include "message.h"
#include "dyn_ip.h"
#include "msgparser.h" 
#include "proxyarp.h"
#include "util.h"
#include "mn.h"
#ifdef INCLUDE_IPAY
#include "mn_ipay.h"
#endif

extern struct mn_data mn;
extern struct mn_config config;
extern struct timeval timers[TIMER_COUNT];
extern int mn_traces;
extern struct ro_entry ro_cache[RO_CACHE_MAX_SIZE];
extern int ro_entry_count;
extern struct nw_iface usable_nw_ifaces[NW_IFACE_ARRAY_LEN];
extern int usable_nw_iface_count;


static int fill_req_header(char *pos, int request_type, int use_reverse, struct in_addr coa)
{
    struct reg_req *req;
	int new_reg_time;
	unsigned int new_reg_rand;
	static int last_reg_time = 0;
	static unsigned int last_reg_rand = 0;
	struct timeval tv;

	req = (struct reg_req *) pos;
    req->type = REG_REQ;
    req->opts = 0;
	
	/* Note: the agent adv should be checked for allowed tunneling modes
	 * and req->opts should be adjusted with that information if needed
	 * (i.e. do not request unsupported mode) */
	if (!config.enable_fa_decapsulation)
		req->opts |= REGREQ_MN_DECAPS;
	if (use_reverse)
		req->opts |= REGREQ_REVERSE_TUNNEL;
	if (request_type != REG_DISC) {
		if (mn.tunnel_mode == API_TUNNEL_FULL_HA)
			req->lifetime =
				htons(config.mn_default_tunnel_lifetime);
		else
			req->lifetime =
				htons(MIN(mn.req_lifetime,
					  config.mn_default_tunnel_lifetime));
		if (ntohs(req->lifetime) == 0) {
			LOG2(LOG_INFO, "send_registration - trying to register"
			     ", but lifetime=0 - aborting\n");
			return -1;
		}
	} else 
		req->lifetime = 0;
		
 	/*----------------------------------------------------------*
 	 | Home Address												|
 	 *----------------------------------------------------------*/
 	req->home_addr.s_addr = config.mn_home_ip_addr.s_addr;
 	
 	/*----------------------------------------------------------*
 	 | HA's Address												|
 	 *----------------------------------------------------------*/
	if (config.ha_ip_addr.s_addr == 0 && config.use_aaa) {
		req->ha_addr.s_addr = config.allow_home_addr_from_foreign_net ?
			0 : -1;
	} else if (config.ha_ip_addr.s_addr == 0 &&
		   config.home_net_addr_plen > -1) {
		/* dynamic HA address resolution */
		req->ha_addr.s_addr = config.home_net_subnet_bc.s_addr;
	} else
		req->ha_addr.s_addr = config.ha_ip_addr.s_addr;

 	/*----------------------------------------------------------*
 	 | CoA														|
 	 *----------------------------------------------------------*/
	req->co_addr.s_addr = coa.s_addr;//mn.co_addr.s_addr;

 	/*----------------------------------------------------------*
 	 | ID														|
 	 *----------------------------------------------------------*/
	switch (config.replay_meth) {
	case REPLAY_PROTECTION_NONE:
		req->id[0] = 0;
		req->id[1] = get_rand32();
		break;

	case REPLAY_PROTECTION_TIMESTAMP:
		/* set the timestamp according to RFC 2002, 5.6.1 (NTP format):
		 * high-order 32 bits: time in seconds from year 1900,
		 * low-order 32 bits: fractional seconds (1/256 sec resolution,
		 * i.e., 8 bits, used) and the last bits random */
		gettimeofday(&tv, NULL);
		new_reg_time = tv.tv_sec + mn.clock_correction;
		new_reg_rand = ((tv.tv_usec * 256 / 1000000) << 24) |
			(get_rand32() & 0xffffff);
		if (new_reg_time == last_reg_time &&
		    new_reg_rand <= last_reg_rand) {
			/* id must always increase - forcing a larger id */
			new_reg_rand = last_reg_rand + 1;
		}
		req->id[0] = htonl(new_reg_time + UNIX_NTP_DIFF);
		req->id[1] = htonl(new_reg_rand);
		last_reg_time = new_reg_time;
		last_reg_rand = new_reg_rand;
		break;

	case REPLAY_PROTECTION_NONCE:
		req->id[0] = mn.last_ha_nonce;
		req->id[1] = get_rand32();
		break;
	}

	return sizeof(struct reg_req);
}


static int nai_equal(struct fa_nai_ext *nai1, struct fa_nai_ext *nai2)
{
	int len = GET_NAI_LEN(nai1);
	if (nai2 == NULL || GET_NAI_LEN(nai2) != len)
		return 0;
	if (memcmp(MSG_NAI_DATA(nai1), MSG_NAI_DATA(nai2), len) == 0)
		return 1;
	return 0;
}


static int add_localized_reg_extensions(char *pos, int left, char *start)
{
	static int last_req_seq_num = 0;
	struct registration_ext_dynamics *dyn_ext;
	int n, added = 0, add_auth = 1;

	if (mn.current_adv == NULL || mn.current_adv->adv.fa_nai == NULL ||
	    mn.last_req_FA_NAI->type == 0 ||
	    (!mn.prev_req_replied &&
	     !nai_equal(mn.last_req_FA_NAI, mn.current_adv->adv.fa_nai))) {
		if (mn.current_adv == NULL)
			DEBUG(DEBUG_INFO, "\tmn.current_adv == NULL\n");
		if (mn.current_adv != NULL &&
		    mn.current_adv->adv.fa_nai == NULL)
			DEBUG(DEBUG_INFO,
			      "\tcurrent_adv->fa_nai == NULL\n");
		if (mn.last_req_FA_NAI->type == 0)
			DEBUG(DEBUG_INFO, "\tlast_req_FA_NAI->type == 0\n");
		if (!mn.prev_req_replied)
			DEBUG(DEBUG_INFO, "\t!mn.prev_req_replied\n");
		if (mn.current_adv != NULL &&
		    !nai_equal(mn.last_req_FA_NAI, mn.current_adv->adv.fa_nai))
			DEBUG(DEBUG_INFO, "\t!nai_equal()\n");

		if (mn.current_adv == NULL ||
		    mn.current_adv->adv.fa_nai != NULL) {
			DEBUG(DEBUG_INFO, "Not adding SK auth because "
			      "current_adv == NULL or FA NAI advertised and "
			      "forcing req. to HA\n");
			add_auth = 0;
		}

		DEBUG(DEBUG_INFO, "Excluding previous FA NAI extension "
		      "(forcing request to HA)\n");
		mn.last_req_FA_NAI->type = 0;
	}

	if (left < sizeof(struct registration_ext_dynamics))
		return -1;

	DEBUG(DEBUG_MESSAGES, " * ext_dynamics\n");
	dyn_ext = (struct registration_ext_dynamics *) pos;
	dyn_ext->type = VENDOR_EXT_TYPE2;
	dyn_ext->length = sizeof(struct registration_ext_dynamics) - 2;
	dyn_ext->reserved = 0;
	dyn_ext->vendor_id = htonl(VENDOR_ID_DYNAMICS);
	dyn_ext->sub_type = htons(VENDOR_EXT_DYNAMICS_OPTIONS);
	dyn_ext->version = VENDOR_EXT_VERSION;
	dyn_ext->opts = 0;
	dyn_ext->seq = htonl(++last_req_seq_num);
	added += sizeof(struct registration_ext_dynamics);
	left -= sizeof(struct registration_ext_dynamics);

	if (mn.current_adv != NULL && mn.current_adv->adv.fa_nai != NULL) {
		/* previous FA NAI ext. */
		if (mn.prev_req_replied && mn.last_req_FA_NAI->type != 0) {
			n = GET_NAI_EXT_LEN(mn.last_req_FA_NAI);
			if (left < n)
				return -1;
			mn.last_req_FA_NAI->sub_type =
				htons(VENDOR_EXT_DYNAMICS_PREVIOUS_FA_NAI);
			memcpy(pos + added, mn.last_req_FA_NAI, n);
			added += n;
			left -= n;
			DEBUG(DEBUG_MESSAGES, " * previous FA NAI (len=%i)\n",
			      n);
		}

		/* current FA NAI ext. */
		n = GET_NAI_EXT_LEN(mn.current_adv->adv.fa_nai);
		memcpy(mn.last_req_FA_NAI, mn.current_adv->adv.fa_nai, n);
		if (left < n)
			return -1;
		memcpy(pos + added, mn.last_req_FA_NAI, n);
		added += n;
		left -= n;
		DEBUG(DEBUG_MESSAGES, " * current FA NAI (len=%i)\n", n);
	}

	if (add_auth) {
		if (left < sizeof(struct vendor_msg_auth) + MAX_SK_LEN)
			return -1;
		DEBUG(DEBUG_MESSAGES, " * sk_auth\n");
		n = auth_add_vendor(
			config.auth_alg, mn.session_key, mn.session_key_len,
			(unsigned char *) start,
			(struct vendor_msg_auth *) (pos + added),
			VENDOR_EXT_DYNAMICS_SK_AUTH, htonl(config.spi));
		added += n;
		left -= n;
	}

	return added;
}
/** 
 * add_nemo_extension: Add NEMO extensions if 
 *  - the NEMO nw list is not empty AND 
 *  - this request is not a deregistration 
 * 
 * Returns: the size of all the NEMO datagrams created 
 */
int add_nemo_extensions(int request_type, char *pos) 
{
	struct mobile_nw_req_ext *nemo_ext;
	struct nemo_nw_entry *nemo_nw_entry;
	int bytes_used = 0, ext_size;
	
	/* Check that there are NEMO networks and that this is not a deregistration */
	if (!list_is_empty(&config.nemo_nw_list) && request_type != REG_DISC) {
		
		/* Get the first NEMO nw entry */
		nemo_nw_entry = (struct nemo_nw_entry *) list_get_first(&config.nemo_nw_list);
		ext_size = sizeof(struct mobile_nw_req_ext);
		
		DEBUG(DEBUG_MESSAGES, " * mobile_nw_req_ext\n");
		
		/* Loop until an empty node in the list */
		while (nemo_nw_entry != NULL) { 
			DEBUG(DEBUG_MESSAGES, "    - %s/%d\n", 
				inet_ntoa(nemo_nw_entry->prefix), nemo_nw_entry->prefix_length);
			
			nemo_ext				= (struct mobile_nw_req_ext *) pos;
			nemo_ext->type 			= MOBILE_NW_EXT;	// dec: 148, hex: 94
			nemo_ext->length 		= 6;
			nemo_ext->subtype 		= 0;
			nemo_ext->prefix_length	= nemo_nw_entry->prefix_length;	// 0 - 32
			nemo_ext->prefix		= nemo_nw_entry->prefix.s_addr;	// IPv4 address in nw byte order

			pos 		+= ext_size;
			bytes_used 	+= ext_size;
			
			nemo_nw_entry = (struct nemo_nw_entry *) list_get_next((struct node *) nemo_nw_entry);
		}
		
	}
	return bytes_used;
}
/**
 * add_mr_cr_auth_extension: Add Mobile-Correspondent Authentication Extension.
 * 
 * Returns: the amount of bytes used for the extension in total.
 */
int add_mr_cr_auth_extension(struct ro_entry *cr, char *pos, char *reg_start) {
	struct mr_cr_auth_ext *auth_ext;
	int ext_size;
	__u64 auth_in_host_byte_order[2];
	int msg_len_until_auth_ext;
	int msg_len_until_auth_field;
	
	msg_len_until_auth_ext 		= (int) (pos - reg_start);
	msg_len_until_auth_field 	= msg_len_until_auth_ext + 8;
	
	auth_ext						= (struct mr_cr_auth_ext *) pos;
	auth_ext->type 					= TBA_T2;						/* 1 byte */
	auth_ext->length				= 22;							/* 1 byte */
	auth_ext->subtype				= TBA_ST_2_2;					/* 1 byte */
	auth_ext->reserved				= 0;							/* 1 byte */
	auth_ext->home_nonce_index		= htons(cr->hot_nonce_index); 	/* 2 bytes */
	auth_ext->care_of_nonce_index	= htons(cr->cot_nonce_index);	/* 2 bytes: Ignored in dereg */
	
	/* Authenticator: HMAC_MD5(KRm, (The whole Reg packet until this field)) */
	hmac_md5(cr->krm, 16, 								/* Key */
			reg_start, msg_len_until_auth_field, 		/* Input data */
			(unsigned char *) auth_in_host_byte_order);	/* Output code */
	
	DEBUG(DEBUG_MESSAGES, "add_mr_cr_auth_extension: KRm %llu %llu, auth %llu %llu\n",
		cr->krm[0], cr->krm[1], auth_in_host_byte_order[0], auth_in_host_byte_order[1]);
	
	/* Convert the authenticator to nw byte order */
	auth_ext->authenticator[0]	= hton64(auth_in_host_byte_order[1]);		/* [12 00] <- [43 21]*/
	auth_ext->authenticator[1]	= hton64(auth_in_host_byte_order[0]);		/* [12 34] <- [43 21] */
	
	/* return the size of the packet including the type and length fields */
	return (2 + auth_ext->length);
}

/* add_req_to_cr_extensions: Add extensions to an inter MR Registration Request.
 * The following extensions are added in the Request:
 *  - Mobile-Correspondent Authentication extension
 *  - Mobile Network Request Extension
 * 
 * The Authenticator field is hashed with the key KRm.
 */
static int add_inter_mr_req_extensions(struct ro_entry *cr, char *pos, int left, char *start,
			      int request_type, int use_reverse)
{
	char *orig = pos;
	struct udp_tnl_req_ext *udp_tnl_req;
	int tnl_req_len, mr_cr_auth_ext_len, nemo_ext_len;
	
	/*--------------------------------------------------------------*
	 | UDP Tunneling Request Extension								|
	 *--------------------------------------------------------------*/
	tnl_req_len = sizeof(struct udp_tnl_req_ext);
	if (left < tnl_req_len) { 
		DEBUG(DEBUG_MESSAGES, "Not enough space to add the UDP TNL Request extension!\n");
		return -1;
	}
	udp_tnl_req 				= (struct udp_tnl_req_ext *) pos;
	udp_tnl_req->type			= UDP_TNL_REQ_EXT;
	udp_tnl_req->length			= 6;
	udp_tnl_req->subtype		= 0;
	udp_tnl_req->reserved1		= 0;
	udp_tnl_req->reserved2		= 0x80;	// Force-flag set by default
	udp_tnl_req->encapsulation	= 4;
	udp_tnl_req->reserved3		= 0;

	pos 	+= tnl_req_len;
	left 	-= tnl_req_len;
	
	/*--------------------------------------------------------------*
	 | Mobile Network Request (NEMO) Extension						|
	 *--------------------------------------------------------------*/
	nemo_ext_len = add_nemo_extensions(request_type, pos);
	if (left < nemo_ext_len) { 
		DEBUG(DEBUG_MESSAGES, "Not enough space to add the NEMO extension(s)!\n");
		return -1;
	}
	pos 	+= nemo_ext_len;
	left 	-= nemo_ext_len;
	
	/*--------------------------------------------------------------*
	 | Mobile-Correspondent Authentication extension				|
	 *--------------------------------------------------------------*/
	mr_cr_auth_ext_len = add_mr_cr_auth_extension(cr, pos, start);
	if (left < mr_cr_auth_ext_len) { 
		DEBUG(DEBUG_MESSAGES, "Not enough space to add the MR-CR Authentication ext!\n");
		return -1;
	}
	pos 	+= mr_cr_auth_ext_len;
	left 	-= mr_cr_auth_ext_len;
	
	return (int) (pos - orig);
}
static int add_req_extensions(char *pos, int left, char *start,
			      int request_type, int use_reverse)
{
	int use_dynamics_ext, n;
	struct fa_spi_entry *fa_spi = NULL;
	struct msg_key *mn_keyreq;
	struct challenge_ext *challenge = NULL;
	char *orig = pos;
	int add_mn_aaa_auth_ext = 0;
	int send_to_ha = 0;
	
	if (request_type == REG_DISC &&
	    mn.fa_addr.s_addr == config.ha_ip_addr.s_addr)
		send_to_ha = 1;
		
	if (config.priv_ha > 0) {
		struct priv_ha_ext *priv = (struct priv_ha_ext *) pos;
		if (left < sizeof(struct priv_ha_ext))
			return -1;
		memset(priv, 0, sizeof(struct priv_ha_ext));
		priv->type = VENDOR_EXT_TYPE2;
		priv->length = sizeof(struct priv_ha_ext) - 2;
		priv->vendor_id = htonl(VENDOR_ID_DYNAMICS);
		priv->sub_type = htons(VENDOR_EXT_DYNAMICS_PRIV_HA);
		priv->priv_ha = htonl(config.priv_ha);
		pos += GET_PRIV_HA_EXT_LEN(priv);
		left -= GET_PRIV_HA_EXT_LEN(priv);
		DEBUG(DEBUG_MESSAGES, " * priv_ha\n");
	}

	if (config.mn_nai_len > 0) {
		struct mn_nai_ext *nai;
		if (left < sizeof(struct mn_nai_ext) + config.mn_nai_len)
			return -1;
		nai = (struct mn_nai_ext *) pos;
		nai->type = MN_NAI_EXT;
		nai->length = config.mn_nai_len;
		memcpy(nai + 1, config.mn_nai, config.mn_nai_len);
		pos += GET_MN_NAI_EXT_LEN(nai);
		left -= GET_MN_NAI_EXT_LEN(nai);
		DEBUG(DEBUG_MESSAGES, " * mn_nai\n");
	}

#ifdef INCLUDE_IPAY
	if (config.mn_nai_len == 0 && mn.nai.nai != NULL) {
		struct mn_nai_ext *nai;
		if (left < sizeof(struct mn_nai_ext) + mn.nai.len)
			return -1;
		nai = (struct mn_nai_ext *) pos;
		nai->type = MN_NAI_EXT;
		nai->length = mn.nai.len;
		memcpy(nai + 1, mn.nai.nai, mn.nai.len);
		pos += GET_MN_NAI_EXT_LEN(nai);
		left -= GET_MN_NAI_EXT_LEN(nai);
		DEBUG(DEBUG_MESSAGES, " * mn_nai (Ipay)\n");
	}
#endif

	/* add Dynamics extensions if the Foreign Agent advertised support for
	 * them or if the registration is made directly to the Home Agent */
	use_dynamics_ext = mn.fa_dynamics_ext.type != 0 ||
		mn.tunnel_mode == API_TUNNEL_FULL_HA;

	if (use_dynamics_ext) {
		// add mn_keyreq extension
		if (left < MIN_KEY_EXT_LEN)
			return -1;
		DEBUG(DEBUG_MESSAGES, " * mn_keyreq\n");
		mn_keyreq = (struct msg_key *) pos;
		init_key_extension(mn_keyreq, VENDOR_EXT_DYNAMICS_MN_KEYREQ,
				   htonl(config.spi), 0);
		pos += GET_KEY_EXT_LEN(mn_keyreq);
		left -= GET_KEY_EXT_LEN(mn_keyreq);

		// add HFA public key hash extension if the hash is available
		// from the agent advertisement
		if (mn.current_adv != NULL) {
			struct msg_key *hashext =
				mn.current_adv->adv.pubkey_hash;
			if (hashext != NULL && hashext->type != 0) {
				if (left < GET_KEY_EXT_LEN(hashext))
					return -1;
				DEBUG(DEBUG_MESSAGES, " * pubkey_hash\n");
				memcpy(pos, hashext, GET_KEY_EXT_LEN(hashext));
				pos += GET_KEY_EXT_LEN(hashext);
				left -= GET_KEY_EXT_LEN(hashext);
			}
		}
	} else {
		DEBUG(DEBUG_INFO, "FA did not advertise support for Dynamics "
		      "extensions - not using them\n");
	}
	
	if (config.use_aaa &&
	    config.mn_ha_key_timestamp != 0 &&
	    config.mn_ha_key_lifetime != 0 &&
	    config.mn_ha_key_timestamp + config.mn_ha_key_lifetime <
	    time(NULL)) 
	{
		DEBUG(DEBUG_INFO, "Dynamic MN-HA security association (from "
		      "AAA) expired\n");
		config.shared_secret_len = -1;
	}
			
	/*--------------------------------------------------------------*
	 | UDP Tunnel Request Extension.								|
	 *--------------------------------------------------------------*/
	n = sizeof(struct udp_tnl_req_ext);
	if (left < n) return -1;
	
	struct udp_tnl_req_ext *tnl_req = (struct udp_tnl_req_ext *) pos;
	tnl_req->type 			= 144;
	tnl_req->length 		= 6;
	tnl_req->subtype 		= 0;
	tnl_req->reserved1 		= 0;
	tnl_req->reserved2		= 0x80; // Force-flag set by default
	tnl_req->encapsulation	= 4;
	tnl_req->reserved3		= htons(0);

	pos 	+= n;
	left 	-= n;
	
	DEBUG(DEBUG_MESSAGES, " * udp_tnl_req\n");	
	
	/*--------------------------------------------------------------*
	 | Mobile Network Request (NEMO) Extension.						|
	 *--------------------------------------------------------------*/
	int nemo_ext_len = add_nemo_extensions(request_type, pos);
	if (left < nemo_ext_len) { 
		DEBUG(DEBUG_MESSAGES, "Not enough space to add the NEMO extension(s)\n!");
		return -1;
	}
	pos 	+= nemo_ext_len;
	left 	-= nemo_ext_len;
	
	/*--------------------------------------------------------------*
	 | MR Route Optimization Capability (HAaRO) Extension.			|
	 *--------------------------------------------------------------*/
	/* Check if there are CRs that need to be solicited from HA 
	 * (RO Cache entries with confirmed_by_ha = 0) 
	 */
	int no_cr_to_solicit = 1;	/* 0: Send RO Capability with S=0 */
	int i;
	
	//DEBUG(DEBUG_MESSAGES, "Iterate through RO Cache to look for CRs to solicit\n");
	for (i = 0; i < ro_entry_count; i++) {
		//DEBUG(DEBUG_MESSAGES, " - ro_cache[%d]", i);
		if (ro_cache[i].confirmed_by_ha == 0)
		{
			//DEBUG(DEBUG_MESSAGES, ": needs to be solicited -> add RO Capab with S=1\n");
			no_cr_to_solicit = 0;
			DEBUG(DEBUG_MESSAGES, " * ro_capab_ext (S=1): solicit CR %s\n",
				inet_ntoa(ro_cache[i].cr_hoa));
			print_trace(mn_traces, " - solicit CR %s from HA\n", inet_ntoa(ro_cache[i].cr_hoa));
			/* Send RO Capability with S=1 */
			struct ro_capab_s_on *ro_capab_s = (struct ro_capab_s_on *) pos;
			
			n = sizeof(struct ro_capab_s_on);
			if (left < n) return -1;
			
			ro_capab_s->type	= TBA_T1;
			ro_capab_s->length	= 6;
			ro_capab_s->subtype	= TBA_ST1_1;
			ro_capab_s->flags	= 0 | RO_CAPAB_S | RO_CAPAB_R;
			ro_capab_s->mr_hoa	= ro_cache[i].cr_hoa.s_addr;
			
			pos 	+= n;
			left 	-= n;
		} else {
		//DEBUG(DEBUG_MESSAGES, ": no need to solicit\n");
		}
	} 
	if (no_cr_to_solicit) 
	{	
		//DEBUG(DEBUG_MESSAGES, " - no RO entries to solicit -> Add RO Capab with S=0\n");
		/* Send RO Capability with S=0 */
		struct ro_capab_s_off *ro_capab = (struct ro_capab_s_off *) pos;
		
		n = sizeof(struct ro_capab_s_off);
		if (left < n) return -1;
		
		DEBUG(DEBUG_MESSAGES, " * ro_capab_ext (S=0)\n");
		
		ro_capab->type		= TBA_T1;
		ro_capab->length	= 2;
		ro_capab->subtype	= TBA_ST1_1;
		ro_capab->flags		= 0 | RO_CAPAB_A | RO_CAPAB_R; // |A|R|S|O|  Rsvd |			
		
		pos 	+= n;
		left 	-= n;
	}	
	/* -- /HAaRO -- */

	if (config.shared_secret_len >= 0) {
		// add MN->HA authentication extension
		if (left < sizeof(struct msg_auth) + MAX_SK_LEN)
			return -1;
		DEBUG(DEBUG_MESSAGES, " * mh_auth, alg %d\n", mn.use_auth_alg);
		n = auth_add(mn.use_auth_alg, config.shared_secret,
			     config.shared_secret_len, (unsigned char *) start,
			     (struct msg_auth *) pos, MH_AUTH,
			     htonl(config.spi));
		pos += n;
		left -= n;
	}
	
	/* Add encapsulating delivery extension if MN decaps and reverse
	 * tunnel is in use [RFC 2344, Chap. 3.3] */
	/**/
	if (!config.enable_fa_decapsulation && use_reverse) {
		struct encaps_delivery_ext *edel =
			(struct encaps_delivery_ext *) pos;
		if (left < sizeof(struct encaps_delivery_ext))
			return -1;
		DEBUG(DEBUG_MESSAGES, " * encaps_delivery\n");
		edel->type = ENCAPS_DELIVERY_EXT;
		edel->length = 0;
		pos += sizeof(struct encaps_delivery_ext);
		left -= sizeof(struct encaps_delivery_ext);
	}
	
	/* add FA NAI extensions and session key based MN->FA authentication
	 * if Dynamics extensions are used and the session key is available
	 * and this is not a reregistration */
	if (use_dynamics_ext && mn.session_key != NULL &&
	    request_type != REG_REREG) {
		n = add_localized_reg_extensions(pos, left, start);
		if (n < 0)
			return -1;
		pos += n;
		left -= n;
	} else if (mn.current_adv != NULL &&
		   mn.current_adv->adv.fa_nai != NULL) {
		/* current FA NAI ext. */
		memcpy(mn.last_req_FA_NAI, mn.current_adv->adv.fa_nai,
		       MAX_FA_NAI_LEN);
		n = GET_NAI_EXT_LEN(mn.last_req_FA_NAI);
		if (left < n)
			return -1;
		memcpy(pos, mn.last_req_FA_NAI, n);
		pos += n;
		left -= n;
		DEBUG(DEBUG_MESSAGES, " * current FA NAI (len=%i)\n", n);
	}

	/* if the agent advertisement from the FA has a Challenge extension,
	 * copy it to the registration request or if the previous registration
	 * reply had a more recent Challenge ext, use it */
	/* FIX: MN must not use the same challenge again, so if no new
	 * challenge is available, it is no use sending this registration
	 * without first acquiring a new challenge fron an agent advertisement.
	 * MN could send an agent soliciation in that case. Now the request
	 * will be denied by FA and MN might get a new challenge from that
	 * denial reply. */
	if (mn.last_challenge_ext != NULL &&
	    (mn.current_adv == NULL ||
	     cmp_timeval(&mn.current_adv->last, &mn.last_challenge_time) < 0))
	{
		DEBUG(DEBUG_INFO, "Using challenge from last reg. reply\n");
		challenge = mn.last_challenge_ext;
	} else if (mn.current_adv != NULL &&
		   mn.current_adv->adv.challenge != NULL) {
		DEBUG(DEBUG_INFO, "Using challenge from last agentadv\n");
		challenge = mn.current_adv->adv.challenge;
	} else
		challenge = NULL;

	if (!send_to_ha)
		fa_spi = get_fa_spi(0, mn.fa_addr);

	if (challenge != NULL && (fa_spi != NULL || config.use_aaa)) {
		n = GET_CHALLENGE_EXT_LEN(challenge);
		DEBUG(DEBUG_MESSAGES, " * challenge (len=%i)\n", n);
		if (left < n)
			return -1;
		memcpy(pos, challenge, n);
		challenge = (struct challenge_ext *) pos;
		challenge->type = MN_FA_CHALLENGE_EXT;
		pos += n;
		left -= n;
		add_mn_aaa_auth_ext = 1;
	}

	/* add shared secret based MN->FA authentication if the security
	 * association is configured */
	
	if (!send_to_ha && fa_spi != NULL && !mn.aaa_rekey) {
		if (left < sizeof(struct msg_auth) + MAX_SK_LEN)
			return -1;
		DEBUG(DEBUG_MESSAGES, " * mf_auth\n");
		n = auth_add(fa_spi->alg, fa_spi->shared_secret,
			     fa_spi->shared_secret_len,
			     (unsigned char *) start, (struct msg_auth *) pos,
			     MF_AUTH, htonl(fa_spi->spi));
		pos += n;
		left -= n;
		add_mn_aaa_auth_ext = 0;
	} else if (!send_to_ha && config.use_aaa) {
		struct generalized_mn_fa_key_req_ext *keyreq;

		add_mn_aaa_auth_ext = 1;

		// add MN-FA key req. from AAA
		keyreq = (struct generalized_mn_fa_key_req_ext *) pos;
		if (left < sizeof(*keyreq))
			return -1;
		DEBUG(DEBUG_MESSAGES, " * MN-FA Key Req from AAA\n");
		keyreq->type = GENERALIZED_MN_FA_KEY_REQ_EXT;
		keyreq->subtype = GEN_MN_FA_KEY_REQ_FROM_AAA;
		keyreq->length = htons(4);
		// FIX: mn_spi is the SPI that MN will assign for the security
		// association; this may need to be uniquely selected for some
		// cases(?)
		keyreq->mn_spi = htonl(1000);
		pos += GET_GEN_MN_FA_KEY_REQ_EXT_LEN(keyreq);
		left -= GET_GEN_MN_FA_KEY_REQ_EXT_LEN(keyreq);
	}
	
	
	if ((config.shared_secret_len < 0 || mn.aaa_rekey) &&
	    config.use_aaa) {
		struct generalized_mn_ha_key_req_ext *keyreq;

		add_mn_aaa_auth_ext = 1;

		// add MN-HA key req. from AAA
		keyreq = (struct generalized_mn_ha_key_req_ext *) pos;
		if (left < sizeof(*keyreq))
			return -1;
		DEBUG(DEBUG_MESSAGES, " * MN-HA Key Req from AAA\n");
		keyreq->type = GENERALIZED_MN_HA_KEY_REQ_EXT;
		keyreq->subtype = GEN_MN_HA_KEY_REQ_FROM_AAA;
		keyreq->length = htons(4);
		// FIX: mn_spi is the SPI that MN will assign for the security
		// association; this may need to be uniquely selected for some
		// cases(?)
		keyreq->mn_spi = htonl(1000);
		pos += GET_GEN_MN_FA_KEY_REQ_EXT_LEN(keyreq);
		left -= GET_GEN_MN_FA_KEY_REQ_EXT_LEN(keyreq);
	}
	
	
	if (add_mn_aaa_auth_ext && config.use_aaa) {
		// add MN-AAA authentication extension
		if (left < sizeof(struct generalized_auth_ext) + MAX_SK_LEN)
			return -1;
		DEBUG(DEBUG_MESSAGES, " * gen_auth_ext(MN-AAA), alg %d\n", config.mn_aaa_auth_alg);
		n = auth_add_gen(config.mn_aaa_auth_alg,
				 config.mn_aaa_shared_secret,
				 config.mn_aaa_shared_secret_len,
				 (unsigned char *) start,
				 challenge != NULL ?
				 MSG_CHALLENGE_EXT_DATA(challenge) : NULL,
				 challenge != NULL ?
				 GET_CHALLENGE_LEN(challenge) : 0,
				 (struct generalized_auth_ext *) pos,
				 GENERALIZED_AUTH_MN_AAA,
				 htonl(config.mn_aaa_spi));
		pos += n;
		left -= n;
	}

	mn.prev_req_replied = 0;
	return (int) (pos - orig);
}

/* send_from_mr_hoa_mip_port_to_ha_tunnel: Add UDP in IP header and write to tunnel.
 *
 * @dst_port: in nw byte order
 */
int send_from_mr_hoa_mip_port_to_ha_tunnel(char *packet, int packet_len, 
			struct in_addr dst_addr, unsigned short dst_port)
{
	int buf_size = 35 + packet_len;	// IP + UDP + MIP TNL ~= 35
	char buffer[buf_size];
	int total_len;
	
	/* Fill an UDP in IP packet */
	total_len = fill_udp_in_ip(buffer, config.mn_home_ip_addr, dst_addr, 
				htons(434), dst_port, packet, packet_len);
	
	/* Send the packet to tunnel to HA */
	return from_buffer_to_ha_tunnel(buffer, total_len);
}

int send_from_mr_coa_mip_port_to_cr_hoa(struct ro_entry *cr, unsigned char *send_buf, int len) 
{
	return send_udp_to(send_buf, len, cr->cr_hoa, cr->cr_co_port, cr->coa_mip_socket);
}
/* MR CoA : 434 -> CR CoA : reg_port */
int send_from_mr_coa_mip_port_to_cr_coa(struct ro_entry *cr, unsigned char *send_buf, int len) 
{
	return send_udp_to(send_buf, len, cr->tnl_dst, cr->cr_co_port, cr->coa_mip_socket);
}
int send_to_cr_hoa_mip_port(struct ro_entry *cr, unsigned char *msg, int msg_len, int socket) {
	struct sockaddr_in cr_hoa;
	int sent_bytes;
	
	assert(cr != NULL && msg != NULL && msg_len >= 0);
	
	/* Fill the receiver info */
	memset(&cr_hoa, 0, sizeof(cr_hoa));
	cr_hoa.sin_family 		= AF_INET;
	cr_hoa.sin_addr.s_addr 	= cr->cr_hoa.s_addr;
	cr_hoa.sin_port 		= htons(434);
		
	DEBUG(DEBUG_MESSAGES, "send_to_cr_hoa_mip_port: send to %s:434 via fd %d\n", 
		inet_ntoa(cr->cr_hoa), socket);
		
	/* Send the CoTI message (src IP: MR CoA) */
	sent_bytes = sendto(socket, msg, msg_len, 0,
    		(struct sockaddr *) &cr_hoa, sizeof(struct sockaddr_in));
	if (sent_bytes < 0) {
 		LOG2(LOG_ERR, "send_to_cr_hoa - sendto: %s\n", strerror(errno));
 	}
	return sent_bytes;
}

/* Send the HoTI message (src IP: MR HoA) */
int send_from_mr_hoa_to_cr_hoa_mip_port(struct ro_entry *cr, unsigned char *msg, int msg_len)
{ 
 	return send_to_cr_hoa_mip_port(cr, msg, msg_len, mn.hoa_socket);
}

/* send_hoti_coti: Send HoTI (from MR HoA) and CoTI (from MR CoA) messages 
 * towards CR HoA:434. 
 * 
 * @cr
 * @coa_struct: If this parameter is not NULL, coa_struct->coa is used 
 * as the src IP address.
 *  
 */
int send_hoti_coti(struct ro_entry *cr, struct mr_coa *coa_struct) 
{
	unsigned char hoti_buf[200];
	struct ro_hoti hoti_msg;
	struct ro_coti coti_msg;
    struct sockaddr_in cr_hoa;
   	struct timeval *init_sending_time;
	unsigned short dst_port	= 434;
	int hoti_len = sizeof(struct ro_hoti);
	int coti_len = sizeof(struct ro_coti);
	int hoti_sent, coti_sent;
	int coti_socket;
	__u64 hoti_cookie;
	__u64 coti_cookie;
	
	/* If Return routability is switched off, send Registration Request directly */
	if (config.disable_rr)
	{	
		print_trace(mn_traces, "RR turned off -> do registration without HoTI + CoTI\n");
		/* Check if if the registration socket is binded to a different iface than CoTI was sent through */
		if (cr->different_coti_iface)
		{
			bind_cr_reg_connection_to_device(cr, cr->last_coti_iface);
		}	
		cr->connection_initialized 	= 1;
		cr->init_send_time.tv_sec 	= 0;
		cr->init_send_time.tv_usec 	= 0;
		return send_regreq_to_cr(cr, REG_CONNECT);
	}

	hoti_cookie 		= (((__u64) rand()) << 32 ) + (__u64) rand();
	coti_cookie 		= (((__u64) rand()) << 32 ) + (__u64) rand();
	coti_socket 		= cr->coti_socket;
	init_sending_time 	= &cr->init_send_time;
	
	cr->hoti_cookie = hoti_cookie;
	cr->coti_cookie = coti_cookie;
	
	DEBUG(DEBUG_MESSAGES, "Send HoTI and CoTI to CR HoA %s/%d\n",
		inet_ntoa(cr->cr_hoa), dst_port);
	
	/* Fill the HoTI message buffer */	
	hoti_msg.type 		= TBA_MIP1;
	hoti_msg.reserved	= 0;
	hoti_msg.cookie		= hoti_cookie;
	
	/* Fill the CoTI message buffer */
	coti_msg.type 		= TBA_MIP2;
	coti_msg.reserved	= 0;
	coti_msg.cookie		= coti_cookie;

	DEBUG(DEBUG_MESSAGES, "HoTI cookie %llu\n",	ntoh64(hoti_cookie)); 
	DEBUG(DEBUG_MESSAGES, "CoTI cookie %llu\n", ntoh64(coti_cookie)); 
		
	/* Send the HoTI message (src IP: MR HoA) */
	hoti_sent = send_from_mr_hoa_to_cr_hoa_mip_port(cr, (unsigned char *) &hoti_msg, hoti_len); 
	
 	if (hoti_sent != hoti_len) {
 		LOG2(LOG_ERR, "sending HoTI - sendto: %s\n", strerror(errno));
        return -1;	
 	}
	DEBUG(DEBUG_MESSAGES, "\tHoTI sent (%d bytes)\n", hoti_sent);
	
	/* Fill the receiver info */
	memset(&cr_hoa, 0, sizeof(cr_hoa));
	cr_hoa.sin_family 		= AF_INET;
	cr_hoa.sin_addr.s_addr 	= cr->cr_hoa.s_addr;
	cr_hoa.sin_port 		= htons(dst_port);
	
	/* Send the CoTI message (src IP: MR CoA) */
	coti_sent = send_to_cr_hoa_mip_port(cr, (unsigned char *) &coti_msg, coti_len, coti_socket); 
	DEBUG(DEBUG_MESSAGES, "\tCoTI sent (%d bytes)\n", coti_sent);
	
	print_trace(mn_traces, "HoTI + CoTI sent to CR %s through %s\n", inet_ntoa(cr->cr_hoa), cr->last_coti_iface);
	gettimeofday(init_sending_time, NULL);
	
	return 0;
}

/* send_registration:
 * @request_type: 
 *    REG_DISC    - deregistration
 *    REG_CONNECT - normal connection request
 *    REG_REREG   - reregistration (do not send sk_auth extension
 *                  and ask full time)
 *
 * Send registration request message.
 * 
 * Returns: 0 on success, -1 on error.
 */
int send_registration(int request_type)
{
    char sendbuf[MAXMSG];
	char *pos; /* pointer to next byte to write in sendbuf */
    int n, len;
    struct sockaddr_in addr;
	int use_reverse;
	struct msg_extensions ext;
	
	if (MAXMSG < (sizeof(struct reg_req) +
    	2 * (sizeof(struct msg_auth) + 16))) 
    {
        DEBUG(DEBUG_INFO, "MN: send_reg: too small send buffer!");
        return -1;
  	}

	/* Use reverse tunneling if tunneling mode is Reverse or Auto Reverse */
	use_reverse = 	config.tunneling_mode == TUNMODE_AUTO_REVERSE ||
					config.tunneling_mode == TUNMODE_REVERSE;

	if (mn.current_adv != NULL) {
		if (mn.current_adv->addr.s_addr != mn.fa_addr.s_addr) {
			DEBUG(DEBUG_INFO, "fa_addr=%s != ",
			      inet_ntoa(mn.fa_addr));
			DEBUG(DEBUG_INFO, "current_adv->addr=%s\n",
			      inet_ntoa(mn.current_adv->addr));
		}

		DEBUG(DEBUG_INFO, "Setting in_use=%i for FA %s (send reg)\n",
		      request_type == REG_DISC ? 0 : 1,
		      inet_ntoa(mn.current_adv->addr));
		mn.current_adv->in_use = request_type == REG_DISC ? 0 : 1;
		if (config.tunneling_mode == TUNMODE_AUTO_REVERSE &&
		    (ntohs(mn.current_adv->adv.ext->opts) &
		     AGENT_ADV_BIDIR_TUNNELING) == 0)
			use_reverse = 0;
		if (config.tunneling_mode == TUNMODE_AUTO_TRIANGLE &&
		    mn.current_adv->adv.own_ext != NULL &&
		    (mn.current_adv->adv.own_ext->opts &
		     AGENT_ADV_OWN_TRIANGLE_TUNNELING) == 0)
			use_reverse = 1;
	} else {
		DEBUG(DEBUG_INFO,
		      "send_registration: current_adv == NULL!?\n");
	}

	if ((use_reverse && config.tunneling_mode == TUNMODE_TRIANGLE) ||
	    (!use_reverse && config.tunneling_mode == TUNMODE_REVERSE)) {
		DEBUG(DEBUG_INFO, "FA does not support wanted tunneling mode -"
		      " aborting registration\n");
		if (mn.current_adv != NULL)
			mn.current_adv->reg_failed = 1;
		return -1;
	}

	if (use_reverse)
		DEBUG(DEBUG_MESSAGES, "Requesting reverse tunneling\n");

	pos = sendbuf;

	DEBUG(DEBUG_MESSAGES, "Sending registration message to HA:\n * header\n");
	n = fill_req_header(pos, request_type, use_reverse, mn.co_addr);
	if (n < 0) return -1;
	pos += n;

	n = add_req_extensions(pos, MAXMSG - n, sendbuf, request_type,
			       use_reverse);
	if (n < 0) {
		LOG2(LOG_ERR, "Registration request buffer overflow - unable"
		     " to send the request\n");
		return -1;
	}
	pos += n;

	len = (int) (pos - sendbuf);
	assert(len <= MAXMSG);

	/* call parse_msg() to parse the message in order to get the details
	 * of the sent message into debug output; in addition, this validates
	 * the format of the request */
	int parse_res = parse_msg(sendbuf, len, &ext);
	if (parse_res != 0) {
		DEBUG(DEBUG_INFO, "WARNING! Own registration request was not "
		      "valid!\n");
	}

	/* add ARP entry, if not already added */
	if (mn.current_adv != NULL &&
	    mn.current_adv->adv_type == MN_ADV_TYPE_FA &&
	    !mn.current_adv->arpentry) 
	{
		struct sockaddr hwaddr;
		DEBUG(DEBUG_INFO, "Adding ARP entry for FA\n");
		memset(&hwaddr, 0, sizeof(hwaddr));
		hwaddr.sa_family = ARPHRD_ETHER;
#ifdef DYN_TARGET_LINUX
		memcpy(hwaddr.sa_data, mn.current_adv->adv.from.sll_addr,
		       mn.current_adv->adv.from.sll_halen);
#else
		memcpy(hwaddr.sa_data, mn.current_adv->adv.eth->h_source,
		       ETH_ALEN);
#endif
		if (arp_add_permanent_item(mn.current_adv->addr,
					   mn.current_adv->ifname,
					   &hwaddr) < 0) 
		{
			LOG2(LOG_WARNING, "arp_add_permanent_item(%s, %s, %s) "
			     "failed\n", inet_ntoa(mn.current_adv->addr),
			     mn.current_adv->ifname,
			     ether_hwtoa((unsigned char *)hwaddr.sa_data));
		}
		mn.current_adv->arpentry = 1;
	}

	/* add host route to FA, if not already added */
	if (config.enable_fa_decapsulation &&
	    mn.current_adv != NULL &&
	    mn.current_adv->adv_type == MN_ADV_TYPE_FA &&
	    !mn.current_adv->routeentry) {
		DEBUG(DEBUG_INFO, "Adding routing entry for FA\n");
		add_fa_host_route(mn.current_adv->ifname, mn.agentadv,
				  mn.current_adv->ifindex, mn.fa_addr);
	}

	/* send the message */
	memset(&addr, 0, sizeof(addr));
	addr.sin_family 		= AF_INET;
	addr.sin_addr.s_addr 	= mn.fa_addr.s_addr;
	addr.sin_port 			= htons(config.udp_port);
	DEBUG(DEBUG_MESSAGES, "sending registration request to %s:%i, "
	      "type=%i\n", inet_ntoa(addr.sin_addr), ntohs(addr.sin_port),
	      request_type);

	gettimeofday(&mn.last_reg_send_time, NULL);
	DEBUG(DEBUG_TIMERS, "last_reg_send_time=%li.%06li (FA=%s)\n",
	      mn.last_reg_send_time.tv_sec,
	      mn.last_reg_send_time.tv_usec,
	      inet_ntoa(addr.sin_addr));

	if (ext.req != NULL) {
		print_trace(mn_traces, "Sending registration request to HA through %s: %s\n", 
			mn.device_in_use, get_req_str(ext.req));
	}
	
 	n = sendto(mn.registration_socket, sendbuf, len, 0,
    		(struct sockaddr *) &addr, sizeof(addr));
 	if (n == len) {
		/* store id for reply */
        mn.registration_id[0] =
			ntohl(((struct reg_req *)sendbuf)->id[0]);
            mn.registration_id[1] =
			ntohl(((struct reg_req *)sendbuf)->id[1]);
		/* set registration time */
		gettimeofday(&timers[TIMER_REQUEST], NULL);
		mn.last_request_sent = timers[TIMER_REQUEST].tv_sec;
        return 0;
	} else {
		LOG2(LOG_ERR, "send_reg - sendto: %s\n", strerror(errno));
        return -1;
 	}
}


/* send_regreq_to_cr: Send Registration Request from MR CoA to CR HoA.
 * @cr
 * @request_type: 
 *    REG_DISC    - deregistration
 *    REG_CONNECT - normal connection request
 *    REG_REREG   - reregistration (do not send sk_auth extension
 *                  and ask full time)
 *
 */
int send_regreq_to_cr(struct ro_entry *cr, int request_type) {

	char sendbuf[MAXMSG];
	char *pos = sendbuf;		/* pointer to next byte to write in sendbuf */
    int n, len;
	int use_reverse = 0;
    struct sockaddr_in dst_addr;
    struct msg_extensions ext;
    
	DEBUG(DEBUG_MESSAGES, "Sending Registration Request to CR HoA %s:434\n",
		inet_ntoa(cr->cr_hoa));

	/* Use reverse tunneling if tunneling mode is Reverse or Auto Reverse */
	use_reverse = 	config.tunneling_mode == TUNMODE_AUTO_REVERSE ||
					config.tunneling_mode == TUNMODE_REVERSE;
		
	if (use_reverse)
		DEBUG(DEBUG_MESSAGES, " - Requesting reverse tunneling\n");
		
	/*----------------------------------------------*
	 | Registration Request header into the buffer	|
	 *----------------------------------------------*/
	DEBUG(DEBUG_MESSAGES, " * header\n");
	
	n = fill_req_header(pos, request_type, use_reverse, cr->mr_coa);
	
	if (n < 0) return -1;
	pos += n;

	/*----------------------------------------------*
	 | Extensions into the buffer					|
	 *----------------------------------------------*/
	n = add_inter_mr_req_extensions(cr, pos, MAXMSG - n, sendbuf, request_type,
			       use_reverse);
	if (n < 0) {
		LOG2(LOG_ERR, "Registration request buffer overflow - unable"
		     " to send the request!\n");
		return -1;
	}
	pos += n;

	len = (int) (pos - sendbuf);
	assert(len <= MAXMSG);
	
	/*----------------------------------------------*
	 | Validate the message format					|
	 *----------------------------------------------*/
	int parse_res = parse_msg(sendbuf, len, &ext);
	if (parse_res != 0) {
		DEBUG(DEBUG_INFO, "WARNING! Own registration request was not "
		      "valid!\n");
	}
	/*----------------------------------------------------------------------*
	 | Save the Request in the RO Cache to be able to identify the Reply	|
	 *----------------------------------------------------------------------*/
	cr->sent_reg_id[0] = ext.req->id[0];
	cr->sent_reg_id[1] = ext.req->id[1];
	
	if (ext.req != NULL) {
		print_trace(mn_traces, "Sending registration request to CR %s through %s: %s\n", 
			inet_ntoa(cr->cr_hoa), cr->device_in_use, get_req_str(ext.req));
	}
	
	/*------------------------------------------------------*
	 | Set regreq_online for the PRE_ACCEPT reply system.	|
	 *------------------------------------------------------*/
	cr->regreq_online = 1;
	set_cr_reply_wait_timer_expiration(cr);
	
	/*----------------------------------------------*
	 | Send the buffer data from MR CoA	to CR CoA	|
	 *----------------------------------------------*/
	int sent_bytes = send_to_cr_hoa_mip_port(cr, (unsigned char *) sendbuf, len, cr->registration_socket); 
	
	DEBUG(DEBUG_INFO, "Registration Request sent (%d bytes, id %08x %08x)\n", sent_bytes, 
		(__u32) ntohl(cr->sent_reg_id[0]), 
		(__u32) ntohl(cr->sent_reg_id[1])); 
	
	return 0;
}

/* send_regrep_to_cr: Send Registration Reply from MR CoA to CR HoA and
 * clear the information saved from the CR's Registration Request. 
 * Function is called when either a Reg Req from a CR or a MR Solicitation
 * Reply from the HA is received.
 * 
 * @cr: The CR this Reply is sent to.  
 */
int send_regrep_to_cr(struct ro_entry *cr) {
	unsigned char send_buf[MAXMSG];
	struct msg_extensions ext_for_parse_msg;	/* Container for the parse_msg function */
	unsigned char *pos = send_buf;				/* pointer to the next byte to write in sendbuf */
	struct reg_rep *reply;
	struct udp_tnl_rep_ext *tnl_reply;
	struct mobile_nw_ack_ext *nemo_ack;
	struct in_addr *current_nemo;
	int current_nemo_len;
	struct coa_ext *coa_ext;
	struct in_addr *coa_ptr;
	int n, len, left = MAXMSG, mr_cr_auth_ext_len, i, j;

	/*------------------------------------------*
 	 | Send failure if HA didn't verify this CR	|
	 *------------------------------------------*/
	if (!cr->confirmed_by_ha) {
		/* TODO: Which code? */
		cr->reply_code = REGREP_HA_FAILED_AUTH_FA;
	}

	/*----------------------------------*
 	 | Fill the RegRep header			|
	 *----------------------------------*/
	DEBUG(DEBUG_MESSAGES, "Reply\n");
	n = sizeof(struct reg_rep);
	if (n > left) {
		DEBUG(DEBUG_MESSAGES, "RegRep buffer overflow!\n");
		return -1;
	}
	reply 					= (struct reg_rep *) send_buf;
	memset(reply, 0, sizeof(struct reg_rep));
	reply->type 			= REG_REP;
	reply->code 			= cr->reply_code;
	
	if (cr->cr_deregistered) 
		reply->lifetime = 0;
	else 
		reply->lifetime = htons(cr->reply_lifetime);

	
	reply->home_addr.s_addr = cr->cr_hoa.s_addr;
	reply->ha_addr.s_addr 	= config.ha_ip_addr.s_addr;
	
	/* Copy the ID for this Reply from the Request */
	reply->id[0] = cr->received_reg_id[0];
	reply->id[1] = cr->received_reg_id[1];
	
	pos 	+= n;
	left 	-= n;

	/*----------------------------------*
 	 | Print some info					|
	 *----------------------------------*/
	if (reply->code == REGREP_ACCEPTED) {
		print_trace(mn_traces, "Sending Registration Accept to CR CoA %s:%d\n",
			inet_ntoa(cr->tnl_dst), ntohs(cr->cr_co_port));
	} else if (reply->code == REGREP_PREACCEPTED) {
		print_trace(mn_traces, "Sending Registration Pre-Accept to CR CoA %s:%d\n",
			inet_ntoa(cr->tnl_dst), ntohs(cr->cr_co_port));
	} else {
		print_trace(mn_traces, "Sending Registration Failure to CR CoA %s:%d\n",
			inet_ntoa(cr->tnl_dst), ntohs(cr->cr_co_port));
	}
	
	/*----------------------------------*
 	 | UDP Tunneling Reply				|
	 *----------------------------------*/
	if (cr->udp_tnl_requested) 
	{
		DEBUG(DEBUG_MESSAGES, " * UDP TNL Reply\n");
		
		n = sizeof(struct udp_tnl_rep_ext);
		if (n > left) {
			DEBUG(DEBUG_MESSAGES, "RegRep buffer overflow!\n");
			return -1;
		}
		
		tnl_reply						= (struct udp_tnl_rep_ext *) pos;
		tnl_reply->type					= UDP_TNL_REP_EXT;	// == 44
		tnl_reply->length				= 6;
		tnl_reply->subtype				= 0;
		tnl_reply->reply_code			= 0;	// == accept tunneling
		tnl_reply->reserved				= 0;
		tnl_reply->keepalive_interval 	= 0;	// 0: MN must use default
		
		pos 	+= n;
		left 	-= n;	
	}
 	
 	/*----------------------------------*
 	 | Add NEMO Ack extension			|
	 *----------------------------------*/
 	for (j = 0; j < cr->cr_nw_count; j++) 
	{
		/* Skip the HoA */
		if (cr->cr_nws[j].prefix.s_addr == cr->cr_hoa.s_addr) continue;
			
		DEBUG(DEBUG_MESSAGES, " * NEMO Ack %s/%d", inet_ntoa(cr->cr_nws[j].prefix),
			cr->cr_nws[j].prefix_length);
			
		n = sizeof(struct mobile_nw_ack_ext);
		if (n > left) {
			DEBUG(DEBUG_MESSAGES, "RegRep buffer overflow!\n");
			return -1;
		}
				
		nemo_ack 			= (struct mobile_nw_ack_ext *) pos;
		nemo_ack->type		= 148;
		nemo_ack->length	= 8;
		nemo_ack->subtype	= 1;	// Explicit mode ACK
		
		if (cr->cr_nws[j].authorized) {
			nemo_ack->code = MOBNET_SUCCESS;
			DEBUG(DEBUG_MESSAGES, " SUCCESS\n");
		} else { 
			/* TODO: Also handle the other error codes (INVALID_PREFIX_LEN, 
			 * FORWARDING_SETUP_FAILED) */
			nemo_ack->code = MOBNET_UNAUTHORIZED;
			DEBUG(DEBUG_MESSAGES, " UNAUTHORIZED\n");
		}	
		
		nemo_ack->prefix_length	= cr->cr_nws[j].prefix_length;
		nemo_ack->reserved		= 0;
		nemo_ack->prefix		= cr->cr_nws[j].prefix.s_addr;
		
		pos 	+= n;
		left 	-= n;
	}
	
	/*----------------------------------*
 	 | Add CoA extension				|
	 *----------------------------------*/
	/* In this version we send only 1 CoA */
	DEBUG(DEBUG_MESSAGES, " * CoA Ext %s\n", inet_ntoa(mn.co_addr));
	//update_usable_nw_ifaces();
	
	/* Size of the extension including all MR CoA's (4 B each) */
	n = sizeof(struct coa_ext);				/* This includes the first CoA */
	n += (usable_nw_iface_count - 1) * 4;	/* Add the rest */
	
	if (n > left) {
		DEBUG(DEBUG_MESSAGES, "RegRep buffer overflow!\n");
		return -1;
	}
	
	coa_ext				= (struct coa_ext *) pos;
	coa_ext->type		= TBA_T2;
	/* length: subtype (1 B) + reserved (1 B) + CoAs (x * 4 B) */
	coa_ext->length		= 2 + 4 * usable_nw_iface_count; 
	coa_ext->subtype	= TBA_ST_2_3;
	coa_ext->reserved	= 0;
	coa_ptr				= (struct in_addr *) &coa_ext->coa;
	for (j = 0; j < usable_nw_iface_count; j++)
	{
		coa_ptr->s_addr	= usable_nw_ifaces[j].ip_addr.s_addr;
		coa_ptr++;
	}
	pos 	+= n;
	left 	-= n;
	
	/*----------------------------------*
 	 | Add Auth reply					|
	 *----------------------------------*/
	DEBUG(DEBUG_MESSAGES, " * MR CR Auth\n");
	mr_cr_auth_ext_len = add_mr_cr_auth_extension(cr, pos, send_buf);
	if (left < mr_cr_auth_ext_len) { 
		DEBUG(DEBUG_MESSAGES, "Not enough space to add the MR-CR Authentication ext!\n");
		return -1;
	}
	pos 	+= mr_cr_auth_ext_len;
	left 	-= mr_cr_auth_ext_len;

	/*----------------------------------------------------*
 	 | Solicit and initialize connection towards this CR. |
	 *----------------------------------------------------*/
	if (cr->connection_initialized) {
		DEBUG(DEBUG_MESSAGES, "Connection towards this CR already initialized\n");
	} else {
		DEBUG(DEBUG_MESSAGES, "Need to initialize connection to this CR\n"); 
		cr->send_init_msgs = 1;
	}
	
	/*------------------------------------------------------*
 	 | Send RegRep from MR HoA:434 to CR CoA:reg_port		|
	 *------------------------------------------------------*/
	cr->send_regrep = 0;
	len = (int) (pos - send_buf);
	assert(len <= MAXMSG);

	/* parse the message in order to get the details of the sent message 
	 * into debug output; in addition, this validates the format of the request */
	int parse_res = parse_msg(send_buf, len, &ext_for_parse_msg);
	if (parse_res != 0) {
		DEBUG(DEBUG_INFO, "WARNING! Own registration request was not "
		      "valid!\n");
	}
	
	send_from_mr_hoa_mip_port_to_ha_tunnel(send_buf, len, cr->tnl_dst, cr->cr_co_port);
		
	if (cr->cr_deregistered)
	{
		DEBUG(DEBUG_INFO, "Reply sent for a deregistration -> delete RO Entry of CR %s\n",
			inet_ntoa(cr->cr_hoa));
		delete_ro_entry(cr);
	} 
	
	return 0;
}
static void show_key_dump(const char *title, const unsigned char *key,
			  int keylen)
{
	int i;

	DEBUG(DEBUG_INFO, "%s: ", title);
	for (i = 0; i < keylen; i++)
		DEBUG(DEBUG_INFO, "%02x", key[i]);
	DEBUG(DEBUG_INFO, "\n");
}

/**
 * handle_reply_aaa:
 * @ext: parsed registration reply extensions
 *
 * Handle AAA related extensions: validate the reply and generate MN-FA and/or
 * MN-HA security association if key material was present.
 *
 * Returns: 1 = valid extensions, 0 = invalid extensions
 */
static int handle_reply_aaa(struct msg_extensions *ext, struct in_addr addr)
{
	unsigned char *nodeaddr;
	unsigned int nodeaddrlen;
	unsigned char fa_secret[MAXSHAREDSECRETLEN];
	int fa_secret_len;
	unsigned char ha_secret[MAXSHAREDSECRETLEN];
	int ha_secret_len;
	struct mn_fa_key_material_from_aaa *fa_key = NULL;
	struct mn_ha_key_material_from_aaa *ha_key = NULL;

	if (config.mn_home_ip_addr.s_addr == 0 &&
	    config.mn_nai_len > 0) {
		nodeaddr = config.mn_nai;
		nodeaddrlen = config.mn_nai_len;
	} else {
		nodeaddr = (unsigned char *) &config.mn_home_ip_addr;
		nodeaddrlen = 4;
	}

	if (!config.use_aaa) {
		DEBUG(DEBUG_INFO, "Registration reply has key material from "
		      "AAA, but AAA is not configured; ignoring extension\n");
		ext->mn_fa_key_material_aaa = NULL;
		ext->mn_ha_key_material_aaa = NULL;
		return 1;
	}

	if (ext->mn_fa_key_material_aaa) {
		fa_key = (struct mn_fa_key_material_from_aaa *)
			(ext->mn_fa_key_material_aaa + 1);

		if (ext->mf_auth == NULL ||
		    !auth_is_protected(ext->mn_fa_key_material_aaa,
				       ext->mf_auth)) {
			LOG2(LOG_ALERT, "MN-FA auth. ext. does not protect "
			     "MN-FA key material from AAA\n");
			return 0;
		}

		/* generate MN-FA key from key material */
		fa_secret_len = MAXSHAREDSECRETLEN;
		if (auth_generate_key(
			    config.mn_aaa_keygen_alg,
			    config.mn_aaa_shared_secret,
			    config.mn_aaa_shared_secret_len,
			    (unsigned char *) (fa_key + 1),
			    GET_GEN_MN_FA_KEY_REP_LEN(
				    ext->mn_fa_key_material_aaa) -
			    sizeof(struct mn_fa_key_material_from_aaa),
			    nodeaddr, nodeaddrlen,
			    fa_secret, &fa_secret_len)) {
			LOG2(LOG_ALERT, "MN-FA key generation with material "
			     "from AAA failed\n");
			return 0;
		}
		show_key_dump("MN-FA key", fa_secret, fa_secret_len);

		/* check that MN-FA auth ext is valid before accepting the
		 * key */
		if (!auth_check(auth_aaa_key_alg_id_to_dynamics(ntohs(
			fa_key->alg_id)),
				fa_secret, fa_secret_len,
				(unsigned char *) ext->rep, ext->mf_auth)) {
			LOG2(LOG_ALERT, "MN-FA key from AAA does not match "
			     "with MN-FA auth. ext.\n");
			return 0;
		}
	}

	if (ext->mn_ha_key_material_aaa) {
		ha_key = (struct mn_ha_key_material_from_aaa *)
			(ext->mn_ha_key_material_aaa + 1);

		if (ext->mh_auth == NULL ||
		    !auth_is_protected(ext->mn_ha_key_material_aaa,
				       ext->mh_auth)) {
			LOG2(LOG_ALERT, "MN-HA auth. ext. does not protect "
			     "MN-HA key material from AAA\n");
			return 0;
		}

		/* generate MN-HA key from key material */
		ha_secret_len = MAXSHAREDSECRETLEN;
		if (auth_generate_key(
			    config.mn_aaa_keygen_alg,
			    config.mn_aaa_shared_secret,
			    config.mn_aaa_shared_secret_len,
			    (unsigned char *) (ha_key + 1),
			    GET_GEN_MN_HA_KEY_REP_LEN(
				    ext->mn_ha_key_material_aaa) -
			    sizeof(struct mn_ha_key_material_from_aaa),
			    nodeaddr, nodeaddrlen,
			    ha_secret, &ha_secret_len)) {
			LOG2(LOG_ALERT, "MN-HA key generation with material "
			     "from AAA failed\n");
			return 0;
		}
		show_key_dump("MN-HA key", ha_secret, ha_secret_len);

		/* check that MN-HA auth ext is valid before accepting the
		 * key */
		if (!auth_check(auth_aaa_key_alg_id_to_dynamics(
			ntohs(ha_key->alg_id)),
				ha_secret, ha_secret_len,
				(unsigned char *) ext->rep, ext->mh_auth)) {
			LOG2(LOG_ALERT, "MN-HA key from AAA does not match "
			     "with MN-HA auth. ext.\n");
			return 0;
		}
	}

	/* Generated keys matched with authentication extensions - add new
	 * dynamic security associations */

	if (ext->mn_fa_key_material_aaa) {
		struct fa_spi_entry *spi;

		spi = malloc(sizeof(struct fa_spi_entry));
		if (spi == NULL) {
			DEBUG(DEBUG_INFO, "Could not allocate memory for new "
			      "FA security association\n");
			return 0;
		}
		memset(spi, 0, sizeof(struct fa_spi_entry));
		list_init_node(&spi->node);

		spi->spi = ntohl(fa_key->fa_spi);
		spi->addr.s_addr = addr.s_addr;
		spi->alg =
			auth_aaa_key_alg_id_to_dynamics(ntohs(fa_key->alg_id));
		time(&spi->created);
		spi->lifetime = ntohl(fa_key->lifetime);
		memcpy(spi->shared_secret, fa_secret, sizeof(fa_secret));
		spi->shared_secret_len = fa_secret_len;
		add_fa_spi(spi, 1);
	}

	if (ext->mn_ha_key_material_aaa) {
		config.spi = ntohl(ha_key->ha_spi);
		mn.use_auth_alg = config.auth_alg =
			auth_aaa_key_alg_id_to_dynamics(ntohs(ha_key->alg_id));
		config.replay_meth = auth_aaa_key_replay_to_dynamics(
			ntohs(ha_key->replay_method));
		memcpy(config.shared_secret, ha_secret, sizeof(ha_secret));
		config.shared_secret_len = ha_secret_len;
		time(&config.mn_ha_key_timestamp);
		config.mn_ha_key_lifetime =
			ntohl(ext->mn_ha_key_material_aaa->lifetime);
	}

	mn.aaa_rekey = 0;

	return 1;
}


/*
 * Check if ext contains valid registration reply.
 * correct reply should have id[1] same as in request.
 * Return nonzero if it's valid, zero otherwise.
 */
static int is_valid_reply(struct msg_extensions *ext, struct sockaddr_in addr)
{
	int auth_alg;
	struct fa_spi_entry *fa_spi;

	ASSERT(ext != NULL);

        /* verify fields */
        if (ext->rep == NULL || ext->rep->type != REG_REP) {
		LOG2(LOG_WARNING, "Received message from %s:%i was not a "
		     "registration reply\n",
		     inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));
		return 0;
	}

#ifndef MN_ENABLE_VALIDATION
	DEBUG(DEBUG_INFO, "MN - no message validation in use\n");
	return 1;
#else

	if ((ext->mn_fa_key_material_aaa != NULL ||
	     ext->mn_ha_key_material_aaa != NULL) &&
	    handle_reply_aaa(ext, addr.sin_addr) != 1) {
		LOG2(LOG_ALERT, "Registration reply validation failed for AAA "
			"related data - dropping reply\n");
		return 0;
	}

        /* verify mac */
	if (ext->mh_auth != NULL) {
		if (config.shared_secret_len < 0) {
			LOG2(LOG_ALERT, "MN-HA in reply, but security "
			     "association does not exists yet\n");
			return 0;
		}
		if (!auth_check(mn.use_auth_alg, config.shared_secret,
				config.shared_secret_len,
				(unsigned char *) ext->rep, ext->mh_auth)) {
			if (mn.use_auth_alg == AUTH_ALG_MD5)
				auth_alg = AUTH_ALG_MD5_RFC2002;
			else if (mn.use_auth_alg == AUTH_ALG_MD5_RFC2002)
				auth_alg = AUTH_ALG_MD5;
			else
				auth_alg = 0;
			if (auth_alg == 0 ||
			    !auth_check(auth_alg, config.shared_secret,
					config.shared_secret_len,
					(unsigned char *) ext->rep,
					ext->mh_auth)) {
				LOG2(LOG_ALERT,
				     "Incorrect HA->MN authentication "
				     "extension received in the reply\n");
				if (config.use_aaa)
					mn.aaa_rekey = 1;
				return 0;
			} else {
				/* non-standard authentication method used */
				DEBUG(DEBUG_MESSAGES,
				      "Non-standard MAC calculation detected "
				      "- trying to use the same method\n");
				mn.use_auth_alg = auth_alg;
				return 0;
			}
		}
	} else if (ext->sk_auth != NULL && mn.session_key_len > 0) {
		if (!auth_check_vendor(config.auth_alg, mn.session_key,
				       mn.session_key_len,
				       (unsigned char *) ext->rep,
				       ext->sk_auth)) {
			LOG2(LOG_ALERT, "Incorrect session key based "
			     "authentication extension in the reply.\n");
			return 0;
		}
		/* FAs cannot send session_keys */
		if (ext->mn_keyrep != NULL) {
			DEBUG(DEBUG_MESSAGES, "Registration reply from FA "
			      "contained mn_keyrep extension - dropped it\n");
			/* Just clear the session key to prevent FAs from
			 * sending them us. */
			ext->mn_keyrep = NULL;
		}
	} else if (ext->rep->code < 64 || ext->rep->code > 127) {
		/* codes 64 .. 127 are for Foreign Agent denial; only require
		 * authentication extension in messages that are from the
		 * Home Agent */
		LOG2(LOG_ALERT, "No authentication in the reply\n");
		if (config.use_aaa)
			mn.aaa_rekey = 1;
		return 0;
	}

	/* check the possible mf_auth */
	fa_spi = get_fa_spi(0, addr.sin_addr);
	if (ext->rep->code == REGREP_MN_FAILED_AUTH_FA && config.use_aaa &&
	    fa_spi != NULL && fa_spi->created != 0) {
		DEBUG(DEBUG_INFO, "FA reported incorrect authentication - "
		      "trying to create a new security association\n");
		remove_fa_spi(fa_spi);
		free(fa_spi);
		fa_spi = NULL;
		mn.try_to_fix_sec_assoc = 1;
	}

	if (fa_spi != NULL) {
		if (ext->mf_auth == NULL) {
			LOG2(LOG_ALERT, "No mf_auth in the reply even though "
			     "there is a security association with the FA.\n");
			return 0;
		} else if (!auth_check(fa_spi->alg, fa_spi->shared_secret,
				       fa_spi->shared_secret_len,
				       (unsigned char *) ext->rep,
				       ext->mf_auth)) {
			LOG2(LOG_ALERT, "Incorrect FA->MN authentication "
			     "extension received the reply\n");
			return 0;
		}
	} else if (ext->mf_auth != NULL && !mn.try_to_fix_sec_assoc) {
		LOG2(LOG_ALERT, "Reply has an MN-FA auth. ext., but there is "
		     "no security association with the FA.\n");
		return 0;
	}

        /* verify id */
        if (ntohl(ext->rep->id[1]) != mn.registration_id[1]) {
                LOG2(LOG_ALERT,
		     "Wrong id field in the reply (was %08x, "
		     "expected %08x)\n",
		     ext->rep->id[1], (u32) htonl(mn.registration_id[1]));
                return 0;
        }


	if (ntohs(ext->rep->lifetime) > 0 &&
	    IS_REGREP_ACCEPTED(ext->rep->code) &&
	    config.ha_ip_addr.s_addr == 0) 
	{
		DEBUG(DEBUG_INFO, "Discovered HA address from reply: %s\n",
		      inet_ntoa(ext->rep->ha_addr));
		config.ha_ip_addr.s_addr = ext->rep->ha_addr.s_addr;
	}

	if (ntohs(ext->rep->lifetime) > 0 &&
	    IS_REGREP_ACCEPTED(ext->rep->code) &&
	    config.mn_home_ip_addr.s_addr == 0) 
	{
		/*--------------------------------------------------------------*
		 | Discovered a HoA.											|
		 *--------------------------------------------------------------*/	
		print_trace(mn_traces, "Discovered MR home address from "
		      "reply: %s\n", inet_ntoa(ext->rep->home_addr));
		mn.local_addr.s_addr =
			config.mn_home_ip_addr.s_addr =
			ext->rep->home_addr.s_addr;
		
		/*--------------------------------------------------------------*
		 | Open the MR HoA sockets: 	.								|
		 *--------------------------------------------------------------*/
		dyn_ip_addr_add(mn.hoa_dev, config.mn_home_ip_addr);
		/* port 434: incoming traffic from CR */
		mn.hoa_listener_socket = create_ro_socket(config.mn_home_ip_addr); 
		if (mn.hoa_listener_socket < 0) {
			fprintf(stderr, "Error in creating HoA listener socket!\n");
		}
				
		/* random port: socket to send the HoTI through and receive the HoT in */
		mn.hoa_socket = socket(AF_INET, SOCK_DGRAM, 0);
		if (mn.hoa_socket < 0) {
			fprintf(stderr, "HoA:random_port - %s\n",
				inet_ntoa(config.mn_home_ip_addr), strerror(errno));
		}
		struct sockaddr_in local_addr; 
		local_addr.sin_addr.s_addr = config.mn_home_ip_addr.s_addr;
		if (bind(mn.hoa_socket, (struct sockaddr *) &local_addr, sizeof(local_addr)) < 0) {
			printf("bind HoA socket - %s", strerror(errno));
		}
		DEBUG(DEBUG_INFO, "is_valid_reply: HoA socket fd %d\n", mn.hoa_socket);
		
		if (mn.current_adv != NULL &&
		    dyn_ip_iface_address(mn.current_adv->ifindex,
					 &config.mn_home_ip_addr) != 1) {
			DEBUG(DEBUG_INFO, "Adding assigned home address to the"
			      " interface\n");
			if (dyn_ip_addr_add(mn.current_adv->ifname,
					    config.mn_home_ip_addr) == 0) {
				mn.dynamic_home_addr.s_addr =
					config.mn_home_ip_addr.s_addr;
				memcpy(mn.dynamic_home_addr_dev,
				       mn.current_adv->ifname, IFNAMSIZ);
			} else
				DEBUG(DEBUG_INFO, "dyn_ip_addr_add failed\n");
		}
	}


	/* don't accept any further messages with same id */
	mn.registration_id[1]++;

        return 1;
#endif
}


/* Handle session key extension. Decrypt and copy to session_key */
static void handle_key_ext(struct reg_rep *rep, struct msg_key *key)
{
    ASSERT(rep != NULL && key != NULL);
    DEBUG(DEBUG_MESSAGES, "mobile-home key extension received\n");
    /* If SPIs are different, something went wrong and we should
     * send another registration request to obtain session key. */

    if (ntohl(key->spi) != config.spi) {
            LOG2(LOG_WARNING, "Wrong SPI in the key reply extension.\n");
            return;
    }

    /* Extract new session key */
	if (mn.session_key) free(mn.session_key);
	mn.session_key = auth_decrypt(config.auth_alg, config.shared_secret,
				   config.shared_secret_len, key, rep,
				   (int *) &mn.session_key_len);
	if (mn.session_key == NULL) {
		LOG2(LOG_ERR,
		       "handle_key_ext: session key decryption failed\n");
		mn.session_key_len = 0;
		return;
	}

	show_key_dump("Session key", mn.session_key, mn.session_key_len);
}


/* Handle registration accepting registration reply */
static void handle_reg_accept(struct msg_extensions *ext)
{
#ifdef MN_LOCUPD_PROFILER
	char tmp[30];
#endif

	ASSERT(ext != NULL && ext->rep != NULL);

	/* Connection succeeded. It's ok to monitor agent adv. expiring */
	mn.expire_check = 1;

	if (mn.current_adv != NULL) {
		/* reset retry wait time for next registration */
		DEBUG(DEBUG_TIMERS, "FA[%s] reg_retry_time: "
		      "%i => %i (reg. reply ok)\n",
		      inet_ntoa(mn.current_adv->addr),
		      mn.current_adv->reg_retry_time,
		      MIN_REGISTRATION_TIME);
		mn.current_adv->reg_retry_time = MIN_REGISTRATION_TIME;
	} else
		mn.HA_reg_retry_time = MIN_REGISTRATION_TIME;

	if (ntohs(ext->rep->lifetime) > 0) {
		DEBUG(DEBUG_INFO, "Registration accepted\n");
#ifdef INCLUDE_IPAY
		if (mn.current_adv != NULL)
			ipay_new_fa(mn.current_adv);
#endif
	} else DEBUG(DEBUG_INFO, "Deregistration accepted\n");

    switch (mn.state) 
    {
        case MN_CLOSE_FOR_HOME:
			mn.ha_error_count = 0;
            if (ntohs(ext->rep->lifetime) == 0)
                    at_home();
            break;
    	case MN_FIND_AGENT:
   		case MN_PASSIVE_FIND:
			/* The answer is coming later than expected,
			 * so we just skip it. */
			DEBUG(DEBUG_MESSAGES, "handle_reg_accept - in find agent or "
			      "passive find state - skipping reg. reply\n");
			break;
    	case MN_REQUEST_TUNNEL:
  		case MN_CONNECTED:
			mn.ha_error_count = 0;
            if (ext->mn_keyrep != NULL)
                    handle_key_ext(ext->rep, ext->mn_keyrep);
			if (ext->rep->lifetime == 0) 
			{
				LOG2(LOG_INFO, "Got registration accepted reply with "
				     "zero lifetime - aborting registration and going "
				     "to passive find state\n");
				reply_waiting_api(API_FAILED, NULL, 0);
				passive_find();
				return;
			} else {
				/* Tunneling gets started here */
				connected((ext->mh_auth) ? CON_HA : CON_FA,
						ntohs(ext->rep->lifetime));
#ifdef MN_LOCUPD_PROFILER
				snprintf(tmp, sizeof(tmp), "CON %s",
			 			inet_ntoa(mn.fa_addr));
				write_profiler(tmp);
#endif
				reply_waiting_api(API_SUCCESS, NULL, 0);
			}
           	break;

			case MN_DISCONNECTED:
#ifdef MN_LOCUPD_PROFILER
				write_profiler("DIS");
#endif
				break;

			default:
				break;
        }
}


/* returns: 1 = try again, 0 = give up */
static int handle_id_mismatch_ha(struct msg_extensions *ext)
{
	/* If reply code is 133, id failed and we should resynchronize
         * clock/nonce and try again */
#ifdef MN_LOCUPD_PROFILER
	write_profiler("DENY_ID");
#endif
	switch (config.replay_meth) {
	case REPLAY_PROTECTION_NONE:
		LOG2(LOG_INFO, "ID mismatch from HA even with "
		     "disabled replay protection\n");
		reply_waiting_api(API_FAILED, &ext->rep->code, 1);
		disconnect();
		break;

	case REPLAY_PROTECTION_TIMESTAMP:
		mn.clock_correction = ntohl(ext->rep->id[0]) -
			UNIX_NTP_DIFF -	time(NULL);
		if (mn.ha_error_count++ < MAX_HA_ERRORS) {
			LOG2(LOG_INFO, "Clock synchronization %d seconds\n",
			     mn.clock_correction);
			print_trace(mn_traces, "Clock synchronization %d seconds\n",
			     mn.clock_correction);
			return 1;
		} else {
			LOG2(LOG_INFO,
			     "Clock resynchronization failed. "
			     "Disconnecting.\n");
			print_trace(mn_traces, "Clock resynchronization failed --> disconnect\n"); 
			reply_waiting_api(API_FAILED, &ext->rep->code, 1);
			disconnect();
		}
		break;

	case REPLAY_PROTECTION_NONCE:
		if (mn.ha_error_count++ < MAX_HA_ERRORS) {
			LOG2(LOG_INFO, "Nonce mismatch - syncronizing\n");
			return 1;
		} else {
			LOG2(LOG_INFO, "Nonce mismatch - "
			     "syncronization failed\n");
			reply_waiting_api(API_FAILED, &ext->rep->code, 1);
			disconnect();
		}
		break;
	}

	return 0;
}


/* returns: 1 = try again, 0 = give up */
static int handle_long_lifetime(struct msg_extensions *ext)
{
#ifdef MN_LOCUPD_PROFILER
	write_profiler("LIFETIME_FA");
#endif
	if (mn.req_lifetime > ntohs(ext->rep->lifetime) &&
		ntohs(ext->rep->lifetime) >= MIN_ALLOWED_LIFETIME) {
		LOG2(LOG_INFO, "FA (closest FA %s), "
		     "requests smaller lifetime (%d) - retrying\n",
		     inet_ntoa(mn.fa_addr), ntohs(ext->rep->lifetime));
		mn.req_lifetime = ntohs(ext->rep->lifetime);
		return 1;
	}

	/* this FA failed - try to find another one */
	LOG2(LOG_INFO, "FA %s denies requested lifetime %d s with bad "
	     "suggestion %d s - changing FA\n",
	     inet_ntoa(mn.fa_addr), mn.req_lifetime,
	     ntohs(ext->rep->lifetime));
	if (mn.current_adv != NULL)
		mn.current_adv->reg_failed = 1;
	else
		LOG2(LOG_INFO, "handle_long_lifetime - current_adv == NULL\n");
	find_agent(STATE_INIT);
	return 0;
}


/* returns: 1 = try again, 0 = give up */
static int handle_unknown_ha(struct msg_extensions *ext)
{
	int retry_hadisc = 0;

	/* dynamic home agent address resolution */

	if (config.ha_ip_addr.s_addr != 0 || config.home_net_addr_plen < 0) {
		LOG2(LOG_INFO, "HA did not accept HA address %s\n",
		     inet_ntoa(config.ha_ip_addr));
		retry_hadisc = 1;
	}

	if (ext->rep->ha_addr.s_addr ==
	    config.home_net_subnet_bc.s_addr) {
		LOG2(LOG_INFO, "HA returned same subnet-directed broadcast "
		     "address in dynamic address resolution\n");
		retry_hadisc = 1;
	}

	if (retry_hadisc) {
		if (config.use_hadisc) {
			DEBUG(DEBUG_MESSAGES, "Try to discover another HA\n");
			mn.info_str = "restarting HA discovery";
			config.ha_ip_addr.s_addr = 0;
			return 1;
		}
		return 0;
	}

	LOG2(LOG_INFO, "Learned HA address (%s) with dynamic address "
	     "resolution\n", inet_ntoa(ext->rep->ha_addr));
	mn.info_str = "HA discovery completed successfully";
	config.ha_ip_addr.s_addr = ext->rep->ha_addr.s_addr;
	return 1; /* try to register with the new HA address */
}


/**
 * handle_reg_denied:
 * ext: received and parsed registration reply
 *
 * Handle registration reply denying registration.
 *
 * Returns 1 if registration should be retried or 0 if registration should be
 * aborted.
 */
static int handle_reg_denied(struct msg_extensions *ext)
{
        ASSERT(ext != NULL && ext->rep != NULL);

	if (mn.state != MN_REQUEST_TUNNEL &&
	    mn.state != MN_CLOSE_FOR_HOME &&
	    mn.state != MN_CONNECTED) {
		DEBUG(DEBUG_MESSAGES, "handle_reg_denied - received denial "
		      "reply while not trying to register\n");
		return 0;
	}

	DEBUG(DEBUG_MESSAGES, "Registration denied, code = %d (%s)\n",
		     ext->rep->code, reply_code_str(ext->rep->code));
	
	print_trace(mn_traces, "Received registration denied, code %d (%s)\n",
		ext->rep->code, reply_code_str(ext->rep->code));
	
	switch (ext->rep->code) {
	case REGREP_ID_MISMATCH_HA:
		return handle_id_mismatch_ha(ext);

	case REGREP_LONG_LIFETIME_FA:
		return handle_long_lifetime(ext);

	case REGREP_UNKNOWN_HA_HA:
		return handle_unknown_ha(ext);

	case REGREP_HA_HOST_UNCREACHABLE_FA:
	case REGREP_HA_PORT_UNCREACHABLE_FA:
	case REGREP_HA_UNREACHABLE_FA:
		if (config.use_hadisc && config.ha_ip_addr.s_addr != 0) {
			DEBUG(DEBUG_MESSAGES, "HA unreachable => try to "
			      "discover another HA\n");
			mn.info_str = "restarting HA discovery";
			config.ha_ip_addr.s_addr = 0;
			return 1;
		}
		return 0;

	case REGREP_UNKNOWN_CHALLENGE_FA:
	case REGREP_STALE_CHALLENGE_FA:
		return 1; /* try to reregister; FIX: should get a new challenge
			   * value, if no new agent advertisement has been
			   * received */

	case REGREP_MN_FAILED_AUTH_FA:
		if (mn.try_to_fix_sec_assoc) {
			/* dynamic security association was removed in
			 * is_valid_reply() so try to reregister and create a
			 * new security association with the FA */
			return 1;
		}

		/* no method for fixing the security association, so just fall
		 * through to the default case */

        default:
#ifdef MN_LOCUPD_PROFILER
		write_profiler("DENY");
#endif
		LOG2(LOG_INFO, "Registration denied, code = %d (%s)\n",
		     ext->rep->code, reply_code_str(ext->rep->code));
		reply_waiting_api(API_FAILED, &ext->rep->code, 1);
		/* Try to change the FA by degrading the priority of the "
		   "current FA. */
		degrade_current_fa_priority();
        }

	return 0;
}

int handle_cr_keepalive_reply(struct ro_entry *cr, unsigned short icmp_seq) 
{
	return handle_keepalive_reply(cr->keepalives, icmp_seq, cr->RTTs, 
			&cr->ping_failure_count);
}
int handle_ha_keepalive_reply(unsigned short icmp_seq) 
{
	return handle_keepalive_reply(mn.ha_keepalives, icmp_seq, mn.RTTs,
			&mn.ping_failure_count); 
}

/* handle_ping_reply: Find out by the sequence number which ping request's
 * reply this is and make the corresponding conclusions about the state of 
 * the tunnel to HA or CR. 
 */
int handle_ping_reply(struct icmphdr *icmph) {
	
	struct timeval now, rtt;
	struct ro_entry *cr;
	int i, rtt_in_ms;
	unsigned short icmp_seq;
	
	gettimeofday(&now, NULL);
	icmp_seq = ntohs(icmph->un.echo.sequence);
		
	/*--------------------------------------------------------------*
	 | Check if this was a keepalive reply from the HA.				|
	 *--------------------------------------------------------------*/
	/* Function returns 0 if the packet was a HA keepalive reply */
	if (handle_ha_keepalive_reply(icmp_seq) == 0)
		return 0;

	/*------------------------------------------------------------------*
	 | Check if this was an init ping or a keepalive reply from a CR.	|
	 *------------------------------------------------------------------*/
	cr = (struct ro_entry *) get_ro_entry_by_ping_seq(icmp_seq);
	if (cr != NULL) {
		DEBUG(DEBUG_MESSAGES, "handle_ping_reply: from CR %s\n",
			inet_ntoa(cr->cr_hoa));
		if (cr->rr_state == RR_IN_PROGRESS ||
			cr->rr_state == RR_INACTIVE) 
		{
			/* Reply for init ping -> tunneling ok */
			DEBUG(DEBUG_MESSAGES, " - RR_IN_PROGRESS: init ping reply\n");
			return start_tunneling_to_cr(cr);
		} else if (cr->rr_state == RR_ACTIVE) {
			DEBUG(DEBUG_MESSAGES, " - RR_ACTIVE: keepalive reply\n");
			if (handle_cr_keepalive_reply(cr, icmp_seq) == 0)
				return 0;
		} else {
			DEBUG(DEBUG_MESSAGES, " - unknown RR State: %d\n", cr->rr_state);
		}
	}

	return 0;
}

/* handle_cr_ping_request: Handle a ping request probably received from a CR
 * testing a direct MR <==> CR tunnel. Send a reply to this request directly
 * to the tunnel without passing the message to the kernel.
 * 
 * ICMP Reply protocol stack
 *   ICMP	: type ICMP Echo Request (0)
 *   IP		: src MR HoA, dst CR HoA 
 *   MIP	: type TNL
 *   UDP	: dst 434
 *   IP		: src MR CoA, dst CR CoA
 * 
 * @src_addr: Should include the CR CoA and port (the lowest IP of the stack)
 * @ip_header: pointer to the IP header that was after the MIP TNL header
 */
int handle_cr_ping_request(struct ro_entry *cr, struct sockaddr_in *src_addr, 
		struct iphdr *ip_header, int receiver_socket, struct in_addr *receiver_ip) 
{
	unsigned char buffer[84];
	struct icmphdr *icmph;
	int icmp_data_len = 56;
	unsigned char icmp_data[icmp_data_len];
	unsigned short icmp_seq;
	unsigned char icmp_type = 0;	// Echo Reply
	struct in_addr tunneled_ip_src;	// Addresses of the IP inside the tunnel
	struct in_addr tunneled_ip_dst;
	struct in_addr tnl_dst_ip;		// Tunnel end address
	int tnl_dst_port;
	int icmp_packet_size;
	char cr_hoa_str[INET6_ADDRSTRLEN];
	char cr_coa_str[INET6_ADDRSTRLEN];
	char new_dev_name[IFNAMSIZ];
	DEBUG(DEBUG_INFO, "Initiation ping request received from CR %s\n",
		inet_ntoa(cr->cr_hoa)); 
	
	icmph = (struct icmphdr *) 
		((unsigned char *) ip_header + sizeof(struct iphdr));

	/*------------------------------------------------------------------*
	 | Switch Request's srcIP to be the Reply's dstIP etc.				|
	 *------------------------------------------------------------------*/

 	/* Make the dst address to be the src address */
	tnl_dst_ip.s_addr	= src_addr->sin_addr.s_addr;
	tnl_dst_port		= src_addr->sin_port;
	
	/* Use the same sequence */
	icmp_seq = ntohs(icmph->un.echo.sequence);
	
	tunneled_ip_src.s_addr = ip_header->daddr;
	tunneled_ip_dst.s_addr = ip_header->saddr;

	/*----------------------------------------------------------*
	 | Check if we are waiting for this CR to init the tunnel 	|
	 *----------------------------------------------------------*/
	if (cr->waiting_for_ping) 
	{
		cr->waiting_for_ping = 0;
		DEBUG(DEBUG_INFO, "Update CR's tunnel addresses to correspond to this ping's src and dst IP:\n");
		/* TODO: Check validity */
		/*--------------------------------------------------------------------------*
		 | Update the CR tunnel's src IP with the new MR CoA in case it has changed	|
		 *--------------------------------------------------------------------------*/
		if (receiver_ip != NULL)
		{
			if (cr->mr_coa.s_addr != receiver_ip->s_addr) 
			{
				/* Bind the CR connection to the new device */
				if (get_dev_name_by_ip(*receiver_ip, new_dev_name))
				{
					bind_cr_connection_to_device(cr, new_dev_name);
					
					DEBUG(DEBUG_INFO, "\tBinded CR tunnel to device %s, CoA %s\n",
						cr->device_in_use, inet_ntoa(cr->mr_coa));
				} else {
					DEBUG(DEBUG_INFO, "CR ping received into an bad IP address %s!\n", 
						inet_ntoa(*receiver_ip));
				}
			}
		}
		/*--------------------------------------------------------------------------*
		 | Replace the CR tunnel dst IP with the new MR CoA in case it has changed	|
		 *--------------------------------------------------------------------------*/
		if (cr->tnl_dst.s_addr != tnl_dst_ip.s_addr) 
		{
			set_cr_tnl_dst(cr, tnl_dst_ip);
			DEBUG(DEBUG_INFO, "\tUpdated tunnel dst CoA: %s\n",
				inet_ntoa(cr->tnl_dst));
		}
		DEBUG(DEBUG_INFO, "Try this connection by sending a ping through the tunnel\n");
		send_initial_ping_to_CR(cr);
	}	

	inet_ntop(AF_INET, &tunneled_ip_dst, cr_hoa_str, INET6_ADDRSTRLEN);
	inet_ntop(AF_INET, &tnl_dst_ip, cr_coa_str, INET6_ADDRSTRLEN);
	
	DEBUG(DEBUG_INFO, "send Ping Reply to CR (%s) CoA %s, seq %d\n",
		cr_hoa_str, cr_coa_str, icmp_seq); 

	strncpy(icmp_data, "This is a ping Reply to CR", icmp_data_len);
	
	icmp_packet_size = fill_icmp_req_in_ip(buffer, tunneled_ip_src, tunneled_ip_dst, 
				icmp_seq, icmp_type, icmp_data, icmp_data_len);
	
	return from_buffer_to_tunnel(buffer, sizeof(buffer), tnl_dst_ip, tnl_dst_port,
				receiver_socket);
}

/* handle_tnl_msg_mn: Handles a message received from the tunnel by 
 * removing the MIP TNL header and writing the remaining datagram into
 * the tun_fd corresponding to the src IP address. After that the datagram
 * will be handled by the kernel. 
 * 
 * If the packet is an initial keepalive request from a CR whose tunnel 
 * is not active (routing to its TUN device is not yet set correctly), 
 * this ping is replied directly back to the tunnel, without passing 
 * it to the kernel. 
 * 
 * The incoming packet at msg_start: [MIP tnl header | IP header | ...]
 * 
 * TODO: Check the old tunnel packet handling!
 */
int handle_tnl_msg_mn(char *msg_start, struct sockaddr_in *outer_src_addr, 
			int len, int receiver_socket, struct in_addr *receiver_ip) 
{	
	char *ip_header_start;
	struct udp_tnl_data *tnl_data;
	struct in_addr inner_ip_src;
	struct in_addr inner_ip_dst;
	char inner_ip_src_str[INET_ADDRSTRLEN];
	struct icmphdr *icmph;
	int i, tun_fd = mn.tun_fd;
	short int ip_length;
	struct iphdr *ip_header;
	struct ro_entry *tnl_src_cr = NULL;	// Defined by the CoA this msg came from
	int inner_protocol_type;
	int inner_msg_len;
	unsigned char *inner_msg_start;
	
	/* Point tnl_data to the MIP TNL Msg */
	tnl_data = (struct udp_tnl_data *) msg_start;
	
	DEBUG(DEBUG_INFO, " <<<<================ \n");
	DEBUG(DEBUG_MESSAGES, "\tMIP Tnl Msg: type %d, next_header %d, reserved %d\n",
		tnl_data->type, tnl_data->next_header, ntohs(tnl_data->reserved));

	/*--------------------------------------------------------------*
	 | Jump over the MIP Tunnel Data Message header.				|
	 *--------------------------------------------------------------*/
	ip_header_start = msg_start + sizeof(struct udp_tnl_data);
	inner_protocol_type = print_ip_stack(ip_header_start);

	/*--------------------------------------------------------------*
	 | The next header is IP.										|
	 *--------------------------------------------------------------*/
	ip_header 			= (struct iphdr *) ip_header_start;
	inner_ip_src.s_addr	= ip_header->saddr;  
	inner_ip_dst.s_addr = ip_header->daddr;
	inet_ntop(AF_INET, &inner_ip_src, inner_ip_src_str, INET_ADDRSTRLEN);
	
	/*--------------------------------------------------------------*
	 | Check if the packet in the tunnel is HoT.					|
	 *--------------------------------------------------------------*/
	if (inner_protocol_type == TBA_MIP3) 
	{
		inner_msg_start = ip_header_start + sizeof(struct iphdr) 
						+ sizeof(struct udphdr);
		return handle_hot(inner_msg_start, outer_src_addr);
	}
	
	/*------------------------------------------------------------------*
	 | Find out whose tunnel it came to get the corresponding tun_fd.	|
	 *------------------------------------------------------------------*/
	if (outer_src_addr->sin_addr.s_addr == config.ha_ip_addr.s_addr) 
	{
		/*--------------------------------------*
		 | Came from the HA tunnel.				|
		 *--------------------------------------*/
		DEBUG(DEBUG_MESSAGES, "\t - Tunnel message from HA\n");
		tun_fd = mn.tun_fd;
		mn.ping_failure_count = 0;
		set_ha_keepalive_timer_expiration();
	} else 	{
		/*--------------------------------------*
		 | See if it came from a CR's tunnel.	|
		 *--------------------------------------*/
		/* Try to identify CR by the tunnel src IP (CoA) */
		tnl_src_cr = (struct ro_entry *) get_ro_entry_by_tnl_dst(outer_src_addr->sin_addr);
		if (tnl_src_cr == NULL)	
		{
			/* Check if the packet is from an old tunnel */
			tnl_src_cr = (struct ro_entry *) get_ro_entry_by_old_tnl_dst(outer_src_addr->sin_addr);
			if (tnl_src_cr != NULL)
			{
				DEBUG(DEBUG_MESSAGES, "\t - Msg from CR's old tunnel which is still alive\n");
			}
		}
		if (tnl_src_cr != NULL)	
		{
			/*--------------------------------------*
			 | Came from a CR's tunnel.				|
			 *--------------------------------------*/
			DEBUG(DEBUG_MESSAGES, "\t - Tunnel message from CR %s\n",
				inet_ntoa(tnl_src_cr->cr_hoa)); 
			tun_fd = tnl_src_cr->tun_fd;
		} else {
			/*--------------------------------------*
			 | Came from an unknown tunnel.			|
			 *--------------------------------------*/
			fprintf(stderr, "Tunnel message from an unknown CR: src CoA %s!\n",
				inet_ntoa(outer_src_addr->sin_addr));
			return -1;
		}
	}
	
	/*--------------------------------------------------------------*
	 | Handle, if it's ICMP.										|
	 *--------------------------------------------------------------*/
	if (ip_header->protocol == 1) {
		icmph = (struct icmphdr *) (ip_header_start + sizeof(struct iphdr));
		if (icmph->type == 0) 
		{
			/* ICMP Echo Reply: update keepalive timers */
			handle_ping_reply(icmph);	
		} else if (icmph->type == 8) 
		{		
			/* ICMP Echo Request: 
			 * If it came from a CR that has an inactive tunnel
			 *  -> reply without passing on to the kernel */
			if (tnl_src_cr != NULL && 
				tnl_src_cr->rr_state != RR_ACTIVE) 
			{
				DEBUG(DEBUG_MESSAGES, "Came from a CR tunnel that is not active "
					"-> reply without passing on to kernel\n");
				/* Add TNL headers and send to the CR CoA */
		   		return handle_cr_ping_request(tnl_src_cr, outer_src_addr, 
		   				ip_header, receiver_socket, receiver_ip);
			}
		}
	}
	
	/*----------------------------------------------------------------------------------*
	 | Check if a CR tnl keepalive timer and ping failure count need to be initialized.	|
	 *----------------------------------------------------------------------------------*/
	/* Initialize if the message is from a CR and it's not a keepalive request */
	if (tnl_src_cr != NULL && icmph != NULL)
	{
		/* Won't init ping failure count and keepalive timer if msg is a 
		 * CR keepalive request:
		 *  - packet is ICMP Request
		 *  - inner ip src is CR's HoA 
		 *  - inner ip dst is my Hoa */
		if (icmph->type == 8 &&
			inner_ip_src.s_addr == tnl_src_cr->cr_hoa.s_addr &&
			inner_ip_dst.s_addr == config.mn_home_ip_addr.s_addr)
		{
			DEBUG(DEBUG_MESSAGES, " - Packet is CR keepalive request\n");
		} else {
			/* init ping failure count and keepalive timer */
			tnl_src_cr->ping_failure_count = 0;
			if (timerisset(&tnl_src_cr->keepalive_timer))
				set_cr_keepalive_timer_expiration(tnl_src_cr);
		}
	}
	
	/*--------------------------------------------------------------*
	 | Write all the data following the MIP header into the tun_fd.	|
	 *--------------------------------------------------------------*/
	DEBUG(DEBUG_MESSAGES, "\tWrite %d bytes into tun_fd %d\n", 
		ntohs(ip_header->tot_len), tun_fd);
	if(write(tun_fd, ip_header_start, ntohs(ip_header->tot_len)) < 0) {
    	perror("Writing into TUN fd");
    	return -1;
  	}	
 	
	return 0;
}
/* cp_adv_hoa_into_ro_entry: Helper function for handle_ro_prefix_adv.
 * Part of the RO Prefix Advertisement extension handling 
 */
int cp_adv_hoa_into_ro_entry(struct ro_entry *ro_entry, 
		struct ro_prefix_adv_data *data, int need_to_send_init, 
		int need_to_send_regrep) 
{
	__u8 info; 
	__u8 flag_mask = 0x3F; /* == 0011 1111 */
	
	if (ro_entry == NULL ||	data == NULL)
	{
		DEBUG(DEBUG_MESSAGES, "cp_hoa_into_ro_cache: a NULL pointer!\n");
		return -1;
	}
	
	/*----------------------------------------------------------*
	 | Data item's M flag is set --> contains Info + MR HoA		|
	 *----------------------------------------------------------*/
	info = data->flags_plen_info & flag_mask; /* Mask the M bit to 0 */
	ro_entry->confirmed_by_ha			= 1;
	ro_entry->not_reachable 			= info;
	ro_entry->cr_hoa.s_addr	 			= data->hoa_or_prefix;
	ro_entry->cr_nws[0].prefix.s_addr	= data->hoa_or_prefix;
	ro_entry->cr_nws[0].prefix_length 	= 32;
	ro_entry->cr_nws[0].nai[0]			= '\0';
	ro_entry->cr_nws[0].nai_length		= 1;
	
	/* Nw flags */
	ro_entry->cr_nws[0].plen_ok			= 1;
	ro_entry->cr_nws[0].forwarding_ok	= 0;
	ro_entry->cr_nws[0].authorized		= 1;
	
	ro_entry->cr_nw_count 				= 1;
	ro_entry->send_init_msgs			= need_to_send_init;
	ro_entry->send_regrep				= need_to_send_regrep;
	if (need_to_send_init) 
	{
		ro_entry->hot_received			= 0;
		ro_entry->cot_received			= 0;
	}
	DEBUG(DEBUG_MESSAGES, "cr_hoa and cr_nws[0] %s/%d, confirmed_by_ha %d, " 
		"not_reachable %d, nai %s, cr_nw_count set to %d\n", 
		inet_ntoa(ro_entry->cr_hoa), ro_entry->cr_nws[0].prefix_length, 
		ro_entry->confirmed_by_ha, ro_entry->not_reachable, 
		ro_entry->cr_nws[0].nai, ro_entry->cr_nw_count);
	
	return 0;
}
/* cp_adv_non_delta_prefix_into_ro_entry: Helper function for 
 * handle_ro_prefix_adv. Part of the RO Prefix Advertisement extension 
 * handling.
 * 
 * Returns the data item length or -1 if an error occurred.
 * 
 * This version doesn't support deltas nor realms so the function returns 
 * always round_up(plen / 4) + 2 as the data item length.
 */
int cp_adv_non_delta_prefix_into_nw(struct network *nw, struct ro_prefix_adv_data *data) 
{
	__u8 plen = data->flags_plen_info; /* No flags set */
	int octets_in_prefix;
	
	/*--------------------------------------------------------------*
	 | Data item's M flag is set --> contains Plen + Prefix + Realm	|
	 *--------------------------------------------------------------*/
	if (nw == NULL || data == NULL)
	{
		DEBUG(DEBUG_MESSAGES, "cp_non_delta_prefix_into_ro_cache: a NULL pointer!\n");
		return -1;
	}
	if (plen > 33 || plen < 1) {
		DEBUG(DEBUG_MESSAGES, "cp_non_delta_prefix_into_ro_cache: Invalid plen %d!\n",
			plen);
		return -1;
	}
	octets_in_prefix = divide_and_round_up(plen, 8);
	
	DEBUG(DEBUG_MESSAGES, "plen %d --> %d octets in the prefix\n",
		plen, octets_in_prefix);
	
	/* Set prefix to 0 and then copy the first octets from the datagram. */
	nw->prefix.s_addr = 0;
	memcpy(&nw->prefix, &data->hoa_or_prefix, octets_in_prefix);
	nw->prefix_length 	= plen;
	nw->nai[0]		 	= '\0';
	nw->nai_length	 	= 1;
	nw->plen_ok			= check_if_plen_ok(plen);
	nw->forwarding_ok	= 0;	// No routing changes at this point
	nw->authorized		= 1;	// Adv from HA --> authorized
	DEBUG(DEBUG_MESSAGES, "\tnw: %s/%d, nai %s\n",
		inet_ntoa(nw->prefix), nw->prefix_length, nw->nai);
	
	print_trace(mn_traces, " - nw: %s/%d, nai %s\n",
		inet_ntoa(nw->prefix), nw->prefix_length, nw->nai);
	
	return octets_in_prefix + 2;
}
/* handle_ro_prefix_adv: Handle the RO Prefix Advertisement
 * extension. Copy the CRs advertised by the HA into the RO Cache.
 * 
 */
int handle_ro_prefix_adv(struct ro_prefix_adv_head *adv_head) 
{
	struct ro_entry *cr			= NULL;
	struct network cr_nw;
	int length		 			= ntohs(adv_head->length);
	int total_length			= length + 4; // 4 B: Type + Subtype + Length
	int header_length			= sizeof(struct ro_prefix_adv_head);
	unsigned char *adv_start 	= (unsigned char *) adv_head;
	int offset 					= 0;
	int data_length, ro_entry_index, new_cr_nw_index;
	__u8 plen, info;
	__u8 flag_mask 				= 0x3F; /* == 0011 1111 */
	struct ro_prefix_adv_data *data;
	int i, need_to_send_init, need_to_send_regrep;
	DEBUG(DEBUG_MESSAGES, "handle_ro_prefix_adv: ");
	 
	if (total_length < header_length) {
		DEBUG(DEBUG_MESSAGES, "total length %d < header length %d !!!\n", 
			total_length, header_length);
		return -1;
	}
	
	DEBUG(DEBUG_MESSAGES, "total packet length %d\n", total_length);
	
	
	for (i = 0; i < total_length; i++) {
		DEBUG(DEBUG_MESSAGES, "%02x ", *(adv_start + i));
	}
	DEBUG(DEBUG_MESSAGES, "\n");
	
	/* Jump over the header to the first data item */
	offset 	+= header_length;
	
	while (offset < total_length)
	{
		data = (struct ro_prefix_adv_data *) (adv_start + offset);
		
		/* The D flag should not be set */
		if (data->flags_plen_info & RO_ADV_D) { 
			fprintf(stderr, "RO Adv.: D=1 but this version does not support "
				"Deltas in Route Optimization Prefix Advertisement!\n");
			return -1; 
		}
		if (data->flags_plen_info & RO_ADV_M) 
		{
			/*------------------------------------------*
			 | M flag set --> Info + MR HoA 			|
			 *------------------------------------------*/
			print_trace(mn_traces, " * advertised CR HoA %s\n", 
				inet_ntoa(*(struct in_addr *) &data->hoa_or_prefix));
			
			/*----------------------------------------------------------------*
			 | Check if an entry with this HoA already exists in the RO Cache |
			 *----------------------------------------------------------------*/
			/* A new MR -> update the pointer */
			cr = (struct ro_entry *) get_ro_entry_by_hoa(
					*(struct in_addr *) &data->hoa_or_prefix);
			if (cr == NULL) 
			{
				/*------------------------------------------------------*
				 | Unknown CR											|
				 *------------------------------------------------------*/
				/* Send HoTI + CoTI to this CR */
				DEBUG(DEBUG_MESSAGES, "Advertised CR not found from the RO Cache"
					"--> create an entry for it and send HoTI + CoTI\n");
				cr = (struct ro_entry *) get_new_ro_entry();
				need_to_send_init 	= 1;
				need_to_send_regrep	= 0;
				print_trace(mn_traces, "HA advertised an unknown RO capable CR %s\n", 
					inet_ntoa(*(struct in_addr *) &data->hoa_or_prefix));
				cp_adv_hoa_into_ro_entry(cr, data, need_to_send_init, need_to_send_regrep);
			} else 
			{
				/*--------------------------------------------------------------*
				 | Known CR														|
				 *--------------------------------------------------------------*/
				DEBUG(DEBUG_MESSAGES, "Advertised CR found from the RO Cache\n");
				if (cr->confirmed_by_ha == 0)
				{
					/*--------------------------------------------------------------*
					 | Solicited CR -> trustable									|
					 *--------------------------------------------------------------*/
					DEBUG(DEBUG_MESSAGES, " - this was a solicitation reply\n");		
					need_to_send_regrep	= 1;	// Send Registration Reply to the CR
								
					if (!cr->connection_initialized) {
						need_to_send_init = 0;	// Don't init before the reply sent
					}
					cp_adv_hoa_into_ro_entry(cr, data, need_to_send_init, need_to_send_regrep);
				} else {		
					/*--------------------------------------------------------------*
					 | Advertised a trusted CR -> Do nothing						|
					 *--------------------------------------------------------------*/
					DEBUG(DEBUG_MESSAGES, " - advertised an already trusted CR -> no changes in the RO Cache\n");
				}
			}	
			offset	+= sizeof(struct ro_prefix_adv_data);
			continue;
		} else { 
			/*------------------------------------------*
			 | M flag not set --> Plen + Prefix + Realm	|
			 *------------------------------------------*/
			if (cr == NULL) {
				DEBUG(DEBUG_MESSAGES, "Invalid RO Prefix Adv (Prefix item before a HoA item)!\n");
				return -1;
			}
			
			/*----------------------------------------------------------*
			 | Add network into the RO Entry if it's not there already	|
			 *----------------------------------------------------------*/
			data_length = cp_adv_non_delta_prefix_into_nw(&cr_nw, data);
				
			if (data_length < 0) {
				fprintf(stderr, "RO Adv.: Error copying prefix!\n");
				return -1;
			}
			add_confirmed_nw_item_into_ro_entry(cr, &cr_nw);
			offset += data_length;
			continue;
		}
	}
	return 0;
}

/* Handle the HoTI from an unknown CR and send a HoT to the CR's HoA.
 */
int handle_hoti(unsigned char *msg, struct sockaddr_in *src_addr) {
	struct ro_hoti *hoti;
	struct ro_hot	hot;	
	int temp_hot_socket;
	int hot_len, hot_sent;
	__u64 nonce 			= get_nonce_with_index(mn.current_nonce_index);
	__u16 nonce_index 		= mn.current_nonce_index;
	__u64 keygen_token;

	/* Calculate the Home Keygen Token */
	get_home_keygen_token(mn.node_key, &src_addr->sin_addr, &nonce, &keygen_token);
	
	hot_len = sizeof(struct ro_hot);
	
	hoti 				= (struct ro_hoti *) msg;
	hot.type			= TBA_MIP3;
	hot.reserved		= 0;
	hot.nonce_index		= htons(nonce_index);	// 16 bit
	hot.cookie			= hoti->cookie;			// 64 bit
	hot.keygen_token	= hton64(keygen_token);	// 64 bit

	print_trace(mn_traces, "HoTI received from CR HoA %s:%d - sending HoT\n", 
		inet_ntoa(src_addr->sin_addr), ntohs(src_addr->sin_port));

	DEBUG(DEBUG_MESSAGES, "Received HoTI with cookie %llu\n", ntoh64(hot.cookie));
	DEBUG(DEBUG_MESSAGES, "Calculated Home token with node_key %llu %llu and nonce %llu\n", 
		mn.node_key[0], mn.node_key[1], nonce);
	DEBUG(DEBUG_MESSAGES, "Sending HoT from MR HoA to CR HoA, token %llu, nonce index %d\n", 
		ntoh64(hot.keygen_token), nonce_index);
	
	/* Send the HoT message to the MR's HoA */
	hot_sent = sendto(mn.hoa_listener_socket, &hot, hot_len, 0, 
			(struct sockaddr *) src_addr, sizeof(struct sockaddr_in));

 	if (hot_sent != hot_len) {
 		LOG2(LOG_ERR, "sending HoT - sendto: %s\n", strerror(errno));
        return -1;	
 	}
 	
	DEBUG(DEBUG_MESSAGES, "\tHoT sent (%d bytes)\n", hot_sent);	
	return 0;	
}
/* Handle the CoTI from CR CoA and send a CoT to the from MR HoA to CR CoA.
 */
int handle_coti(unsigned char *msg, struct sockaddr_in *src_addr) {
	struct ro_coti *coti;
	struct ro_cot cot;
	int temp_cot_socket;
	int cot_len, cot_sent;
	__u64 nonce 		= get_nonce_with_index(mn.current_nonce_index);
	__u16 nonce_index 	= mn.current_nonce_index;
	__u64 keygen_token; 

	get_co_keygen_token(mn.node_key, &src_addr->sin_addr, &nonce, &keygen_token);
	
	cot_len = sizeof(struct ro_cot);
	
	coti 				= (struct ro_coti *) msg;
	cot.type			= TBA_MIP4;
	cot.reserved		= 0;
	cot.nonce_index		= htons(nonce_index);	// 16 bit
	cot.cookie			= coti->cookie;			// 64 bit
	cot.keygen_token	= hton64(keygen_token);	// 64 bit
	
	print_trace(mn_traces, "CoTI received from CR CoA %s:%d - sending CoT\n",
		inet_ntoa(src_addr->sin_addr), ntohs(src_addr->sin_port));
		
	DEBUG(DEBUG_MESSAGES, "Received CoTI with cookie %llu\n", ntoh64(cot.cookie));
	DEBUG(DEBUG_MESSAGES, "Calculated Care-of token with node_key %llu %llu and nonce %llu\n", 
		mn.node_key[0], mn.node_key[1], nonce);
	DEBUG(DEBUG_MESSAGES, "Sending CoT from MR HoA to CR CoA, token %llu, nonce index %d\n", 
		ntoh64(cot.keygen_token), nonce_index);;
	
	/* Send the CoT message to the MR's CoA */
	cot_sent = sendto(mn.hoa_listener_socket, &cot, cot_len, 0, 
		(struct sockaddr *) src_addr, sizeof(struct sockaddr_in));
 	
 	if (cot_sent != cot_len) {
 		LOG2(LOG_ERR, "sending CoT - sendto: %s\n", strerror(errno));
        return -1;	
 	}
	DEBUG(DEBUG_MESSAGES, "\tCoT sent (%d bytes)\n", cot_sent);
		
	return 0;	
}

/* handle_hot: Handle the HoT from the CR. A registration request is sent from my 
 * MR CoA to this CR's HoA.
 */
int handle_hot(unsigned char *msg, struct sockaddr_in *src_addr) {
	struct ro_hot *hot;
	struct ro_entry *cr;
	
	hot = (struct ro_hot *)	msg;
	
	/* Get the RO Cache Entry corresponding to this cookie */
	cr = (struct ro_entry *) get_ro_entry_by_cookie(&hot->cookie);
	if (cr == NULL) {
		DEBUG(DEBUG_MESSAGES, "HoT received but no HoTI sent with this cookie (%llu)!\n",
			ntoh64(hot->cookie));
		print_trace(mn_traces, "HoT received with an invalid cookie!\n");
		return -1;
	}
	
	/* Store nonce index and keygen token */
	cr->hot_nonce_index 	= ntohs(hot->nonce_index);
	cr->hot_keygen_token 	= ntoh64(hot->keygen_token);
	cr->hot_received		= 1;
	
	print_trace(mn_traces, "HoT received from CR %s.\n", inet_ntoa(cr->cr_hoa));
	DEBUG(DEBUG_MESSAGES, "Home Test (HoT) message received: cookie %llu, token %llu, nonce index %d.\n", 
		ntoh64(hot->cookie), cr->hot_keygen_token, cr->hot_nonce_index);
	
	/* If no CoT received from this CR --> wait for it 
	 * else                            --> register */
	if (!cr->cot_received) {
		DEBUG(DEBUG_MESSAGES, "No CoT corresponding to this cookie received yet --> wait for that.\n");
	} else {
		DEBUG(DEBUG_MESSAGES, "A CoT corresponding to this cookie already received --> register to the CR.\n");
		
		/* Check if if the registration socket is binded to a different iface than CoTI was sent through */
		if (cr->different_coti_iface)
		{
			bind_cr_reg_connection_to_device(cr, cr->last_coti_iface);
		}
		
		calculate_krm_for_ro_entry(cr);
		cr->connection_initialized 	= 1;
		cr->init_send_time.tv_sec 	= 0;
		cr->init_send_time.tv_usec 	= 0;
		
		return send_regreq_to_cr(cr, REG_CONNECT);
	}
	return 0;
}

/* handle_cot: Handle the CoT from the CR. A registration request is sent from my 
 * MR CoA to this CR's HoA.
 */
int handle_cot(unsigned char *msg, struct sockaddr_in *src_addr) {
	struct ro_cot *cot;
	struct ro_entry *cr;
	
	cot = (struct ro_cot *)	msg;
	
	/* Get the RO Cache Entry corresponding to this cookie */
	cr = (struct ro_entry *) get_ro_entry_by_cookie(&cot->cookie);
	if (cr == NULL) {
		DEBUG(DEBUG_MESSAGES, "CoT received but no CoTI sent with this cookie (%llu)!\n",
			cot->cookie);
		print_trace(mn_traces, "CoT received with an invalid cookie!\n");
		return -1;
	}
	
	/* Store nonce index and keygen token */
	cr->cot_nonce_index 	= ntohs(cot->nonce_index);
	cr->cot_keygen_token 	= ntoh64(cot->keygen_token);
	cr->cot_received		= 1;
	
	print_trace(mn_traces, "CoT received from CR %s.\n", inet_ntoa(cr->cr_hoa));
	DEBUG(DEBUG_MESSAGES, "CoT received: cookie %llu, token %llu, nonce index %d.\n", 
		ntoh64(cot->cookie), cr->cot_keygen_token, cr->cot_nonce_index);
	
	/* If no HoT received from this CR --> wait for it 
	 * else                            --> register */
	if (!cr->hot_received) {
		DEBUG(DEBUG_MESSAGES, "No HoT corresponding to this cookie received yet --> wait for that.\n");
	} else {
		DEBUG(DEBUG_MESSAGES, "A HoT corresponding to this cookie already received --> register to the CR.\n");		
		/* Check if if the registration socket is binded to a different iface than CoTI was sent through */
		if (cr->different_coti_iface)
		{
			bind_cr_reg_connection_to_device(cr, cr->last_coti_iface);
		}
		calculate_krm_for_ro_entry(cr);
		cr->connection_initialized = 1;
		cr->init_send_time.tv_sec 	= 0;
		cr->init_send_time.tv_usec 	= 0;
		return send_regreq_to_cr(cr, REG_CONNECT);
	}
		
	return 0;
}

int handle_coa_ext(struct ro_entry *cr, struct coa_ext *coa_ext) {
	int coa_amount = (coa_ext->length - 2) / 4;
	int i;
	char *coa_ptr = (char *) &coa_ext->coa;
	
	DEBUG(DEBUG_MESSAGES, "handle_coa_ext: %d CoA(s) in the extension\n", coa_amount);
	
	assert(cr != NULL && coa_ext != NULL);
	
	print_trace(mn_traces, "CR's Route Optimization CoAs:\n");
	
	for (i = 0; i < coa_amount; i++)
	{
		print_trace(mn_traces, " - %s\n", 
			inet_ntoa(*(struct in_addr *) coa_ptr));
			
		DEBUG(DEBUG_MESSAGES, "\tCoA %d: %s\n", 
			i, inet_ntoa(*(struct in_addr *) coa_ptr));
		
		/* Insert CoAs in the RO Entry */
		add_cr_coa_into_ro_entry(cr, *(struct in_addr *) coa_ptr, RR_INACTIVE);
		coa_ptr += 4; /* Jump to the next 4 byte address */
	}
	return 0;
}
/* handle_cr_regreq: From CR CoA port_x to MR HoA port 434. 
 * 
 * If the registering CR has not been solicited by HA, it needs to be
 * solicited before replying.
 * 
 * If Reg is ok and
 *   If MR HoA > CR HoA
 *     - the registration is ACCEPTED
 *   If MR HoA < CR HoA
 *     If tunnel is ACTIVE
 *       - the registration is ACCEPTED
 *     Else 
 *       - the registration is PRE_ACCEPTED
 *       - Now Register to CR and put up a tunnel towards its CoA:434
 */
int handle_cr_regreq(unsigned char *msg, struct sockaddr_in *src_addr, int bytes_received) {
	struct ro_entry *cr;
	struct in_addr *cr_hoa;
	struct in_addr *nw_ptr_list[20];
	int nw_list_count;
	int parse_res, i;
	char hoa_str[INET6_ADDRSTRLEN];
	char coa_str[INET6_ADDRSTRLEN];
	char nai_str[MAX_NAI_LEN + 1];
	int known_cr = 0, coa_update = 0;
	struct msg_extensions ext;					/* Msg will be parsed here */
	
	/* CR HoA is found from the MIP Reg message */
	cr_hoa = (struct in_addr *) (msg + 4);
	
	DEBUG(DEBUG_MESSAGES, "Registration request received from CR (%s) ",
		inet_ntoa(*cr_hoa));
	DEBUG(DEBUG_MESSAGES, "CoA %s:%d. Checking the RO Cache...\n",
		inet_ntoa(src_addr->sin_addr), ntohs(src_addr->sin_port));

	/*--------------------------------------------------------------*
	 | Get the CR with this HoA from the RO Cache					|
	 *--------------------------------------------------------------*/
	cr = (struct ro_entry *) get_ro_entry_by_hoa(*cr_hoa);
	if (cr == NULL)
	{
		/*----------------------------------------------------------*
		 | An unknown CR --> create a new RO Cache entry			|
		 *----------------------------------------------------------*/
		DEBUG(DEBUG_MESSAGES, "No binding found for this CR HoA --> "
			"insert in the RO Cache and solicit from HA\n");
		
		/* Create a new ro_entry with flag confirmed_by_ha = 0 (gets solicited) */
		cr = (struct ro_entry *) get_new_ro_entry();
		cr->cr_hoa.s_addr	= cr_hoa->s_addr; 
		cr->confirmed_by_ha	= 0;		/* Need to solicit */
		cr->send_init_msgs 	= 0; 		/* No need to send HoTI and CoTI */
		cr->send_regrep		= 0;		/* No RegRep to the CR before soliced from HA */
		
		/* Save CR's CoA and registration port to be the tunnel's destination */
		add_cr_coa_into_ro_entry(cr, src_addr->sin_addr);
		set_cr_tnl_dst(cr, src_addr->sin_addr);
	} else { 
		/*----------------------------------------------------------*
		 | A known CR 												|
		 *----------------------------------------------------------*/
		DEBUG(DEBUG_MESSAGES, "Found from RO Cache: current tnl dst CoA: %s\n",
			inet_ntoa(cr->tnl_dst));
		known_cr = 1;
		if (cr->tnl_dst.s_addr != src_addr->sin_addr.s_addr) {
			/*----------------------------------------------------------*
			 | CR changed CoA --> update routing if tunnel up			|
			 *----------------------------------------------------------*/
			DEBUG(DEBUG_MESSAGES, "CR updating its CoA to %s\n", inet_ntoa(src_addr->sin_addr));
			coa_update = 1;
			/* Save CR's CoA and registration port to be the tunnel destination */
			add_cr_coa_into_ro_entry(cr, src_addr->sin_addr);
			/* Set the new tnl dst */
			DEBUG(DEBUG_MESSAGES, " - update tunnel destination\n");
			set_cr_tnl_dst(cr, src_addr->sin_addr);
			cr->ping_failure_count = 0;
		}
	}
	cr->cr_co_port	= src_addr->sin_port;
	
	/*--------------------------------------------------------------*
	 | Parse the Request and its extensions							|
	 *--------------------------------------------------------------*/
	parse_res = parse_msg(msg, bytes_received, &ext);
	if (parse_res == -3 || parse_res == -4) {
		/*--------------------------------------------------------------*
		 | Bad extension -> Reply with a Registration Failure			|
		 *--------------------------------------------------------------*/
		DEBUG(DEBUG_MESSAGES, "Unknown critical vendor extension in a "
		     "request from %s:%i\n", inet_ntoa(src_addr->sin_addr),
		     ntohs(src_addr->sin_port));
		cr->reply_code = parse_res == -3 ? 
			REGREP_UNSUPP_VENDOR_ID_MN_HA : REGREP_UNSUPP_VENDOR_ID_FA_HA;
		return send_regrep_to_cr(cr);
		
	} else if ((parse_res != 0 || ext.req == NULL)) {
		/*--------------------------------------------------------------*
		 | Bad message -> Discard and don't reply						|
		 *--------------------------------------------------------------*/
		char *reason = "N/A";
		if (parse_res == -1) {
			reason = "unknown extension";
		} else if (parse_res == -2) {
			reason = "malformed message";
		} else if (ext.req == NULL) {
			reason = "not a request";
		}
		DEBUG(DEBUG_MESSAGES, "Message parser failed or not a request "
		     "(from %s:%i): %s\n", inet_ntoa(src_addr->sin_addr),
		     ntohs(src_addr->sin_port), reason);
		/* Print trace */
		print_trace(mn_traces, 
			"Registration message parsing failed: %s\n", reason);
		return -1;
	}
	
	if (ext.req != NULL) 
	{
		/* Save the Req ID for the Reply */
		cr->received_reg_id[0] 	= ext.req->id[0];
		cr->received_reg_id[1] 	= ext.req->id[1];
		cr->reply_lifetime		= ntohs(ext.req->lifetime);
		/* Get HoA and CoA strings */
		inet_ntop(AF_INET, &ext.req->home_addr, hoa_str, INET6_ADDRSTRLEN);	
		inet_ntop(AF_INET, &ext.req->co_addr, coa_str, INET6_ADDRSTRLEN);
		
		if (ext.mn_nai != NULL) 
		{
			get_nai_str(ext.mn_nai, nai_str);
			print_trace(mn_traces, "CR %s requesting registration with "
				"HoA %s, CoA %s, id %08x %08x\n", nai_str, hoa_str, coa_str, 
				(u32) ntohl(ext.req->id[0]), (u32) ntohl(ext.req->id[1]));
		} else 	{
			print_trace(mn_traces, "CR requesting registration with "
			"HoA %s, CoA %s, id %08x %08x\n", hoa_str, coa_str, 
			(u32) ntohl(ext.req->id[0]), (u32) ntohl(ext.req->id[1]));
		}
		if (coa_update) print_trace(mn_traces, " - CoA update\n");
	} else {	
		DEBUG(DEBUG_MESSAGES, "Error: ext.req == NULL!\n");
		return -1;
	}

	/*--------------------------------------------------------------*
	 | Check the authentication										|
	 *--------------------------------------------------------------*/
	if (config.disable_rr) {
		print_trace(mn_traces, "RR turned off -> no auth check\n");
	} else {	
		if (ext.mr_cr_auth == NULL) {
			print_trace(mn_traces, "No MR-CR Auth Extension found from the Registration Request!\n");
			cr->reply_code = REGREP_MN_FAILED_AUTH_FA;
			return send_regrep_to_cr(cr);
		} else {
			if (check_mr_cr_auth(&ext, cr->cr_hoa, cr->tnl_dst, ext.mr_cr_auth) == 0) {
				DEBUG(DEBUG_MESSAGES, "MR-CR Auth Extension ok\n");
			} else {
				print_trace(mn_traces, "Invalid MR-CR Auth Extension!\n");
				cr->reply_code = REGREP_MN_FAILED_AUTH_FA;
				return send_regrep_to_cr(cr);
			}
		}
	}
	/* UDP TNL Req */
	if (ext.tnl_req != NULL)
	{
		cr->udp_tnl_requested = 1;	/* This will be noticed when sending reply */
		/* TODO: Handle the whole protocol */
	}

	/*--------------------------------------------------------------*
	 | Check if the message is a deregistration						|
	 *--------------------------------------------------------------*/
	if (ntohs(ext.req->lifetime) == 0) {
		print_trace(mn_traces, "CR %s deregistering\n", hoa_str);
		
		DEBUG(DEBUG_MESSAGES, "Request is a Deregistration -> "
			"remove from RO Cache and send accept\n");
		cr->reply_code 		= REGREP_ACCEPTED;
		cr->cr_deregistered = 1;
		return send_regrep_to_cr(cr);
	}
	
	/*--------------------------------------------------------------*
	 | Update CR's network list										|
	 *--------------------------------------------------------------*/
	/* Copy NEMO networks into the RO Entry unless it is in there already */
	for (i = 0; i < ext.nemo_nw_amount; i++) {
		if (ext.nemo_req[i] != NULL) {
			/* Add the NEMO network for this CR */
			add_unconfirmed_nw_into_ro_entry(cr, (struct in_addr *) &ext.nemo_req[i]->prefix, 
				ext.nemo_req[i]->prefix_length);
		} else {
			DEBUG(DEBUG_MESSAGES, "NEMO nw number %d not parsed correctly!!\n", i);
		}
	}
	
	/*--------------------------------------------------------------*
	 | Set the Reply Code to be ACCEPT or PRE_ACCEPT				|
	 *--------------------------------------------------------------*/
	/* If RegRep received from a CR whose tunnel is not ACTIVE and MR HoA < CR HoA */ 
	if (config.mn_home_ip_addr.s_addr < cr->cr_hoa.s_addr)
	{
		print_trace(mn_traces, "Reply with REGREP_PREACCEPT "
			"(my HoA smaller than the CR's)\n");
		cr->tunneling_as_mr	= 1;  
		cr->reply_code 		= REGREP_PREACCEPTED;
	} else {
		print_trace(mn_traces, "Reply with REGREP_ACCEPT "
			"(my HoA bigger than the CR's)\n"); 
		cr->reply_code 		= REGREP_ACCEPTED;
		cr->tunneling_as_mr	= 0;
	}

	/*--------------------------------------------------------------*
	 | Solicit CR from HA or send a Reply							|
	 *--------------------------------------------------------------*/
	if (!cr->confirmed_by_ha) {
		DEBUG(DEBUG_MESSAGES, "Solicit CR %s from HA\n", inet_ntoa(cr->cr_hoa));
		/* Send Reg Req with RO Capab extension S=1 */
		return send_registration(REG_CONNECT);
	} else {
		return send_regrep_to_cr(cr);
	}
}

int handle_cr_regrep(unsigned char *msg, struct sockaddr_in *src_addr, int bytes_received) { 	
	int parse_res;
	struct msg_extensions ext;
	struct ro_entry *cr;
	int i, plen;
	char ip_str[INET6_ADDRSTRLEN];
	
	DEBUG(DEBUG_MESSAGES, "Registration reply received from CR.\n");

	/*----------------------------------------------------------*
 	 | Parse reply message and check its validity				|
	 *----------------------------------------------------------*/
	parse_res = parse_msg(msg, bytes_received, &ext);
   	
   	if (parse_res != 0 || ext.double_auth_ext != 0) {
		char *reason = "N/A";
		DEBUG(DEBUG_MESSAGES, "Dropping invalid message\n");
		mn.discarded_msgs++;
		if (parse_res == -1)
			reason = "unknown extension";
		else if (parse_res == -2)
			reason = "malformed message";
		else if (parse_res == -3 || parse_res == -4 || parse_res == -5)
			reason = "unknown vendor extension";
		else if (ext.double_auth_ext != 0)
			reason = "double authentication extension";
		print_trace(mn_traces, "Dropping an invalid Registration Reply from CR. Reason: %s\n",
			reason);
		report_discarded_msg(msg, bytes_received, src_addr, reason);
		return 0;
	}

	/*----------------------------------------------*
	 | Find the CR from the RO Cache				|
 	 *----------------------------------------------*/
	assert(ext.rep != NULL);
	cr = (struct ro_entry *) 
		get_ro_entry_by_sent_reg_id(&ext.rep->id); 
   	
	if (cr == NULL)
	{
		DEBUG(DEBUG_MESSAGES, "No Request sent with this id %08x %08x -> discard Reply\n",
			(__u32) ntohl(((__u32 *) &ext.rep->id)[0]), 
			(__u32) ntohl(((__u32 *) &ext.rep->id)[1]));
			
		print_trace(mn_traces, "Received Registration Reply from a CR with an unknown "
			"id %08x %08x -> discard Reply\n",
			(__u32) ntohl(((__u32 *) &ext.rep->id)[0]), 
			(__u32) ntohl(((__u32 *) &ext.rep->id)[1]) );	
		
		return -1;
	}	
	
	cr->regreq_online 				= 0;
	cr->reply_wait_timer.tv_sec		= 0;
	cr->reply_wait_timer.tv_usec	= 0;
	cr->reg_attempt_count			= 0;
	
	/* Send reregistration in ext.rep->lifetime seconds */
	if (ext.rep->lifetime > 0)
	{
		DEBUG(DEBUG_TIMERS, "Reply's lifetime: update CR's binding in %d seconds\n", ntohs(ext.rep->lifetime));
		add_seconds_to_timer(&cr->rereg_timer, ntohs(ext.rep->lifetime));
	}
	/*----------------------------------------------*
	 | Check the Reply Code.						|
 	 *----------------------------------------------*/
 	/* ACCEPT: Tunneling as MR
 	 * PREACCEPT: Tunneling as CR 
 	 */	
	if (ext.rep->code == REGREP_ACCEPTED ||
		ext.rep->code == REGREP_ACCEPTED_NO_SB)
	{
		cr->tunneling_as_mr	= 1;
		print_trace(mn_traces, "Received Registration ACCEPT from CR %s with id %08x %08x\n",
			inet_ntoa(cr->cr_hoa),
			(__u32) ntohl(((__u32 *) &ext.rep->id)[0]), 
			(__u32) ntohl(((__u32 *) &ext.rep->id)[1]));
		DEBUG(DEBUG_MESSAGES, "Registration ACCEPT -> Tunneling role MR\n");
	} else if (ext.rep->code == REGREP_PREACCEPTED) {
		cr->tunneling_as_mr	= 0;
		print_trace(mn_traces, "Received Registration PRE-ACCEPT from CR %s with id %08x %08x\n",
			inet_ntoa(cr->cr_hoa),
			(__u32) ntohl(((__u32 *) &ext.rep->id)[0]), 
			(__u32) ntohl(((__u32 *) &ext.rep->id)[1]));
		DEBUG(DEBUG_MESSAGES, "Registration PRE-ACCEPT -> Tunneling role CR\n");
	} else {
		cr->tunneling_as_mr	= -1;
		print_trace(mn_traces, "Received Registration FAILURE from CR %s with id %08x %08x\n",
			inet_ntoa(cr->cr_hoa),
			(__u32) ntohl(((__u32 *) &ext.rep->id)[0]), 
			(__u32) ntohl(((__u32 *) &ext.rep->id)[1]));
		
		/* If the KRm was not valid -> send HoTI + CoTI */
		if (ext.rep->code == REGREP_MN_FAILED_AUTH_FA)
		{
			print_trace(mn_traces, "Authenticator was not valid -> need to calculate a new KRm\n");
			return go_to_resend_hoti_coti_state(cr);
		}
	}
	
	/*----------------------------------------------*
	 | UDP Tunnel Reply Extension.					|
 	 *----------------------------------------------*/
	if (ext.tnl_rep != NULL) 
	{
		DEBUG(DEBUG_INFO, " * UDP Tunnel Reply: ");
		if (ext.tnl_rep->reply_code < 64) 
		{
			DEBUG(DEBUG_INFO, "ACCEPT\n");
			cr->udp_tnl_accepted = 1;
						
		} else  {
			DEBUG(DEBUG_INFO, "DECLINE\n");
		}
		if (ext.tnl_rep->keepalive_interval == 0) {
			DEBUG(DEBUG_INFO, "  - keepalive interval 0 --> use default\n");
		} else  {
			DEBUG(DEBUG_INFO, "  - keepalive interval: %d sec\n", ext.tnl_rep->keepalive_interval);
			// TODO: if a keepalive interval defined here --> use that instead of the average RRT?
		}
	}
	
	/*----------------------------------------------*
	 | Mobile NW (NEMO) Ack Extension.				|
	 *----------------------------------------------*/
	for (i = 0; i < ext.nemo_nw_amount; i++) {
		if (ext.nemo_ack[i] != NULL) {
			
			inet_ntop(AF_INET, &ext.nemo_ack[i]->prefix, ip_str, INET6_ADDRSTRLEN);
	 		plen = ext.nemo_ack[i]->prefix_length;

			/* code */
			/* TODO React to these codes */
			if (ext.nemo_ack[i]->code == MOBNET_SUCCESS) 
			{
				print_trace(mn_traces, 	" - NEMO network %s/%d accepted\n", ip_str, plen);
				DEBUG(DEBUG_INFO,		" * NEMO network %s/%d accepted\n", ip_str, plen);
			} else if (ext.nemo_ack[i]->code == MOBNET_INVALID_PREFIX_LEN) 
			{
				print_trace(mn_traces, " - NEMO network %s/%d rejected: invalid prefix length\n", 
					ip_str, plen);
				DEBUG(DEBUG_INFO, " - NEMO network %s/%d rejected: invalid prefix length\n", 
					ip_str, plen);
			} else if (ext.nemo_ack[i]->code == MOBNET_UNAUTHORIZED) 
			{
				print_trace(mn_traces, " - NEMO network %s/%d rejected: unauthorized\n", 
					ip_str, plen);
				DEBUG(DEBUG_INFO, " - NEMO network %s/%d rejected: unauthorized\n", 
					ip_str, plen);
			} else if (ext.nemo_ack[i]->code == MOBNET_FWDING_SETUP_FAILED) {
				print_trace(mn_traces, " - NEMO network %s/%d rejected: forwarding setup failed\n", 
					ip_str, plen);
				DEBUG(DEBUG_INFO, " - NEMO network %s/%d rejected: forwarding setup failed\n", 
					ip_str, plen);
			}
		} else {
			DEBUG(DEBUG_INFO, "Error in NEMO nw parsing (nw amount %d but nw %d NULL)!\n", 
				ext.nemo_nw_amount, i);
		}
	}

	/*------------------------------*
	 | CoA extension				|
	 *------------------------------*/
	if (ext.coa_ext != NULL) {	
		DEBUG(DEBUG_INFO, " * CoA ext\n");
		handle_coa_ext(cr, ext.coa_ext);
	}

	/*--------------------------------------------------------------*
	 | Mobile-Correspondent Authentication extension				|
	 *--------------------------------------------------------------*/
	if (ext.mr_cr_auth != NULL) {
		DEBUG(DEBUG_INFO, " * MR-CR Auth\n");
		
	}
	
	if (cr->udp_tnl_accepted) 
	{
		if (cr->rr_state == RR_INACTIVE) {
			DEBUG(DEBUG_INFO, "UDP tunneling accepted && RR-state INACTIVE "
				"-> Open a TUN device\n");
			/*--------------------------------------------------*
		 	 | Open a TUN device for this direct CR connection	|
			 *--------------------------------------------------*/
			/* Open a TUN device but don't start routing through it until 
			 * a ping has succeeded through MR CoA - CR CoA */
			open_tunnel_to_cr(cr);
		}
		if (ext.rep->code == REGREP_ACCEPTED)
		{
			DEBUG(DEBUG_INFO, "Registration ACCEPT -> Start tunneling "
				"MR CoA:reg_port <==> CR CoA:434\n");
			if (cr->cr_coa_count >= 1) {
				/*--------------------------------------------------*
			 	 | ACCEPT: Send ping from MR CoA to CR CoA 			|
				 *--------------------------------------------------*/
				/* If ping reply gets received 
				 *  -> start tunneling to this CR CoA
				 * Else 
				 *  -> send ping to another CR CoA if there was more in the CoA Ext */
				return send_initial_ping_to_CR(cr);
			}
		} else if (ext.rep->code == REGREP_PREACCEPTED) 
		{
			DEBUG(DEBUG_INFO, "Registration PRE ACCEPT: wait for connection init to "
				"one of my CoAs to port 434\n");
			/*----------------------------------------------------------------------*
		 	 | PRE ACCEPT: Wait for the CR get an initial ping through to my 434	|
			 *---------------------------------------------------------------------*/
			 cr->waiting_for_ping = 1;
		}
	}
	
	return 0;
}

/* handle_cr_msg: Handle message from other MR ( == CR).
 * 
 */
int handle_cr_msg(int s) {
	char msg[MAXMSG];
	char cdata[256];
	struct iovec iov;
	unsigned int addr_len;
	struct sockaddr_in src_addr;
	int bytes_received;	
	struct msghdr mh;	/* Structure describing messages sent by sendmsg and received by recvmsg */
	struct cmsghdr *cmsg;
	struct in_pktinfo *in_pkt;
	struct in_addr pkt_receive_addr;
	addr_len = sizeof(src_addr);

	memset(&src_addr, 0, sizeof(src_addr));
	memset(&mh, 0, sizeof(mh));
	iov.iov_base 		= msg;
	iov.iov_len 		= MAXMSG;
	mh.msg_name 		= &src_addr;
	mh.msg_namelen 		= sizeof(src_addr);	
	mh.msg_iov 			= &iov;				/* Vector of data to send/receive into. */
	mh.msg_iovlen 		= 1;
	mh.msg_control 		= &cdata;			/* Ancillary data (eg BSD filedesc passing). */
	mh.msg_controllen 	= sizeof(cdata);

	/*--------------------------------------------------------------*
	 | Get message from the socket to the mh.						|
	 *--------------------------------------------------------------*/
	bytes_received = recvmsg(s, &mh, 0);

	DEBUG(DEBUG_MESSAGES, "Received %d bytes from %s:%d, fd %d ", bytes_received,
	      inet_ntoa(src_addr.sin_addr), ntohs(src_addr.sin_port), s);

	if (bytes_received < 0) {
		LOG2(DEBUG_MESSAGES, "recvfrom failed - %s\n",
		     strerror(errno));
		return -1;
	}
	
	/*------------------------------------------------------------------*
	 | Get the IP address the message was sent to.						|
	 *------------------------------------------------------------------*/
	/* Receive auxiliary data in msgh - check man ip and man cmsg */
	for (cmsg = CMSG_FIRSTHDR(&mh); cmsg != NULL; cmsg = CMSG_NXTHDR(&mh, cmsg)) 
	{
		if (cmsg->cmsg_level == IPPROTO_IP && 
       		cmsg->cmsg_type == IP_PKTINFO) 
		{
			in_pkt = (struct in_pktinfo *) CMSG_DATA(cmsg);
			pkt_receive_addr.s_addr = in_pkt->ipi_addr.s_addr;
			DEBUG(DEBUG_MESSAGES, "into IP %s\n",
			inet_ntoa(pkt_receive_addr));
			break;
		}
	}
	if (cmsg == NULL) {
		/* Error: IP_TTL not enabled or small buffer
		 * or I/O error.
		 */
		DEBUG(DEBUG_MESSAGES, "I/O error with recvmsg!\n");
	}

	/*------------------------------------------------------------------*
	 | Check the MIP msg type.											|
	 *------------------------------------------------------------------*/	 
	if (*msg == TBA_MIP1) {					/* HoTI */
		return handle_hoti(msg, &src_addr);
	} else if (*msg == TBA_MIP2) {			/* CoTI */
		return handle_coti(msg, &src_addr);
	} else if (*msg == REG_REQ) {			/* Registration Request from a CR */
		return handle_cr_regreq(msg, &src_addr, bytes_received);
	} else if (*msg == TNL) {				/* TNL */
		return handle_tnl_msg_mn(msg, &src_addr, bytes_received, s, &pkt_receive_addr);
	} else if (*msg == TBA_MIP3) {			/* HoT */
		return handle_hot(msg, &src_addr);
	} else if (*msg == TBA_MIP4) {			/* CoT */
		return handle_cot(msg, &src_addr);
	} else if (*msg == REG_REP) {			/* Registration Reply from a CR */
		struct ro_entry *cr;
		cr = (struct ro_entry *) get_ro_entry_by_hoa(src_addr.sin_addr);
		if (cr != NULL) {
			DEBUG(DEBUG_MESSAGES, "Reply's src address %s is a CR HoA "
				"-> Reply from a CR\n",	inet_ntoa(src_addr.sin_addr));
			return handle_cr_regrep(msg, &src_addr, bytes_received);
		}
	}
	return 0;
}
/* Handle registration message from socket s.
 * Returns 1 if registration should be re-tried 
 */
int handle_registration(int s)
{
	char msg[MAXMSG];
	unsigned int len;
	struct sockaddr_in addr;
	int n, res;
	struct msg_extensions ext;
	len = sizeof(addr);

	/* Reg rerty timer off */
	mn.reg_retry_time.tv_sec 	= 0;
	mn.reg_retry_time.tv_usec 	= 0;
	mn.init_reply_received		= 1;
	
  	n = recvfrom(s, msg, MAXMSG, 0, (struct sockaddr*) &addr, &len);
  	if (n < 0) {
		LOG2(LOG_ERR, "registration recv: %s\n", strerror(errno));
		return 0;
	}
  	DEBUG(DEBUG_MESSAGES, "Received %d bytes from %s:%d\n", n,
    	inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));
    	
	/*------------------------------------------------------*
	 | Check the message type (if not Reply from the HA)	|
	 *------------------------------------------------------*/	
	if (*msg == TNL) {
		/* MIP Tunnel Data Message */
		handle_tnl_msg_mn(msg, &addr, n, s, NULL);
		/* Never return bigger than 0 (this would call reinitiate_state()) */
		return 0;
	} else if (*msg == TBA_MIP4) {
		/* CoT (shouldn't come through this socket) */
		DEBUG(DEBUG_MESSAGES, "Warning: CoT received in the HA registration socket!\n");
		return handle_cot(msg, &addr);		
	} else if (*msg == REG_REP) {
		/* Registration Reply from a CR */
		struct ro_entry *cr;
		cr = (struct ro_entry *) get_ro_entry_by_hoa(addr.sin_addr);
		if (cr != NULL) {
			DEBUG(DEBUG_MESSAGES, "Reply's src address %s is a CR HoA "
				"-> Reply from a CR\n",	inet_ntoa(addr.sin_addr));
			DEBUG(DEBUG_MESSAGES, "Warning: CR Reply received in the HA registration socket!\n");
			return handle_cr_regrep(msg, &addr, n);
		}
	}
	
	/*----------------------------------------------------------*
 	 | Parse the Reply and it's extensions.						|
	 *----------------------------------------------------------*/
	res = parse_msg(msg, n, &ext);
	
   	if (res != 0 || ext.double_auth_ext != 0) {
		char *reason = "N/A";
		DEBUG(DEBUG_MESSAGES, "Dropping invalid message\n");
		mn.discarded_msgs++;
		if (res == -1)
			reason = "unknown extension";
		else if (res == -2)
			reason = "malformed message";
		else if (res == -3 || res == -4 || res == -5)
			reason = "unknown vendor extension";
		else if (ext.double_auth_ext != 0)
			reason = "double authentication extension";
		report_discarded_msg(msg, n, &addr, reason);
		return 0;
	}
  	if (IS_REGREP_ACCEPTED(ext.rep->code)) {
		print_trace(mn_traces, "Received registration ACCEPT from HA\n");
  	}
	mn.try_to_fix_sec_assoc = 0;
	if (is_valid_reply(&ext, addr)) {
		/*----------------------------------------------*
 		 | Reply was valid.								|
	 	 *----------------------------------------------*/
		mn.prev_req_replied = 1;
		time(&mn.last_reply_rcvd);
		mn.last_reply_code = ext.rep->code;
		mn.last_ha_nonce = ext.rep->id[0];
		
		/*----------------------------------------------*
 		 | UDP Tunnel Reply Extension.					|
	 	 *----------------------------------------------*/
		if (ext.tnl_rep != NULL) {
			DEBUG(DEBUG_INFO, "[mn_reg.c] handle_registration(): \n\tUDP Tunnel Extension\n");
			if (ext.tnl_rep->reply_code < 64) {
				DEBUG(DEBUG_INFO, "\t\tHA accepts UDP tunneling\n");
			} else  {
				DEBUG(DEBUG_INFO, "\t\tHA does not accept UDP tunneling\n");
			}
			if (ext.tnl_rep->keepalive_interval == 0) {
				DEBUG(DEBUG_INFO, "\t\tkeepalive interval: 0 --> Use default\n");
			} else  {
				DEBUG(DEBUG_INFO, "\t\tkeepalive interval: %d\n", ext.tnl_rep->keepalive_interval);
				// TODO: if a keepalive interval defined here --> use that instead of the average RRT?
			}	
		}
		
		/*----------------------------------------------*
		 | Mobile NW (NEMO) Ack Extension.				|
		 *----------------------------------------------*/
		if (ext.nemo_nw_amount >= 0) {
			int i, plen;
	 		char ip_str[INET6_ADDRSTRLEN];
			for (i = 0; i < ext.nemo_nw_amount; i++) {
				if (ext.nemo_ack[i] != NULL) {
					inet_ntop(AF_INET, &ext.nemo_ack[i]->prefix, ip_str, INET6_ADDRSTRLEN);
			 		plen = ext.nemo_ack[i]->prefix_length;
			 		
			 		DEBUG(DEBUG_INFO, "\tNEMO Ack Extension %d\n", i);
			 		/* subtype */
			 		if (ext.nemo_ack[i]->subtype == 1) {
			 			DEBUG(DEBUG_INFO, "\t\tsubtype: Explicit Mode Ack\n");
			 		} else if (ext.nemo_ack[i]->subtype == 2) {
			 			DEBUG(DEBUG_INFO, "\t\tsubtype: Implicit Mode Ack\n");
			 		}
					/* code */
					/* TODO React to these codes */
					if (ext.nemo_ack[i]->code == MOBNET_SUCCESS) {
						print_trace(mn_traces, " - NEMO network %s/%d accepted\n", ip_str, plen);
						DEBUG(DEBUG_INFO, "\t\tcode: MOBNET_SUCCESS\n");
					} else if (ext.nemo_ack[i]->code == MOBNET_INVALID_PREFIX_LEN) {
						print_trace(mn_traces, " - NEMO network %s/%d rejected: invalid prefix length\n", 
							ip_str, plen);
						DEBUG(DEBUG_INFO, "\t\tcode: MOBNET_INVALID_PREFIX_LEN\n");
					} else if (ext.nemo_ack[i]->code == MOBNET_UNAUTHORIZED) {
						print_trace(mn_traces, " - NEMO network %s/%d rejected: unauthorized\n", 
							ip_str, plen);
						DEBUG(DEBUG_INFO, "\t\tcode: MOBNET_UNAUTHORIZED\n");
					} else if (ext.nemo_ack[i]->code == MOBNET_FWDING_SETUP_FAILED) {
						print_trace(mn_traces, " - NEMO network %s/%d rejected: forwarding setup failed\n", 
							ip_str, plen);
						DEBUG(DEBUG_INFO, "\t\tcode: MOBNET_FWDING_SETUP_FAILED\n");
					}
					/* prefix length */
					DEBUG(DEBUG_INFO, "\t\tprefix length: %d\n", plen);
					/* prefix */
					DEBUG(DEBUG_INFO, "\t\tprefix: %s\n", ip_str);
				}
			}
		}
		
		/*----------------------------------------------*
		 | Route Optimization (HAaRO) Reply.			|
		 *----------------------------------------------*/
		if (ext.ro_reply != NULL) 
		{
			DEBUG(DEBUG_INFO, "\tRO Reply (HAaRO) Extension\n");
			if (ext.ro_reply->code == RO_OK) {
				DEBUG(DEBUG_INFO, "\t\tRoute Optimization ok!\n");
			} else {
				DEBUG(DEBUG_INFO, "\t\tRoute Optimization declined!\n");
			}
		}
		
		/*----------------------------------------------*
		 | RO (HAaRO) Prefix Advertisement.				|
		 *----------------------------------------------*/
		if (ext.ro_prefix_adv.head != NULL)
		{
			DEBUG(DEBUG_INFO, "\tRO (HAaRO) Prefix Advertisement\n");
			DEBUG(DEBUG_INFO, "\t\ttype %d\n", 		ext.ro_prefix_adv.head->type);
			DEBUG(DEBUG_INFO, "\t\tsubtype %d\n", 	ext.ro_prefix_adv.head->subtype);
			DEBUG(DEBUG_INFO, "\t\tlength %d\n", 	ntohs(ext.ro_prefix_adv.head->length));

			handle_ro_prefix_adv(ext.ro_prefix_adv.head);
		
		}
		
		if (ext.challenge != NULL) {
			int len;
			gettimeofday(&mn.last_challenge_time, NULL);
			len = GET_CHALLENGE_EXT_LEN(ext.challenge);
			if (mn.last_challenge_ext == NULL ||
			    GET_CHALLENGE_EXT_LEN(mn.last_challenge_ext) !=
			    len) {
				if (mn.last_challenge_ext != NULL)
					free(mn.last_challenge_ext);
				mn.last_challenge_ext =
					(struct challenge_ext *) malloc(len);
			}

			if (mn.last_challenge_ext != NULL)
				memcpy(mn.last_challenge_ext, ext.challenge,
				       len);
			else
				DEBUG(DEBUG_INFO, "Could not malloc() memory "
				      "for challenge extension\n");
		}
      	if (IS_REGREP_ACCEPTED(ext.rep->code)) {
        	handle_reg_accept(&ext);
      	} else {
			return handle_reg_denied(&ext);
      	}
 	} else {
		report_discarded_msg(msg, n, &addr, "invalid reply");
		mn.discarded_msgs++;
	}

	return 0;
}
