/* $Id: binding.c,v 1.33 2001/02/17 15:29:55 jm Exp $
 * Mobility bindings used in FAs and HAs.
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 * 
 * Additional modifications Copyright 2010 Juho Paaso, Aalto 
 * University / Department of Communication and Networking"
 */

#define DEBUG_FLAG 'B'

#include "owntypes.h"

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <time.h>
#include <assert.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <string.h>	

#include "binding.h"
#include "debug.h"

/* Home and foreign agents keep track of
 * registered mobile nodes with mobility bindings. */

/*
 * Bindings have a lifetime. The current time in seconds
 * since 0:00:00 UTC, January 1, 1970 can be retrieved with the time
 * system call (see man 2 time). The expiration time for a binding
 * is set in exp_time. When the current time is equal to or
 * more than exp_time the binding should be removed.
 *
 * Both foreign and home agents need to know where to forward
 * packets destined for the mobile node; this is indicated by the
 * lower_addr field. The session key for the tunnel is stored
 * in the key field. The corresponding security parameter index for
 * authentication method used is stored in the spi field.
 */

struct bindingtable {
	struct hashtable *hashtable;    /* hash table for bindings */
	unsigned int hashtablesize;     /* number of entries in hash table */
	struct list *shorttimelist;     /* short time list (STL) table */
	unsigned int shorttimelistsize; /* amount of entries in STL table */
	struct list longtimelist;		/* should be empty all the time */
	struct list noexptimelist;		/* entries that do not expire */
	unsigned int firstindex;        /* entry to be reported next */
	unsigned int currentindex;		/* current time index */
};

/* defines */

#define ASSERT assert

#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE 1
#endif

/* Test cases */

/*#define BINDING_TEST_CASE_I05*/
/*TEST CASE I10: MANUALLY*/
/*#define BINDING_TEST_CASE_I12*/

/* prototypes */

static int binding_hash(void *key, const int tablesize);
static int binding_compare(void *key, struct node *cmprd);
static int binding_getexpired_alreadyexpired(struct bindingtable *table,
					     int idx, int *checklongtimelist,
					     int *listempty);
static int binding_getexpired_advance(struct bindingtable *table, int advance,
				      int idx, int *checklongtimelist);
static int binding_getexpired_checkmaxadvancetime(struct bindingtable *table,
						  int left, int *advance);
static void binding_check_longtimelist(struct bindingtable *table,
				       const int advance);

/* "Ixx" are test case IDs to help in the design of the test module */

/* Init binding table
 *
 * INPUT VALUES:
 *
 * I01) maxbindings is negative
 * I02) maxbindings is zero
 * I03) maxlifetime is negative
 * I04) maxlifetime is zero (valid value)
 * I05) not enough memory for struct bindingtable (difficult to test autom.)
 *      - requires changes in code, will be tested manually
 * I06) maxbindings is too large for malloc to succeed (not enough memory)
 * I07) maxlifetime is too large for malloc to succeed (not enough memory)
 * I08) realistic values
 */

/**
 * binding_init:
 * @maxbindings:
 * @maxlifetime:
 *
 * Initializes a bindingtable for storing bindings.
 *
 * Returns: A reference that is used in further operations on 
 *          the bindingtable.
 */
struct bindingtable *
binding_init(const int maxbindings, const int maxlifetime)
{
	int i;
	struct bindingtable *table;

	if (maxbindings < 1) {
		DEBUG(DEBUG_FLAG,
		      "binding_init: maxbindings must be at least 1\n");
		return NULL;
	}
	if (maxlifetime < 0) {
		DEBUG(DEBUG_FLAG,
		      "binding_init: maxlifetime must be positive\n");
		return NULL;
	}
#ifdef BINDING_TEST_CASE_I05
	table = NULL;
#else
	table = malloc(sizeof(struct bindingtable));
#endif
	if (table == NULL) {
		DEBUG(DEBUG_FLAG,
			"binding_init: not enough memory for "
			"struct bindingtable\n");
		return NULL;
	}

	i = (maxbindings > MAX_BINDING_HASH ? MAX_BINDING_HASH : maxbindings);
	table->hashtablesize = hashtable_find_prime(2 * i);
	ASSERT(table->hashtablesize > 0);
	ASSERT(table->hashtablesize >= 2 * i);
	table->hashtable = hashtable_init(table->hashtablesize);
	if (table->hashtable == NULL) {
		DEBUG(DEBUG_FLAG,
			"binding_init: hash table creation failed\n");
		free(table);
		return NULL;
	}
	if (maxlifetime >= MAX_BINDING_SHORT_LIST)
		table->shorttimelistsize = MAX_BINDING_SHORT_LIST;
	else
		table->shorttimelistsize = maxlifetime + 1;
	table->shorttimelist = malloc(table->shorttimelistsize *
				      sizeof(struct list));
	if (table->shorttimelist == NULL) {
		DEBUG(DEBUG_FLAG,
			"binding_init: not enough memory for "
			"a table of %d 'struct list' items\n",
			table->shorttimelistsize);
		hashtable_destroy(table->hashtable);
		free(table);
		return NULL;
	}
	for (i = 0; i < table->shorttimelistsize; i++) {
		list_init(&table->shorttimelist[i]);
	}
	list_init(&table->longtimelist);
	list_init(&table->noexptimelist);
	table->firstindex = 0;
	table->currentindex = 0;
	return table;
}

/* Destroy all bindings
 *
 * NOTE! All entries must be removed before calling this!
 *
 * INPUT VALUES:
 *
 * I09) Table is empty
 * I10) Table is not empty (difficult to test automatically, assertions fail)
 *      - test will be done manually
 */

/**
 * binding_destroy:
 * @table:
 *
 * Destroy all bindings. binding_destroy() should be called to destroy a
 * bindingtable when it is not needed any more. All entries from the table
 * must be removed before doing this.
 */
void
binding_destroy(struct bindingtable *table)
{
	int i;

	ASSERT(table != NULL);
	for (i = 0; i < table->shorttimelistsize; i++) {
		ASSERT(list_is_empty(&table->shorttimelist[i]) == TRUE);
	}
	ASSERT(list_is_empty(&table->longtimelist) == TRUE);
	ASSERT(list_is_empty(&table->noexptimelist) == TRUE);
	free(table->shorttimelist);
	hashtable_destroy(table->hashtable);
	free(table);
}

/* Add binding
 *
 * INPUT VALUES:
 *
 * I11) Timeout is less than zero
 * I12) Unable to add to hashtable (this requires a buggy hash function)
 *      - requires changes in code, test will be done manually
 * I13) Timeout is longer than shorttimelistsize
 * I14) Timeout is longer than maxuseshortlist (expired bindings exist)
 * I15) Timeout is a reasonable value (~ less than firstindex - currentindex)
 *
 */

/**
 * binding_add:
 * @table:
 * @entry:
 *
 * Adds a bindingentry into the bindingtable.
 *
 * NOTE: Use binding_getexpired() to advance the timer to current time before
 * adding new bindings. And for optimal performance, process all expired
 * bindings before adding new ones.
 *
 * Returns: TRUE on success
 */
int
binding_add(struct bindingtable *table, struct bindingentry *entry)
{
	int idx, maxuseshortlist;
	struct list *addlist;

	ASSERT(table != NULL);
	ASSERT(entry != NULL);
	if (entry->timeout < 0) {
		DEBUG(DEBUG_FLAG,
			"binding_add: unable to add entry with timeout %d "
			"that is less than zero\n",
			entry->timeout);
		return FALSE;
	}
	DEBUG(DEBUG_FLAG,
	      "binding_add: entry = 0x%08x, &entry->mn_addr = 0x%08x\n",
	      (int) entry, (int) &entry->mn_addr);
	DEBUG(DEBUG_FLAG, "\tkey: (MN=%s,HA=", inet_ntoa(entry->mn_addr));
	DEBUG(DEBUG_FLAG, "%s)\n", inet_ntoa(entry->ha_addr));

	list_init_node(&entry->hashnode);
	list_init_node(&entry->timenode);

	if (hashtable_add(table->hashtable, binding_hash,
			  &entry->mn_addr, &entry->hashnode) == FALSE) {
		DEBUG(DEBUG_FLAG,
			"binding_add: unable to add entry to hashtable\n");
		return FALSE;
	}
	idx = table->currentindex;
	ASSERT(idx >= 0);
	ASSERT(idx < table->shorttimelistsize);
	maxuseshortlist = table->firstindex + (table->shorttimelistsize - 1) -
			  table->currentindex;
	if (maxuseshortlist >= table->shorttimelistsize) {
		maxuseshortlist -= table->shorttimelistsize;
	}
	DEBUG(DEBUG_FLAG,
	      "binding_add: shorttimelistsize = %d, maxuseshortlist = %d\n",
	      table->shorttimelistsize, maxuseshortlist);
	ASSERT(maxuseshortlist >= 0);
	ASSERT(maxuseshortlist < table->shorttimelistsize);
	if (entry->timeout >= INFINITE_TIMEOUT)
		addlist = &table->noexptimelist;
	else
		addlist = &table->longtimelist;
	if (entry->timeout >= table->shorttimelistsize) {
		DEBUG(DEBUG_FLAG,
			"binding_add: timeout %d longer than "
			"shorttimelistsize %d, long list used\n",
			entry->timeout, table->shorttimelistsize);
	} else if (entry->timeout > maxuseshortlist) {
		DEBUG(DEBUG_FLAG,
			"binding_add: timeout %d longer than "
			"maxuseshortlist %d, long list used\n"
			"binding_add: firstindex %d, currentindex %d\n",
			entry->timeout, maxuseshortlist,
			table->firstindex, table->currentindex);
	} else {
		ASSERT(entry->timeout < INFINITE_TIMEOUT);
		idx = idx + entry->timeout;
		if (idx >= table->shorttimelistsize) {
			idx -= table->shorttimelistsize;
		}
		ASSERT(idx >= 0);
		ASSERT(idx < table->shorttimelistsize);
		addlist = &table->shorttimelist[idx];
	}
	list_add_tail(addlist, &entry->timenode);
	return TRUE;
}

/* Fetch binding
 *
 * INPUT VALUES:
 *
 * I16) Bindingentry is not found
 * I17) Bindingentry is found
 *
 */

/**
 * binding_fetch:
 * @table:
 * @key:
 *
 * Fetch binding.
 * 
 * Returns: a bindingentry structure if any binding for the requested
 * mobile node is found.  
 */
struct bindingentry *
binding_fetch(struct bindingtable *table, struct bindingkey *key)
{
	struct bindingentry *entry;
	struct node *node;

	DEBUG(DEBUG_FLAG, "binding_fetch: key: (MN=%s,HA=",
	      inet_ntoa(key->mn_addr));
	DEBUG(DEBUG_FLAG, "%s,privHA=%i)\n", inet_ntoa(key->ha_addr),
	      key->priv_ha);
	node = hashtable_fetch(table->hashtable, binding_hash, key,
			       binding_compare);
	if (node == NULL) {
		DEBUG(DEBUG_FLAG, "binding_fetch: entry %08x,%08x not found in"
		      " the hashtable\n", key->mn_addr.s_addr,
		      key->ha_addr.s_addr);
		return NULL;
	}
	entry = (struct bindingentry *)
		(((char *) node) - offsetof(struct bindingentry, hashnode));
	return entry;
}

static struct bindingentry *found_binding;

/* fetch_by_nai_iter: iterator function for binding_fetch_by_nai.
 *
 * @node: the new binding_entry in the iteration
 * @nai_str: the NAI string
 * 
 * The function binding_iterator calls this function for each binding.
 * 
 * Returns: 1 to continue iteration, 0 to stop
 */
static int fetch_by_nai_iter(struct node *node, void *nai_str)
{
	struct bindingentry *entry = (struct bindingentry *) node;
	char *nai_param = (char *) nai_str;
	
	DEBUG(DEBUG_FLAG, "fetch_by_nai_iter: NAI %s, strlen %d\n", entry->nai, entry->nai_length);
	if (entry->nai_length == strlen(nai_param))
	{
		if (strncmp(entry->nai, nai_param, entry->nai_length) == 0) {
			DEBUG(DEBUG_FLAG, " - match!\n");
			found_binding = entry;
			return 0;
		}
		DEBUG(DEBUG_FLAG, " - lengths match but strings don't\n");
		return 1;
	}
	DEBUG(DEBUG_FLAG, " - lengths don't match\n");
	return 1;
}

/* binding_fetch_by_nai: 
 * 
 * Returns: bindingentry corresponding to the nai_str or NULL if none corresponds.
 */
struct bindingentry *binding_fetch_by_nai(struct bindingtable *table, char *nai_str)
{
	found_binding = NULL;
	DEBUG(DEBUG_FLAG, "binding_fetch_by_nai: Look for NAI %s, strlen %d\n", 
		nai_str, strlen(nai_str));
	hashtable_iterator(table->hashtable, fetch_by_nai_iter, nai_str);
	if (found_binding == NULL) {
		DEBUG(DEBUG_FLAG, "entry corresponding to the NAI not found!\n");
		return NULL;
	}
	return found_binding;
}

static int fetch_iter2(struct node *node, void *data)
{
	struct bindingentry *entry = (struct bindingentry *) node;
	struct in_addr *addr = (struct in_addr *) data;

	if (entry->mn_addr.s_addr == addr->s_addr) {
		found_binding = entry;
		return 0;
	}
	return 1;
}

/**
 * binding_fetch2:
 * @table:
 * @mn_addr:
 * 
 * This function is like binding_fetch(), but it does not use the full
 * binding key (only @mn_addr). It is a temporary fix due to the binding key
 * change. It should also be noted, that this function has to go through all
 * the binding entries as it cannot use hashtable without proper key.
 *
 * Returns: a bindingentry structure
 * if any binding for the requested mobile node is found.
 */
struct bindingentry *
binding_fetch2(struct bindingtable *table, struct in_addr *mn_addr)
{
	found_binding = NULL;
	hashtable_iterator(table->hashtable, fetch_iter2, mn_addr);
	if (found_binding == NULL) {
		DEBUG(DEBUG_FLAG, "binding_fetch2: entry %08x not found in "
		      "hashtable\n", mn_addr->s_addr);
		return NULL;
	}
	return found_binding;
}

/* fetch_by_coa_iter: helper function for binding_fetch_by_coa. */
static int fetch_by_coa_iter(struct node *node, void *co_addr)
{
	struct bindingentry *entry = (struct bindingentry *) node;
	struct in_addr *coa = (struct in_addr *) co_addr;

	if (entry->lower_addr.s_addr == coa->s_addr) {
		found_binding = entry;
		return 0;
	}
	return 1;
}
/* binding_fetch_by_coa: 
 * 
 * Returns: a pointer to a bindingentry corresponding to the given CoA
 * or NULL if none was found.
 */
struct bindingentry *
binding_fetch_by_coa(struct bindingtable *table, struct in_addr *coa)
{
	found_binding = NULL;
	hashtable_iterator(table->hashtable, fetch_by_coa_iter, coa);
	if (found_binding == NULL) {
		DEBUG(DEBUG_FLAG, "binding_fetch_by_coa: entry for CoA %s not found!\n",
			inet_ntoa(*coa));
		return NULL;
	}
	return found_binding;
}


static int fetch_iter_func(struct node *node, void *data)
{
	struct bindingentry *entry = (struct bindingentry *) node;
	int (*func)(struct bindingentry *) = data;

	if (func(entry)) {
		found_binding = entry;
		return 0;
	}
	return 1;
}


/**
 * binding_fetch_func:
 * @table: pointer to the binding table
 * @func: function to be used to select the binding
 * 
 * This function is like binding_fetch(), but it goes through all the binding
 * entries until a correct one is found. The given @func is used to decide
 * which binding entry is selected.
 *
 * Returns: a bindingentry structure for the selected binding or NULL if no
 * binding is selected.
 */
struct bindingentry *
binding_fetch_func(struct bindingtable *table,
		   int (*func)(struct bindingentry *binding))
{
	found_binding = NULL;
	hashtable_iterator(table->hashtable, fetch_iter_func, func);
	if (found_binding == NULL) {
		DEBUG(DEBUG_FLAG, "binding_fetch_func: entry not found in "
		      "hashtable\n");
		return NULL;
	}
	return found_binding;
}


/* Remove binding
 *
 * INPUT VALUES:
 *
 * I18) Bindingentry is valid
 *    ) Bindingentry is NULL
 *    ) Bindingentry is invalid (cannot be checked)
 *
 */

/**
 * binding_remove:
 * @entry:
 *
 * Removes the bindingentry from the bindingtable.
 */
void
binding_remove(struct bindingentry *entry)
{

	ASSERT(entry != NULL);
	if (entry->hashnode.succ == NULL)
		return; /* already removed */
	list_remove(&entry->hashnode);
	list_remove(&entry->timenode);
}

/* Get the amount of units (for example seconds) until the next binding
 * is going to be expired
 * 0 = expired
 * -1 = no bindings
 *
 * INPUT VALUES:
 *
 * I19) No bindings in the table
 * I20) Expired binding with timeout of 0 (firstindex == currentindex)
 * I21) Earlier expired binding (firstindex != currentindex)
 * I22) Next binding to be expired is in shorttimelist
 * I23) Next binding to be expired is the first in longtimelist
 * I24) Next binding to be expired is not the first but in longtimelist
 *    ) Table is NULL
 *
 */

/**
 * binding_nextexpiretime:
 * @table:
 *
 * Get the amount of units (for example seconds) until the next binding
 * is going to be expired
 *
 * Returns: the amount of time units until the
 * next entry expires. 0 = expired, -1 = no bindings
 */

int
binding_nextexpiretime(struct bindingtable *table)
{
	int idx, time;
	struct bindingentry *entry;
	struct node *timenode;

	ASSERT(table != NULL);
	idx = table->firstindex;
	for (;;) {
		ASSERT(idx >= 0);
		ASSERT(idx < table->shorttimelistsize);
		if (list_is_empty(&table->shorttimelist[idx]) == FALSE) {
			return 0;
		}
		if (idx == table->currentindex) break;
		idx++;
		if (idx >= table->shorttimelistsize) {
			idx -= table->shorttimelistsize;
		}
		table->firstindex = idx;
	}
	ASSERT(table->firstindex == table->currentindex);
	idx = table->currentindex;
	for (time = 1; time < table->shorttimelistsize - 1; time++) {
		idx++;
		if (idx >= table->shorttimelistsize) {
			idx -= table->shorttimelistsize;
		}
		if (list_is_empty(&table->shorttimelist[idx]) == FALSE) {
			return time;
		}
	}
	if (list_is_empty(&table->longtimelist) == TRUE) {
		return -1;
	}
	DEBUG(DEBUG_FLAG,
		"binding_nextexpiretime: shorttimelist empty, using "
		"longtimelist (slower)\n");
	timenode = list_get_first(&table->longtimelist);
	if (timenode == NULL) {
		return -1;
	}
	time = (((unsigned int) (-1)) >> 1); /* large one :) */
	while (timenode != NULL) {
		entry = (struct bindingentry *)
			(((char *) timenode) - offsetof(struct bindingentry,
							timenode));
		ASSERT(entry->timeout >= 0);
		if (entry->timeout < time) {
			time = entry->timeout;
		}
		timenode = list_get_next(&entry->timenode);
	}
	if (time == (((unsigned int) (-1)) >> 1)) {
		return -1;
	}
	return time;
}

/* Advance time and check if there are expired bindings
 *
 * INPUT VALUES:
 *
 * I25) Time to advance is less than zero
 * I26) Time to advance is zero
 * I27) Time to advance is greater than zero
 * I28) Advancetime is longer than allowed (expired bindings)
 * I29) No bindings in the table
 * I30) No expired bindings
 * I31) Expired bindings in shorttimelist
 * I32) Expired bindings in longtimelist
 *    ) Table is NULL
 *
 */

/**
 * binding_getexpired:
 * @table:
 * @advance:
 * 
 * Advance the time by specified amount of units and
 * returns a bindingentry if there is at least one that is expired.
 * If there are no expired entries, %NULL is returned.
 *
 * Usually this function is first called once to advance the timer to current
 * time. If there are expired bindings, the function is repeated with advance
 * value of zero until there are no longer expired bindings.
 */
struct bindingentry *
binding_getexpired(struct bindingtable *table, int advance)
{
	int idx, checklongtimelist, listempty, left;
	struct node *timenode;
	struct bindingentry *entry;

	DEBUG(DEBUG_FLAG,
	      "binding_getexpired: firstindex = %d, currentindex = %d, "
	      "advance = %d\n",
	      table->firstindex, table->currentindex, advance);
	if (advance < 0) {
		DEBUG(DEBUG_FLAG,
			"binding_getexpired: time to advance is %d that is "
			"less than zero, set to zero\n", advance);
		advance = 0;
	}
	idx = table->firstindex;
	checklongtimelist = FALSE;
	listempty = TRUE;
	idx = binding_getexpired_alreadyexpired(table, idx,
						&checklongtimelist,
						&listempty);
	ASSERT(idx == table->firstindex);
	left = advance;
	if (listempty == TRUE) {
		left = binding_getexpired_advance(table, left, idx,
						  &checklongtimelist);
		idx = table->firstindex;
	}
	left = binding_getexpired_checkmaxadvancetime(table, left, &advance);
	ASSERT(idx == table->firstindex);
	ASSERT(idx >= 0);
	ASSERT(idx < table->shorttimelistsize);
	table->currentindex = table->firstindex + left;
	if (table->currentindex >= table->shorttimelistsize) {
		table->currentindex -= table->shorttimelistsize;
	}
	ASSERT(table->currentindex >= 0);
	ASSERT(table->currentindex < table->shorttimelistsize);
	if (checklongtimelist == TRUE) {
		DEBUG(DEBUG_FLAG,
		      "binding_getexpired: binding_check_longtimelist()\n");
		binding_check_longtimelist(table, advance);
	}
	if (listempty == FALSE) {
		ASSERT(list_is_empty(&table->shorttimelist[idx]) == FALSE);
	}
	DEBUG(DEBUG_FLAG, "firstindex = %d, currentindex = %d, left = %d\n",
	      table->firstindex, table->currentindex, left);
	ASSERT(left >= 0);
	if (left == 0 && listempty == TRUE) {
		ASSERT(table->firstindex == table->currentindex);
		ASSERT(idx == table->firstindex);
	}
	timenode = list_get_first(&table->shorttimelist[idx]);
	if (timenode == NULL) {
		return NULL;
	}
	entry = (struct bindingentry *)
		(((char *) timenode) - offsetof(struct bindingentry,
						timenode));
	binding_remove(entry);
	return entry;
}

/**
 * binding_iterator:
 * @table: pointer to the binding table
 * @iterfunc: iterator function to be called for each binding entry
 * @data: argument for @iterfunc
 *
 * Iterate through each binding entry and call @iterfunc for each with @data
 * as the second argument.
 *
 * Returns: TRUE if the iteration was not stopped by iterator function
 * and FALSE if it was stopped.
 * 
 */
int
binding_iterator(struct bindingtable *table,
		 int (*iterfunc)(struct node *node, void *data),
		 void *data)
{
	if (table == NULL || table->hashtable == NULL)
		return 0;

	return hashtable_iterator(table->hashtable, iterfunc, data);
}


/**
 * binding_update:
 * @table: binding table
 * @entry: entry to be updated
 *
 * Updates the bindingentry timeout in the bindingtable. This needs to be
 * called when the lifetime of the binding has changed.
 *
 * Returns TRUE on success, FALSE on failure
 */
int
binding_update(struct bindingtable *table, struct bindingentry *entry)
{
	binding_remove(entry);
	return binding_add(table, entry);
}


/* Check for already expired bindings and update firstindex
 *
 * Sets listempty to FALSE if there are expired bindings
 * Sets checklongtimelist to TRUE if the longtimelist should be checked
 * (i.e. if the time is advanced)
 * Updates firstindex of the table
 * Returns the new index
 */

static int
binding_getexpired_alreadyexpired(struct bindingtable *table, int idx,
				  int *checklongtimelist, int *listempty)
{
	for (;;) {
		ASSERT(idx >= 0);
		ASSERT(idx < table->shorttimelistsize);
		if (list_is_empty(&table->shorttimelist[idx]) == FALSE) {
			*listempty = FALSE;
			break;
		}
		if (idx == table->currentindex) break;
		idx++;
		if (idx >= table->shorttimelistsize) {
			idx -= table->shorttimelistsize;
		}
		table->firstindex = idx;
		*checklongtimelist = TRUE;
	}
	ASSERT(idx == table->firstindex);
	return idx;
}

/* Advances the time to find the next expired binding
 *
 * Sets checklongtimelist to TRUE if the longtimelist should be checked
 * (i.e. if the time is advanced)
 * Updates firstindex and currentindex of the table
 * Returns the amount of time units left after advance
 */

static int
binding_getexpired_advance(struct bindingtable *table, int advance, int idx,
			   int *checklongtimelist)
{
	int i, left;

	left = advance;
	for (i = 0; i < advance; i++) {
		idx++;
		left--;
		if (idx >= table->shorttimelistsize) {
			idx -= table->shorttimelistsize;
		}
		if (idx == table->firstindex) {
			ASSERT(idx == table->currentindex);
			left = 0;
			idx = 0;
			table->currentindex = idx;
			break;
		}
		if (list_is_empty(&table->shorttimelist[idx]) == FALSE) {
			break;
		}
	}
	if (i > 0) {
		*checklongtimelist = TRUE;
	}
	table->firstindex = idx;
	table->currentindex = idx;
	return left;
}

/* Check the maximum time able to advance
 *
 * Makes advance smaller if too high
 * Returns the time left to advance
 */

static int
binding_getexpired_checkmaxadvancetime(struct bindingtable *table, int left,
				       int *advance)
{
	int maxadvancetime;

	maxadvancetime = table->firstindex - table->currentindex;
	if (maxadvancetime <= 0) {
		maxadvancetime += table->shorttimelistsize - 1;
	}
	DEBUG(DEBUG_FLAG,
	      "maxadvancetime = %d, firstindex = %d, currentindex = %d\n",
	      maxadvancetime, table->firstindex, table->currentindex);
	ASSERT(maxadvancetime < table->shorttimelistsize);
	if (left > maxadvancetime) {
		DEBUG(DEBUG_FLAG,
			"binding_getexpired_checkmaxadvancetime: time to "
			"advance is %d\n"
			"that is more than allowed because of expired "
			"bindings\n",
			*advance);
		left = maxadvancetime + left - *advance;
		*advance = maxadvancetime;
	}
	return left;
}

/* Check longtimelist */

/* NOTE:
 * To reduce debug-message flood: do _not_ use DEBUG_FLAG 'B'
 * when running the system as this function might be called quite frequently.
 * The printings are mainly to notify the administrator that
 * the system is using a slower method (longtimelist) and the
 * efficiency could be improved.
 */

static void
binding_check_longtimelist(struct bindingtable *table, const int advance)
{
	int maxuseshortlist;
	struct node *timenode;
	struct bindingentry *entry;

	if (advance < 0) {
		DEBUG(DEBUG_FLAG,
			"binding_check_longtimelist: time to advance is %d "
			"that is less than zero\n", advance);
		return;
	}
	timenode = list_get_first(&table->longtimelist);
	if (timenode == NULL) return;
	DEBUG(DEBUG_FLAG,
	      "binding_check_longtimelist: list is not empty!\n");
	maxuseshortlist = table->firstindex +
		(table->shorttimelistsize - 1) - table->currentindex;
	if (maxuseshortlist >= table->shorttimelistsize) {
		maxuseshortlist -= table->shorttimelistsize;
	}
	DEBUG(DEBUG_FLAG, "maxuseshortlist = %d\n", maxuseshortlist);
	ASSERT(maxuseshortlist >= 0);
	ASSERT(maxuseshortlist <= table->shorttimelistsize - 1);
	do {
		entry = (struct bindingentry *)
			(((char *) timenode) - offsetof(struct bindingentry,
							timenode));
		timenode = list_get_next(&entry->timenode);
		ASSERT(entry->timeout >= 0);
		if (entry->timeout >= INFINITE_TIMEOUT)
			continue; /* this entry will not expire */
		if (entry->timeout < advance) {
			DEBUG(DEBUG_FLAG,
				"binding_check_longtimelist: timeout %d lower "
				"than advance %d, set to zero\n",
				entry->timeout, advance);
			entry->timeout = 0;
		} else {
			DEBUG(DEBUG_FLAG,
				"binding_check_longtimelist: updating timeout "
				"%d to %d\n",
				entry->timeout, entry->timeout - advance);
			entry->timeout -= advance;
		}
		if (entry->timeout < maxuseshortlist) {
			binding_remove(entry);
			binding_add(table, entry);
		}
	} while (timenode != NULL);
}

/* Hash Function used in bindings */

static int
binding_hash(void *key, const int tablesize)
{
	struct bindingkey *bkey;

	bkey = (struct bindingkey *) key;
	/* The key is a bindingentry structure */
	return (int) (((__u32)bkey->mn_addr.s_addr +
		       (__u32)bkey->ha_addr.s_addr) % tablesize);
}

/* compare function used in bindings */

static int
binding_compare(void *key, struct node *cmprd)
{
	struct bindingkey *bkey;
	struct bindingentry *entry;

	bkey = (struct bindingkey *) key;
	entry = (struct bindingentry *)
		(((char *) cmprd) - offsetof(struct bindingentry, hashnode));
	if (entry->mn_addr.s_addr == bkey->mn_addr.s_addr &&
	    entry->ha_addr.s_addr == bkey->ha_addr.s_addr &&
	    entry->priv_ha == bkey->priv_ha)
		return TRUE;
	return FALSE;
}
