/* $Id: tunnel.c,v 1.87 2001/09/26 18:25:24 jm Exp $
 * Tunnel interface functions
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 * 
 * Additional modifications Copyright 2010 Juho Paaso, Aalto 
 * University / Department of Communication and Networking"
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <syslog.h>
#include <assert.h>
//#include <linux/if.h>
#include <linux/if_tun.h>

#include "tunnel.h"
#include "hashtable.h"
#include "debug.h"
#include "list.h"
#include "dyn_ip.h"
#include "util.h"
#include "../ha/ha.h"
#include "../ha/ha_config.h"

/* defines */
#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define DEBUG_FLAG 'T'
#define PURGE_TUNNEL_TIME 3 /* seconds */

#define LOG2(lev, fmt, args...) { DEBUG(DEBUG_FLAG, fmt, ## args); \
  syslog(lev, fmt, ## args); }

static char tunnel_dev_start[IFNAMSIZ];
static TUNNEL *found_tunnel;
/* Maximum numbers of concurrent tunnels
 * This can be increased if needed, but Linux kernel (at least 2.2.3) seems
 * to take quite a long time, when the number of interfaces goes beyond
 * 2048. */
#define MAX_TUNNEL_COUNT 2048
#define MAX_TUNNEL_INTS (MAX_TUNNEL_COUNT >> 5)
static unsigned int tunnel_numbers[MAX_TUNNEL_INTS];

#define MAX_TABLE_COUNT 256
#define MAX_TABLE_INTS (MAX_TABLE_COUNT >> 5)
static unsigned int table_numbers[MAX_TABLE_INTS];

#define DYN_SET_BIT(bits, bit) \
        (bits[(bit) / 32] |= (1 << ((bit) & 0x1f)))
#define DYN_CLR_BIT(bits, bit) \
        (bits[(bit) / 32] &= ~(1 << ((bit) & 0x1f)))
#define DYN_BIT(bits, bit) \
        (bits[(bit) / 32] & (1 << ((bit) & 0x1f)))

static void
debug_print_bitfield(char *name, unsigned int bits[], int ints)
{
#if 0
	int i;
	DEBUG(DEBUG_FLAG, "%s[", name);
	for (i = 0; i < ints; i++)
		DEBUG(DEBUG_FLAG, "%08x", bits[i]);
	DEBUG(DEBUG_FLAG, "]\n");
#endif
}


#define DELAY_UNCONNECTION_USEC 200000

struct delayed_connection_deletion {
	struct delayed_connection_deletion *next;
	struct timeval timeout; /* when the connection can be deleted */

	struct in_addr mn_addr;
	int reverse;
	int delete_route;
	int to_ha_table_id;
	int to_mn_table_id;
	char route_dev[IFNAMSIZ];
	char reverse_dev[IFNAMSIZ];
};

/* priority queue (sorted by increasing timeout) of delayed connection
 * deletions */
static struct delayed_connection_deletion *delayed = NULL;



#define TUNNEL_HASH_SIZE 256
static int
hashfunc(void *key, const int hashsize)
{
	__u8 hash;
	__u32 data;

	data = ((struct in_addr *) key)->s_addr;

	hash = (__u8)((data & 0xff000000) >> 24) ^
		(__u8)((data & 0x00ff0000) >> 16) ^
		(__u8)((data & 0x0000ff00) >> 8) ^
		(__u8)(data & 0x000000ff);

	return ((int) hash);
}


/*
 * Comparison function for hash table
 *
 * Arguments:
 *  key        First data structure
 *  cmprd      Second data structure
 *
 * Return:
 *  0          Not equal
 *  1          Equal
 */
static int
cmpfunc(void *key, struct node *cmprd)
{
	struct tunnel_key *data1;
	struct tunnel *data2;

	if (key == NULL) {
		DEBUG(DEBUG_FLAG, "tunnel module, cmpfunc:\n"
		      "value for argument 'key' was NULL !\n"
		      "Illegal value, returning 0 ! \n");
		return 0;
	}
	if (cmprd == NULL) {
		DEBUG(DEBUG_FLAG, "tunnel module, cmpfunc:\n"
		      "value for argument 'cmpd' was NULL !\n"
		      "Illegal value, returning 0 ! \n");
		return 0;
	}

	data1 = (struct tunnel_key *) key;
	data2 = (struct tunnel *) cmprd;

	/* Compare the tunnel destination addresses */
        if (data1->dst_addr.s_addr == data2->dst_addr.s_addr &&
	    data1->type == data2->type &&
	    (data1->type == TUNNEL_IPIP ||
	     (data1->key == data2->key)))
		return 1;

	return 0;
}


static void
print_tunnel_data(struct tunnel *data)
{
	DEBUG(DEBUG_FLAG, "\tdevice=%s, num=%i, dst=%s, ref=%i\n",
	      data->device, data->number, inet_ntoa(data->dst_addr),
	      data->ref);
}


/**
 * tunnel_init:
 * @start_dev: The prefix to be used with tunnel device names, e.g., TUNL
 * @table_start: The first routing table that can be used, e.g., 1
 * @table_end: The last routing table that can be used, e.g., 253
 *
 * Initialize hash table for tunnels. Every tunnel has one entry
 * in the hash table.
 *
 * Returns: A pointer to the empty hash table or NULL on error
 * 
 */
HASH *
tunnel_init(char *start_dev, int table_start, int table_end)
{
	int i;

	dynamics_strlcpy(tunnel_dev_start, start_dev, IFNAMSIZ);
	tunnel_dev_start[IFNAMSIZ - 1] = '\0';

	for (i = 0; i < MAX_TUNNEL_INTS; i++)
		tunnel_numbers[i] = 0;

	/* only allow tables between table_start and table_end (including
	 * both ends), i.e. mark other tables used */
	for (i = 0; i < MAX_TABLE_INTS; i++)
		table_numbers[i] = 0;
	for (i = 0; i < table_start; i++)
		DYN_SET_BIT(table_numbers, i);
	for (i = table_end + 1; i < 256; i++)
		DYN_SET_BIT(table_numbers, i);
	debug_print_bitfield("TABLE_NUMBERS", table_numbers, MAX_TABLE_INTS);

	return (hashtable_init(TUNNEL_HASH_SIZE));
}


/* Destroy tunnel. Called from the iterator. This function removes the real
 * tunnel and frees the released resources. 
 * 
 */
static int 
tunnel_destroy(struct tunnel *data)
{
	int ret;

	DEBUG(DEBUG_FLAG, "tunnel_destroy - dst=%s, dev=%s, close fd %d\n",
	      inet_ntoa(data->dst_addr), data->device, data->tun_fd);

	close_tun_fd(data->tun_fd);

	if (data->direction == TUNNEL_UP && data->to_ha_table_id != -1) {
		DEBUG(DEBUG_FLAG, "Removing to_ha_table %i\n",
		      data->to_ha_table_id);

		if (!DYN_BIT(table_numbers, data->to_ha_table_id))
			LOG2(LOG_WARNING,
			     "tunnel_destroy - to HA table already freed?\n");
		DYN_CLR_BIT(table_numbers, data->to_ha_table_id);
		debug_print_bitfield("TABLE_NUMBERS", table_numbers,
				     MAX_TABLE_INTS);

		/* remove the default route from the to_ha_table */
		if (dyn_ip_route_del_table(data->device, data->to_ha_table_id,
					   NULL) != 0) {
			LOG2(LOG_WARNING, "tunnel_destroy: could not remove "
			     "default route from to_ha_table\n");
			return -1;
		}
	}

	if (data->direction == TUNNEL_UP && data->to_mn_table_id != -1) {
		DEBUG(DEBUG_FLAG, "Removing to_mn_table %i\n",
		      data->to_mn_table_id);
		if (dyn_ip_rule_del_table(NULL, NULL, data->to_mn_table_id,
					  data->device) != 0) {
			LOG2(LOG_ALERT, "tunnel_destroy: could not remove rule"
			     " to MN routing table (table=%i, dev=%s)\n",
			     data->to_mn_table_id, data->device);
		}
		if (dyn_ip_route_del_blackhole(data->to_mn_table_id) != 0) {
			LOG2(LOG_ALERT, "tunnel_destroy: could not remove "
			     "blackhole from MN routing table %i\n",
			     data->to_mn_table_id);
		}
		if (!DYN_BIT(table_numbers, data->to_mn_table_id))
			LOG2(LOG_WARNING,
			     "tunnel_destroy - to MN table already freed?\n");
		DYN_CLR_BIT(table_numbers, data->to_mn_table_id);
		debug_print_bitfield("TABLE_NUMBERS", table_numbers,
				     MAX_TABLE_INTS);
	}

	ret = 1;
	/* remove the tunnel device */
	/* using tun device
	DEBUG(DEBUG_FLAG, "\tdyn_ip_tunnel_del(%s)\n", data->device);
	if (data->type != TUNNEL_NO_ENCAPS &&
	    dyn_ip_tunnel_del(data->device) != 0) {
		DEBUG(DEBUG_FLAG, "\tdyn_ip_tunnel_del(%s) failed!\n",
		      data->device);
		ret = -1;
	}
	*/
	if (ret != -1 && data->type != TUNNEL_NO_ENCAPS) {
		if (!DYN_BIT(tunnel_numbers, data->number)) {
			LOG2(LOG_WARNING, "tunnel_destroy: bit not set!?\n");
		} else
			DYN_CLR_BIT(tunnel_numbers, data->number);
		debug_print_bitfield("TUNNEL_NUMBERS", tunnel_numbers,
				     MAX_TUNNEL_INTS);
	} else if (data->type != TUNNEL_NO_ENCAPS) {
		LOG2(LOG_ALERT, "Tunnel #%i could not be released - keeping "
		     "the number reserved\n", data->number);
	}

	hashtable_remove(&data->hashnode);
	free(data);

	return ret;
}


/* Iterator function for delayed tunnel deletion tunnel_check_delayed().
 * tv = current time value or NULL for forced removal */
static int 
check_tunnel(struct node *node, void *tv)
{
	TUNNEL *data = (struct tunnel *) node;
	struct timeval *now;
	
	if (tv != NULL && (data->ref > 0 || data->do_not_remove))
		return 1;

	now = (struct timeval *) tv;
	if (now == NULL || now->tv_sec < data->zero_ref_tstamp.tv_sec ||
	    now->tv_sec - data->zero_ref_tstamp.tv_sec > PURGE_TUNNEL_TIME)
		tunnel_destroy(data);
	return 1;
}


/**
 * tunnel_destroy_hash:
 * @hash: The pointer to the tunnel hash table
 *
 * Remove the possibly forgotten tunnel entries and destroy the hashtable.
 *
 * Returns 0 on success or 1 on error
 */
int
tunnel_destroy_hash(HASH *hash)
{
	DEBUG(DEBUG_FLAG, "Removing delayed tunnel deletions..\n");
	tunnel_check_delayed(hash, 1);

	if (hashtable_destroy(hash) == 0)
		return 1;
	else
		return 0;
}

/**
 * tunnel_add:
 * @hash:        The pointer to the tunnel hashtable
 * @dst_addr:    The remote address of the tunnel
 * @local_addr:  The local address to be added to the tunnel.
 * @set_link_up: 1 = set the tunnel device up,
 *               0 = do not set the device up (caller takes care of it later)
 * @type:        The type of the tunnel to be created
 * @key:         The key of the tunnel to be created
 * @dev:         If dev != NULL, the device name of the created or modified
 *               tunnel will be copied to it. At least char[IFNAMSIZ].
 *
 * Add or update an entry in the hash table. The entry is identified with the
 * source and destination addresses of the tunnel. The new tunnel is created
 * if it does not exist. This function does not change the routing tables.
 *
 * Returns: a pointer to the added/reused tunnel entry or NULL on failure
 * 
 * jpaaso - just calls the extended tunnel_add2
 */
TUNNEL *
tunnel_add(HASH *hash, struct in_addr dst_addr, char *dev,
	   struct in_addr local_addr, int set_link_up, int type, int key)
{
	DEBUG(DEBUG_FLAG, "tunnel_add: This function should not be called!!!");
	return tunnel_add2(hash, dst_addr, -1, dev, local_addr, set_link_up, 
		type, key);
}

/* tunnel_add2: A modified version of the funciton tunnel_add. Includes the 
 * destination port number parameter needed with UDP tunneling. 
 */
TUNNEL *
tunnel_add2(HASH *hash, struct in_addr dst_addr, unsigned short dst_port, char *dev,
	   struct in_addr local_addr, int set_link_up, int type, int key)
{
	TUNNEL *data, *new = NULL;
	char device[IFNAMSIZ];
	int number = 0, ok;
	struct tunnel_key tkey;

	int tun_fd = 0;

	DEBUG(DEBUG_FLAG, "tunnel_add %s:%d, type=%i, key=%i\n",
	      inet_ntoa(dst_addr), ntohs(dst_port) , type, key);

	tkey.dst_addr 	= dst_addr;
	tkey.type 		= type;
	tkey.key 		= key;
	data = (struct tunnel *) hashtable_fetch(hash, hashfunc, &tkey,
						 cmpfunc);
	if (data != NULL) {
		/* tunnel already exists */
		/* jpaaso - Destroy tunnel to this CoA and create a new one */
		tunnel_destroy(data);
		/*
		data->ref++; 
		if (dev != NULL) dynamics_strlcpy(dev, data->device, IFNAMSIZ);
		DEBUG(DEBUG_FLAG, "\ttunnel already exists, %s (ref = %d)\n",
		      inet_ntoa(dst_addr), data->ref);
		return data;
		*/
	}
	
	if (type == TUNNEL_NO_ENCAPS) {
		if (dyn_ip_get_ifname(key, device) < 0) {
			LOG2(LOG_WARNING,
			     "tunnel_add - could not get ifname\n");
			return NULL;
		}
	} else {
		/* create tunnel */
		number = 0;
		while (number < MAX_TUNNEL_COUNT && DYN_BIT(tunnel_numbers,
							    number))
			number++;
		if (number >= MAX_TUNNEL_COUNT) {
			DEBUG(DEBUG_FLAG, "\tnot enough space for another "
			      "tunnel (%i/%i)\n", number, MAX_TUNNEL_COUNT);
			return NULL;
		}
		snprintf(device, IFNAMSIZ, "%s%d", tunnel_dev_start, number);
	}
	
	/*--------------------------------------------------------------*
 	 | Create UDP in IP -tunnel										|
	 *--------------------------------------------------------------*/
	/* Create TUN device */
	if (dyn_ip_tunnel_add_udpip(device, dst_addr, local_addr, &tun_fd) != 0) 
	{
		DEBUG(DEBUG_FLAG, "\tdyn_ip_tunnel_add_udpip failed!\n");
		return NULL;
	}
	
	ok = TRUE;
	
	/*--------------------------------------------------------------*
 	 | Allocate memory for the new tunnel							|
	 *--------------------------------------------------------------*/
	if (ok) {
		/* update the hashtable */
		new = (struct tunnel *) calloc(1, sizeof(struct tunnel));

		if (new == NULL) {
			DEBUG(DEBUG_FLAG, "\tmemory allocation failure!\n");
			ok = FALSE;
		}
	}
	/*--------------------------------------------------------------*
 	 | Fill the new tunnel info										|
	 *--------------------------------------------------------------*/
	if (ok) {
		list_init_node(&new->hashnode);
		new->to_ha_table_id = -1;
		new->to_mn_table_id = -1;
		new->ref = 1;
		new->dst_addr = dst_addr;
		new->number = number;
		dynamics_strlcpy(new->device, device, IFNAMSIZ);
		new->type = type;
		new->key = key;
		new->direction = TUNNEL_NOT_CONNECTED;
		new->do_not_remove = 0;
		new->dst_port = dst_port;	
		new->tun_fd = tun_fd;

		if (dev != NULL)
			dynamics_strlcpy(dev, device, IFNAMSIZ);
	}
	
	/*--------------------------------------------------------------*
 	 | Add the new tunnel into the hashtable						|
	 *--------------------------------------------------------------*/
	if (ok &&
	    hashtable_add(hash, hashfunc, &dst_addr, &new->hashnode) != TRUE) {
		DEBUG(DEBUG_FLAG, "\thashtable_add failed\n");
		ok = FALSE;
	}

	if (!ok) {
		DEBUG(DEBUG_FLAG, "\tdyn_ip_tunnel_del(%s)\n", device);
		if (dyn_ip_tunnel_del(device) != 0)
			DEBUG(DEBUG_FLAG, "\tdyn_ip_tunnel_dev(%s) failed\n",
			      device);
		if (new != NULL)
			free(new);
		return NULL;
	}

	/* okay, tunnel adding succeeded, update the bitfields etc. */
	if (type != TUNNEL_NO_ENCAPS) {
		DYN_SET_BIT(tunnel_numbers, number);
		debug_print_bitfield("TUNNEL_NUMBERS", tunnel_numbers,
				     MAX_TUNNEL_INTS);
	}

	return new;
}


/**
 * tunnel_set_remove:
 * @tunl:	A tunnel entry
 * @remove:	0 = allow removal of unneeded tunnel,
 *		1 = deny removal of the tunnel even if reference count reaches
 *		    zero
 *
 * Set 'tunnel removing possible' flag for a tunnel entry.
 */
void tunnel_set_remove(TUNNEL *tunl, int remove)
{
	tunl->do_not_remove = remove;
}


/* Iterator function for tunnel_set_remove_all() */
static int set_remove_iter(struct node *node, void *a)
{
	TUNNEL *data = (struct tunnel *) node;
	struct in_addr *addr = (struct in_addr *) a;
	if (data->dst_addr.s_addr == addr->s_addr &&
	    data->do_not_remove) {
		DEBUG(DEBUG_FLAG, "Marking tunnel %s to %s removable\n",
		      data->device, inet_ntoa(data->dst_addr));
		data->do_not_remove = 0;
	}
	return 1;
}


/* tun_fds_into_fdset_iter: Used with hashtable_iterator.
 * 
 * Returns 1 to continue iteration, 0 to stop.
 */
int tun_fds_into_fdset_iter(struct node *node, void *set) {

	TUNNEL *tunnel_entry 	= (struct tunnel *) node;
	fd_set *fdset 			= (fd_set *) set;
	int fd					= tunnel_entry->tun_fd; 
	
	FD_SET(fd, fdset);
	DEBUG(DEBUG_FLAG, "Tun fd %d into fd_set\n", fd);
	
	return 1;
}

/**
 * tunnel_set_remove_all:
 * @hash:	The pointer to the tunnel hash table
 * @dst_addr:	The remote end-point of the tunnels to be modified
 *
 * Mark all the tunnels to @dst_addr removable.
 */
void tunnel_set_remove_all(HASH *hash, struct in_addr dst_addr)
{
	hashtable_iterator(hash, set_remove_iter, &dst_addr);
}


/* Help function for tunnel_connect() */
static int new_table_to_ha(char *dev)
{
	int i;

	i = 0;
	while (i < MAX_TABLE_COUNT && DYN_BIT(table_numbers, i)) i++;
	if (i >= MAX_TABLE_COUNT) {
		LOG2(LOG_WARNING, "tunnel_connect: "
		     "no space for a new routing table\n");
		return -1;
	}

	if (dyn_ip_route_add_table(dev, i, NULL) != 0) {
		LOG2(LOG_WARNING, "tunnel_connect: "
		     "dyn_ip_route_add_table failed\n");
		return -1;
	}

	DEBUG(DEBUG_FLAG, "\treserving to HA table %i\n", i);
	DYN_SET_BIT(table_numbers, i);
	debug_print_bitfield("TABLE_NUMBERS", table_numbers, MAX_TABLE_INTS);

	return i;
}


/* Help function for tunnel_connect() */
static int new_table_to_mn(char *dev)
{
	int i;

	i = 0;
	while (i < MAX_TABLE_COUNT && DYN_BIT(table_numbers, i)) i++;
	if (i >= MAX_TABLE_COUNT) {
		DEBUG(DEBUG_FLAG, "\tno space for new routing table\n");
		return -1;
	}

	DEBUG(DEBUG_FLAG, "\treserving to MN table %i\n", i);
	DYN_SET_BIT(table_numbers, i);
	debug_print_bitfield("TABLE_NUMBERS", table_numbers, MAX_TABLE_INTS);

	/* add default blackhole */
	if (dyn_ip_route_add_blackhole(i) != 0) {
		LOG2(LOG_WARNING, "tunnel_connect: Could not "
		     "initialize a routing table for MNs\n");
		return -1;
	}

	/* add rule for MN host routes */
	if (dyn_ip_rule_add_table(NULL, NULL, i, dev) != 0) {
		LOG2(LOG_WARNING, "tunnel_connect: "
		     "could not add rule for the MN routing table\n");
		return -1;
	}

	return i;
}


/**
 * tunnel_connect:
 * @up:       The pointer to the upwards tunnel device
 * @down:     The pointer to the downwards tunnel device
 * @mn_addr:  The home address of the mobile node
 * @reverse:  1 = add reverse tunneling rule, 0 = no reverse tunneling
 * @force_route_dev:	a device name forced to the routing entry (to be used
 *                      with FA decaps and encapsulating delivery) or
 *			%NULL to not force the device name
 * @force_reverse_dev:	a device name forced to the reverse routing rule (to be
 *                      used with MN decaps without encapsulating delivery) or
 *			%NULL to not force the device name
 *
 * Connect two tunnels (@up and @down) to each other, i.e., set the routing
 * entries and rules to forward the packets between these tunnels for the
 * given @mn_addr.
 *
 * Returns: -1 on failure or 0 on success
 */
int
tunnel_connect(TUNNEL *up, TUNNEL *down, struct in_addr mn_addr, int reverse,
	       char *force_route_dev, char *force_reverse_dev)
{
	struct delayed_connection_deletion *tmp;
	char *dev;

	if (up == NULL || down == NULL)
		return -1;

	DEBUG(DEBUG_FLAG, "tunnel_connect: %s => %s\n", down->device,
	      up->device);

	down->direction = TUNNEL_DOWN;
	up->direction = TUNNEL_UP;

	if (up->to_ha_table_id == -1) {
		/* make a default route to a tunnel device to a new routing
		 * table */
		up->to_ha_table_id = new_table_to_ha(up->device);
		if (up->to_ha_table_id < 0)
			return -1;
	}

	if (up->to_mn_table_id == -1) {
		/* make the routing table for MN host routes */
		up->to_mn_table_id = new_table_to_mn(up->device);
		if (up->to_mn_table_id < 0)
			return -1;
	}

	/* add host route to the MN */
	dev = (force_route_dev && force_route_dev[0] != '\0')
		? force_route_dev : down->device;
	DEBUG(DEBUG_FLAG, "\troute replace: addr=%s, dev=%s, table=%i\n",
	      inet_ntoa(mn_addr), dev, up->to_mn_table_id);
	if (dyn_ip_route_replace_table(mn_addr, dev, up->to_mn_table_id) == -1)
	{
		LOG2(LOG_WARNING, "tunnel_connect: could not add host route to"
		     " the MN\n");
		return -1;
	}

	/* remove the deletion of the possible replaced routing entry from the
	 * delayed unconnect priority queue */
	tmp = delayed;
	while (tmp != NULL) {
		if (tmp->delete_route &&
		    tmp->mn_addr.s_addr == mn_addr.s_addr &&
		    tmp->to_mn_table_id == up->to_mn_table_id) {
			DEBUG(DEBUG_FLAG,
			      "\tremoved delayed route deletion (addr=%s, "
			      "to_mn_table_id=%i, route_dev=%s, "
			      "reverse_dev=%s\n",
			      inet_ntoa(tmp->mn_addr), tmp->to_mn_table_id,
			      tmp->route_dev, tmp->reverse_dev);
			tmp->delete_route = 0;
		}
			   
		tmp = tmp->next;
	}

	if (reverse) {
		/* set reverse tunnel - use source routing
		 * (MN->HA direction data) */
		dev = (force_reverse_dev && force_reverse_dev[0] != '\0')
			? force_reverse_dev : down->device;

		DEBUG(DEBUG_FLAG, "\tdyn_ip_rule_add_table: "
		      "addr=%s, table_id=%i, dev=%s\n",
		      inet_ntoa(mn_addr), up->to_ha_table_id, dev);
		if (dyn_ip_rule_add_table(&mn_addr, NULL, up->to_ha_table_id,
					  dev) != 0) {
			DEBUG(DEBUG_FLAG, "\tdyn_ip_rule_add_table failed!\n");
			return -1;
		}
	}

	return 0;
}


/* The local function to remove the real tunnel connection after some delay
 * (or immediately if the unconnect is forced) */
static int
tunnel_unconnect_real(struct delayed_connection_deletion *entry)
{
	int ret = 0;

	if (entry == NULL)
		return -1;

	DEBUG(DEBUG_FLAG, "tunnel_unconnect_real - MN=%s, reverse=%i, "
	      "to_ha_table_id=%i, to_mn_table_id=%i, "
	      "route_dev=%s, reverse_dev=%s\n",
	      inet_ntoa(entry->mn_addr), entry->reverse, entry->to_ha_table_id,
	      entry->to_mn_table_id, entry->route_dev, entry->reverse_dev);

	if (entry->reverse && entry->to_ha_table_id != -1 &&
	    dyn_ip_rule_del_table(&entry->mn_addr, NULL, entry->to_ha_table_id,
				  entry->reverse_dev) != 0) {
		DEBUG(DEBUG_FLAG, "tunnel_unconnect_real: could not remove "
		      "associated rule\n");
		ret = -1;
	}

	/* remove the host route from to_mn_table_id (down tunnels) */
	if (entry->delete_route && entry->to_mn_table_id != -1) {
		DEBUG(DEBUG_FLAG, "\tdelete route: addr=%s, dev=%s, "
		      "table=%i\n", inet_ntoa(entry->mn_addr),
		      entry->route_dev, entry->to_mn_table_id);
	}
	if (entry->delete_route && entry->to_mn_table_id != -1 &&
	    dyn_ip_route_del_table(entry->route_dev, entry->to_mn_table_id,
				   &entry->mn_addr) == -1) {
		LOG2(LOG_WARNING, "tunnel_unconnect_real: could not remove "
		     "host route to the MN\n");
		ret = -1;
	}

	return ret;
}


/**
 * tunnel_unconnect:
 * @up:       The pointer to the upwards tunnel device
 * @down:     The pointer to the downwards tunnel device
 * @mn_addr:  The home address of the mobile node
 * @reverse:  1 = add reverse tunneling rule, 0 = no reverse tunneling
 * @force:   0 = delayed unconnect, 1 = forced unconnect 
 * @force_route_dev:	a device name forced to the routing entry (to be used
 *                      with FA decaps and encapsulating delivery) or
 *			%NULL to not force the device name
 * @force_reverse_dev:	a device name forced to the reverse routing rule (to be
 *                      used with MN decaps without encapsulating delivery) or
 *			%NULL to not force the device name
 *
 * Unconnect the tunnels connected with tunnel_connect().
 * 
 * Returns: -1 on failure or 0 on success 
 */
int
tunnel_unconnect(TUNNEL *up, TUNNEL *down, struct in_addr mn_addr, int reverse,
		 int force, char *force_route_dev, char *force_reverse_dev)
{
	struct delayed_connection_deletion *entry, *tmp;

	if (up == NULL || down == NULL)
		return -1;

	DEBUG(DEBUG_FLAG, "tunnel_unconnect - up[dev=%s,tables=%i,%i], "
	      "down[dev=%s,tables=%i,%i], MN=%s, reverse=%i, force=%i\n",
	      up->device, up->to_ha_table_id, up->to_mn_table_id,
	      down->device, down->to_ha_table_id, down->to_mn_table_id,
	      inet_ntoa(mn_addr), reverse, force);

	entry = (struct delayed_connection_deletion *)
		malloc(sizeof(struct delayed_connection_deletion));
	if (entry == NULL) {
		LOG2(LOG_WARNING, "tunnel_unconnect - out of memory\n");
		return -1;
	}
	memset(entry, 0, sizeof(struct delayed_connection_deletion));
	gettimeofday(&entry->timeout, NULL);
	entry->timeout.tv_usec += DELAY_UNCONNECTION_USEC;
	if (entry->timeout.tv_usec > 999999) {
		entry->timeout.tv_usec += entry->timeout.tv_usec / 1000000;
		entry->timeout.tv_usec %= 1000000;
	}
	entry->mn_addr = mn_addr;
	entry->reverse = reverse;
	entry->delete_route = 1;
	entry->to_ha_table_id = up->to_ha_table_id;
	entry->to_mn_table_id = up->to_mn_table_id;
	memcpy(entry->route_dev,
	       (force_route_dev && force_route_dev[0] != '\0')
	       ? force_route_dev : down->device, IFNAMSIZ);
	memcpy(entry->reverse_dev,
	       (force_reverse_dev && force_reverse_dev[0] != '\0')
	       ? force_reverse_dev : down->device, IFNAMSIZ);

#ifdef REMOVE_BINDING_OPTIM
	/* remove delayed unconnect optimization for comparison measurements */
	force = 1;
#endif
	if (force) {
		int ret = tunnel_unconnect_real(entry);
		free(entry);
		return ret;
	}

	/* add the entry to the priority queue to be removed after some time */
	if (delayed == NULL) {
		/* first entry in the priority queue */
		delayed = entry;
		return 0;
	}
	if (cmp_timeval(&delayed->timeout, &entry->timeout) > 0) {
		/* replaces the head of the priority queue */
		entry->next = delayed;
		delayed = entry;
		return 0;
	}

	tmp = delayed;
	while (tmp->next != NULL &&
	       cmp_timeval(&tmp->next->timeout, &entry->timeout) <= 0)
		tmp = tmp->next;
	entry->next = tmp->next;
	tmp->next = entry;
	return 0;
}


/**
 * tunnel_check_delayed:
 * @hash:	The pointer to the tunnel hash table
 * @force:
 *    0 = do not force the deletion,
 *    1 = force deletion even if the entries have not timed out yet,
 *    2 = force only the deletion of not yet timed out connections
 *
 * Check and remove the delayed connection and tunnel deletions that have
 * expired.
 */
void
tunnel_check_delayed(HASH *hash, int force)
{
	static time_t prev_check = 0;
	struct timeval now;

	gettimeofday(&now, NULL);

	/* check delayed priority queue */
	while (delayed != NULL &&
	       (force || cmp_timeval(&delayed->timeout, &now) <= 0)) {
		struct delayed_connection_deletion *prev;
		DEBUG(DEBUG_FLAG, "tunnel_check_delayed - deleting entry - "
		      "now=%li.%li, timeout=%li.%li\n", now.tv_sec,
		      now.tv_usec, delayed->timeout.tv_sec,
		      delayed->timeout.tv_usec);
		tunnel_unconnect_real(delayed);
		prev = delayed;
		delayed = delayed->next;
		free(prev);
	}

	/* check the tunnels if enought time has passed from the previous
	 * check */
	if (force != 1 && prev_check <= now.tv_sec &&
	    now.tv_sec < prev_check + PURGE_TUNNEL_TIME)
		return;

	prev_check = now.tv_sec;
	hashtable_iterator(hash, check_tunnel, force == 1 ? NULL : &now);
}

/** 
 * tunnel_delete_ptr:
 * @data:   The pointer to the tunnel entry to be deleted.
 * @force:  1 = the tunnel will be removed immediately,
 *          0 = the deletion is delayed
 *
 * Remove the given tunnel.
 *
 * Returns:	-1 on error, 0 if the tunnel was not deleted
 *		(i.e., still in use), 1 if the tunnel was deleted
 */
int tunnel_delete_ptr(TUNNEL *data, int force)
{
	int ret = 1;

	if (data == NULL) {
		DEBUG(DEBUG_FLAG, "tunnel_delete_ptr: no such entry\n");
		return -1;
	}

	DEBUG(DEBUG_FLAG, "tunnel_delete_ptr %s,%i,%i, force=%i\n",
	      inet_ntoa(data->dst_addr), data->type, data->key, force);

	DEBUG(DEBUG_FLAG, "\treference count: %i => %i\n",
	      data->ref, data->ref - 1);
	data->ref--;
	print_tunnel_data(data);

	if (data->ref > 0) {
		DEBUG(DEBUG_FLAG, "\ttunnel not deleted\n");
		return 0;
	}

#ifdef REMOVE_TUNNEL_OPTIM
	force = 1;
#endif

	if (force) {
		ret = tunnel_destroy(data);
	} else {
		/* Reference count is zero. Delay tunnel deletion */
		gettimeofday(&data->zero_ref_tstamp, NULL);
	}
	
	return ret;
}


/**
 * tunnel_delete:
 *    @hash:        The pointer to the tunnel hash table
 *    @dst_addr:    The remote address of the tunnel
 *    @force:       1 = Remove tunnel immediately if reference count is zero,
 *                  0 = Use delayed tunnel deletion
 *    @type:        Tunnel type (IPIP / GRE)
 *    @key:         Key for GRE tunnels
 *
 * Delete tunnel from the hash table and decrease the tunnel reference count.
 * If the reference count is zero remove the kernel tunnel device.
 *
 * Returns:
 *    0		Reference count decreased,
 *    1		The tunnel was destroyed,
 *   -1		Error
 */
int
tunnel_delete(HASH *hash, struct in_addr dst_addr, int force, int type,
	      __u32 key)
{
	TUNNEL *data;
	struct tunnel_key tkey;

	DEBUG(DEBUG_FLAG, "tunnel_delete %s, %i, %i\n", inet_ntoa(dst_addr),
	      type, key);
	
	tkey.dst_addr = dst_addr;
	tkey.type = type;
	tkey.key = key;
	data = (struct tunnel *) hashtable_fetch(hash, hashfunc, &tkey,
						 cmpfunc);
	/*
	DEBUG(DEBUG_FLAG, "\tdel_tun_device(%s)\n", data->device);
	if (data->type != TUNNEL_NO_ENCAPS &&
	    del_tun_device(data->tun_fd) != 0) {
		DEBUG(DEBUG_FLAG, "\tdyn_ip_tunnel_del(%s) failed!\n",
		      data->device);
	}
	*/
	return tunnel_delete_ptr(data, force);
}


/**
 * tunnel_fetch:
 * @hash:	The pointer to the tunnel hash table
 * @dst_addr:	The remote address ot the tunnel
 * @type:	The tunnel type (#TUNNEL_NO_ENCAPS, #TUNNEL_IPIP, #TUNNEL_GRE)
 * @key:	The key used with the tunnel
 *
 * Fetch the tunnel data from the hash table.
 *
 * Returns:	A pointer to the tunnel entry or NULL on failure
 */
TUNNEL *
tunnel_fetch(HASH *hash, struct in_addr dst_addr, int type, __u32 key)
{
	TUNNEL *entry;
	struct tunnel_key tkey;

	DEBUG(DEBUG_FLAG, "tunnel_fetch: %s\n", inet_ntoa(dst_addr));

	tkey.dst_addr = dst_addr;
	tkey.type = type;
	tkey.key = key;
	entry = (TUNNEL *) hashtable_fetch(hash, hashfunc, &tkey, cmpfunc);
	if (entry != NULL)
		print_tunnel_data(entry);
	else
		DEBUG(DEBUG_FLAG, "\tno such entry\n");
	return entry;
}

static int tunnel_fetch2_iter(struct node *node, void *dst_addr) {
	TUNNEL *tunnel 			= (TUNNEL *) node;
	struct in_addr *dst = (struct in_addr *) dst_addr;
	
	if (tunnel->dst_addr.s_addr == dst->s_addr) {
		found_tunnel = tunnel;
		return 0;
	}
	return 1;
}

/**
 * Get tunnel entry with destination address *coa.
 * 
 * Returns the tunnel entry pointer or NULL if it was not found.
 */
TUNNEL *tunnel_fetch2(HASH *hash, struct in_addr *coa) {
	found_tunnel = NULL;
	hashtable_iterator(hash, tunnel_fetch2_iter, coa);
	if (found_tunnel == NULL) {
		DEBUG(DEBUG_FLAG, "tunnel_fetch2: Tunnel entry with CoA %s not found!\n",
			inet_ntoa(*coa));
		return NULL;
	}
	return found_tunnel;
}

static int 
find_delayed(struct node *node, void *val)
{
	int *found = (int *) val;
	TUNNEL *data = (struct tunnel *) node;
	
	if (data->ref > 0 || data->do_not_remove)
		return 1;

	*found = 1;
	return 0;
}


/**
 * tunnel_delayed_exists:
 * @hash: The pointer to the tunnel hash table
 *
 * Checks whether any delayed deletions exists in the tunnel hash table.
 *
 * Returns: 0 = no delayed deletions, 1 = at least one delayed deletion found
 */
int
tunnel_delayed_exists(HASH *hash)
{
	int found;

	if (delayed != NULL)
		return 1;

	found = 0;
	hashtable_iterator(hash, find_delayed, &found);
	return found;
}
