/* $Id: util.c,v 1.45 2001/09/30 13:53:38 jm Exp $
 * Help routines for the agents
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/types.h>
#include <unistd.h>
#include <assert.h>
#include <syslog.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#ifdef DYN_TARGET_WINDOWS
#include "windows_extra.h"
#else
#include <netinet/udp.h>
#endif
#include <sys/stat.h>
#include <fcntl.h>

#include "debug.h"
#include "dyn_ip.h"
#include "util.h"

#define DEBUG_FLAG 'u'

int opt_foreground = 0;
char *opt_config = NULL;
int opt_debug = 0;

static struct option
#ifndef DYN_TARGET_WINDOWS
const
#endif
long_options[] =
{
	{"help", no_argument, NULL, 'h'},
	{"version", no_argument, NULL, 'v'},
	{"fg", no_argument, &opt_foreground, 1},
	{"debug", no_argument, &opt_debug, 1},
	{"config", required_argument, NULL, 'c'},
	{0, 0, 0, 0}
};

char *
parse_long_options(int argc, char *argv[], const char *command_name,
		   const char *package, const char *version,
		   void (*usage)(int, char *))
{
	int c;
	int saved_opterr;
	char *program_name;

	if (strchr(argv[0], '/'))
		program_name = strrchr(argv[0], '/') + 1;
	else
		program_name = argv[0];

	saved_opterr = opterr;
	/* do not print an error message for unrecognized options */
	opterr = 0;

	while ((c = getopt_long(argc, argv, "+hv", long_options, NULL)) !=
	       EOF) {
		switch (c) {
		case 'h':
			(*usage) (0, program_name);
			break;

		case 'v':
			printf("%s (%s) %s\n", command_name, package,
			       version);
			printf("Copyright (C) 1998-2001, Dynamics group\n"
			       "This program comes with NO WARRANTY,\n"
			       "to the extent permitted by law.\n"
			       "You may redistribute copies of this program\n"
			       "under the terms of the GNU General Public "
			       "License.\n"
			       "For more information about these matters,\n"
			       "see the files named COPYING.\n");

			exit(0);

		case 'c':
			opt_config = optarg;
			assert(opt_config != NULL);
			printf("Using configuration file '%s'\n", opt_config);
			break;

		default:
			/* skip other parameters */
			break;
		}
	}

	opterr = saved_opterr;
	/* Reset the getopt internals if it is called again */
	optind = 1;

	return program_name;
}


/* print usage information */
void
dynamics_usage(int status, char *name)
{
	if (status != 0) {
		fprintf(stderr, "Try `%s --help' for more information.\n",
			name);
	} else {
		printf("Usage: %s [OPTION]\n\n"
		       "      -h / --help         display this help\n"
		       "      --fg                do not fork the daemon\n"
		       "      --version           output version information\n"
		       "      --debug             print debug information "
		       "(use with --fg)\n"
		       "      --config <file>     use given file as a "
		       "configuration file\n",
		       name);
		/* each agent shows it own specific parameters after this */
	}
	if (status != 0)
		exit(status);
}


int
dynamics_write_pid_file(char *pid_file)
{
	FILE *f;

	if (pid_file != NULL) {
		f = fopen(pid_file, "w");
		if (f != NULL) {
			fprintf(f, "%u\n", (unsigned int) getpid());
			fclose(f);
		}
		return 0;
	}
	return -1;
}


/* Forks a daemon process and exits the original process
 * Returns:
 *    -1: fork failed
 *     0: fork succeeded, running in daemon process
 *     1: -fg used, running the same process
 */
int
dynamics_fork_daemon(void)
{
	pid_t child;
	int s;

        /* fork daemon and exit */
        child = fork();
        if (child < 0) {
                fprintf(stderr, "fork failed\n");
                return -1;
        }
	if (child > 0) {
                /* original process - exit return immediately
		 * (no & needed in the end of the command line) */
                exit(0);
        }

	/* point stdin, stdout, and stderr to /dev/null */
	s = open("/dev/null", O_RDWR);
	if (s >= 0) {
		dup2(s, STDIN_FILENO);
		dup2(s, STDOUT_FILENO);
		dup2(s, STDERR_FILENO);
		if (s > 2)
			close(s);
	} else {
		close(STDIN_FILENO);
		close(STDOUT_FILENO);
		close(STDERR_FILENO);
	}

	return 0;
}


unsigned int
get_rand32(void)
{
	unsigned int data;
	static FILE *rand_file = NULL;

	if (rand_file == NULL) {
		rand_file = fopen("/dev/urandom", "rb");
		if (rand_file == NULL) {
			DEBUG(DEBUG_FLAG, "Cannot open /dev/urandom - "
			      "reverting to rand()\n");
			return rand();
		}
	}

	fread(&data, 1, 4, rand_file);
	return data;
}


/** 
 * allow_ip_forwarding:
 * 
 * Sets up kernel to allow IPv4 forwarding
 * 
 * Returns: -1 on failure and 0 on success 
 */
int
allow_ipv4_forwarding(void)
{
	int ret;
	FILE *f;

	f = fopen("/proc/sys/net/ipv4/ip_forward", "w");
	if (f == NULL) return -1;
	ret = fputs("1", f) > 0;
	fclose(f);
	return ret;
}


void
check_kernel_support(int options)
{
	int ok = 1;

	if ((options & CHECK_KERNEL_ADV_ROUTING) &&
	    dyn_ip_check_adv_routing() < 0) {
		fprintf(stderr, "Kernel does not support advanced policy "
			"routing.\n"
			"You will need to recompile the kernel with that "
			"support\n"
			"(options CONFIG_IP_ADVANCED_ROUTER and "
			"CONFIG_IP_MULTIPLE_TABLES).\n");
		ok = 0;
	}
	/*
	// jpaaso - no need to check IPIP
	if ((options & CHECK_KERNEL_IPIP) && dyn_ip_check_ipip() < 0) {
		fprintf(stderr, "Kernel does not support IP-within-IP "
			"encapsulation (ipip.o module)\n"
			"Try to load it with 'modprobe ipip'.\n"
			"If that does not work, you will need to recompile "
			"the kernel with IPIP support (CONFIG_NET_IPIP).\n");
		ok = 0;
	}

	if ((options & CHECK_KERNEL_GRE) && dyn_ip_check_gre() < 0) {
		fprintf(stderr, "Kernel does not support GRE "
			"encapsulation (ip_gre.o module)\n"
			"Try to load it with 'modprobe ip_gre'.\n"
			"If that does not work, you will need to recompile "
			"the kernel with GRE support (CONFIG_NET_IPGRE).\n");
		ok = 0;
	}
	*/
	if ((options & CHECK_KERNEL_NETLINK) && dyn_ip_check_netlink() < 0) {
		fprintf(stderr, "Kernel does not support netlink socket and/or"
			" routing messages\n"
			"You will need to recompile the kernel with networking"
			" options CONFIG_NETLINK and CONFIG_RTNETLINK.\n");
		ok = 0;
	}

	if (!ok) {
		fprintf(stderr, "At least one of the required kernel options "
			"was missing. Aborting.\n");
		exit(1);
	}
}

void init_key_extension(struct msg_key *ext, __u16 sub_type, __u32 spi,
			__u8 keylen)
{
	assert(ext != NULL && keylen <= (255-7));

	ext->type = VENDOR_EXT_TYPE2;
	ext->length = keylen + MIN_KEY_EXT_LEN;
	ext->reserved = 0;
	ext->vendor_id = htonl(VENDOR_ID_DYNAMICS);
	ext->sub_type = htons(sub_type);
	ext->spi = spi;
}


#ifndef ETH_ALEN
#define ETH_ALEN 6
#endif

/* Return a pointer to hardware address string */
char* ether_hwtoa(unsigned char *addr)
{
	static char ret[ETH_ALEN * 3 + 1];
	int i;

	for (i = 0; i < ETH_ALEN; i++)
		sprintf(&ret[i * 3], "%02x:", addr[i]);
	ret[sizeof(ret) - 2] = '\0';
	return ret;
}


int ether_atohw(const char *hwtxt, unsigned char *addr)
{
	int n;
	unsigned int _addr[6];
	n = sscanf(hwtxt, "%02x:%02x:%02x:%02x:%02x:%02x",
		   &_addr[0], &_addr[1], &_addr[2], &_addr[3], &_addr[4],
		   &_addr[5]);
	if (n != 6) {
		DEBUG(DEBUG_FLAG, "ether_atohw: invalid addr '%s' (n=%i)\n",
		      hwtxt, n);
		return -1;
	}
	for (n = 0; n < 6; n++)
		addr[n] = _addr[n] & 0xff;
	return 0;
}


char* reply_code_str(__u8 code)
{
	switch (code) {
	case REGREP_ACCEPTED:
		return "registration accepted";
	case REGREP_ACCEPTED_NO_SB:
		return "registration accepted, but simultaneous mobility "
			"binding unsupported";
	case REGREP_PREACCEPTED: /* jpaaso - Added to support HAaRO */
		return "registration pre-accepted: wait for CR's initial keepalive "
			"request with some CR CoA - MR CoA, and use those in tunneling";
	case REGREP_REASON_UNSPEC_FA:
		return "reason unspecified (FA)";
	case REGREP_ADMIN_PROHIBITED_FA:
		return "administratively prohibited (FA)";
	case REGREP_NO_RESOURCES_FA:
		return "insufficient resources (FA)";
	case REGREP_MN_FAILED_AUTH_FA:
		return "mobile node failed authentication (FA)";
	case REGREP_HA_FAILED_AUTH_FA:
		return "home agent failed authentication (FA)";
	case REGREP_LONG_LIFETIME_FA:
		return "requested Lifetime too long (FA)";
	case REGREP_BAD_REQUEST_FA:
		return "poorly formed Request (FA)";
	case REGREP_BAD_REPLY_FA:
		return "poorly formed Reply (FA)";
	case REGREP_ENCAP_UNAVAIL_FA:
		return "requested encapsulation unavailable (FA)";
	case REGREP_VJ_UNAVAIL_FA:
		return "requested Van Jacobson compression unavailable (FA)";
	case REGREP_INVALID_CAREOF_FA:
		return "invalid care-of address (FA)";
	case REGREP_REGISTRATION_TIMEOUT_FA:
		return "registration timeout (FA)";
	case REGREP_DELIVERY_STYLE_NOT_SUPPORTED_FA:
		return "delivery style not supported (FA)";
	case REGREP_HN_UNCREACHABLE_FA:
		return "home network unreachable (FA)";
	case REGREP_HA_HOST_UNCREACHABLE_FA:
		return "home agent host unreachable (FA)";
	case REGREP_HA_PORT_UNCREACHABLE_FA:
		return "home agent port unreachable (FA)";
	case REGREP_HA_UNREACHABLE_FA:
		return "home agent unreachable (FA)";
	case REGREP_NONZERO_HOMEADDR_REQD_FA:
		return "non-zero home address required (FA)";
	case REGREP_MISSING_NAI_FA:
		return "missing network access identifier (FA)";
	case REGREP_MISSING_HOME_AGENT_FA:
		return "missing home agent (FA)";
	case REGREP_MISSING_HOMEADDR_FA:
		return "missing home address (FA)";
	case REGREP_UNKNOWN_CHALLENGE_FA:
		return "unknown challenge (FA)";
	case REGREP_MISSING_CHALLENGE_FA:
		return "missing challenge (FA)";
	case REGREP_STALE_CHALLENGE_FA:
		return "stale challenge (FA)";
	case REGREP_MISSING_MN_FA_FA:
		return "missing MN-FA key material from AAA extension";
	case REGREP_UNSUPP_VENDOR_ID_MN_FA:
		return "unsupported vendor extension (MN->FA)";
	case REGREP_UNSUPP_VENDOR_ID_HA_FA:
		return "unsupported vendor extension (HA->FA)";
	case REGREP_REASON_UNSPEC_HA:
		return "reason unspecified (HA)";
	case REGREP_ADMIN_PROHIBITED_HA:
		return "administratively prohibited (HA)";
	case REGREP_NO_RESOURCES_HA:
		return "insufficient resources (HA)";
	case REGREP_MN_FAILED_AUTH_HA:
		return "mobile node failed authentication (HA)";
	case REGREP_FA_FAILED_AUTH_HA:
		return "foreign agent failed authentication (HA)";
	case REGREP_ID_MISMATCH_HA:
		return "registration Identification mismatch (HA)";
	case REGREP_BAD_REQUEST_HA:
		return "poorly formed Request (HA)";
	case REGREP_TOO_MANY_SIMULTANEOUS:
		return "too many simultaneous mobility bindings (HA)";
	case REGREP_UNKNOWN_HA_HA:
		return "unknown home agent address (HA)";
	case REGREP_REVERSE_TUNNEL_UNAVAIL_FA:
		return "requested reverse tunnel unavailable (FA)";
	case REGREP_REVERSE_TUNNEL_MANDATORY_FA:
		return "reverse tunnel is mandatory and 'T' bit not set (FA)";
	case REGREP_REVERSE_TUNNEL_MN_TOO_DISTANT_FA:
		return "mobile node too distant (FA)";
	case REGREP_REVERSE_TUNNEL_UNAVAIL_HA:
		return "requested reverse tunnel unavailable (HA)";
	case REGREP_REVERSE_TUNNEL_MANDATORY_HA:
		return "reverse tunnel is mandatory and 'T' bit not set (HA)";
	case REGREP_ENCAP_UNAVAIL_HA:
		return "requested encapsulation unavailable (HA)";
	case REGREP_UNSUPP_VENDOR_ID_MN_HA:
		return "unsupported vendor extension (MN->HA)";
	case REGREP_UNSUPP_VENDOR_ID_FA_HA:
		return "unsupported vendor extension (FA->HA)";
	}

	return "UNKNOWN CODE";
}


unsigned int usec_passed(struct timeval *tv, struct timeval *now)
{
	struct timeval t;
	unsigned long diff;
	
	if (now == NULL) {
		gettimeofday(&t, NULL);
		diff = (t.tv_sec - tv->tv_sec) * 1000000 + 
			t.tv_usec - tv->tv_usec;
	}
	else {
		diff = (now->tv_sec - tv->tv_sec) * 1000000 + 
			now->tv_usec - tv->tv_usec;
	}

	return diff;
}


unsigned int sec_passed(struct timeval *tv)
{
	struct timeval now;
	gettimeofday(&now, NULL);
	return (now.tv_sec - tv->tv_sec);
}

/* returns 1, 0, -1 when tv1 is greater, equal or less than tv2 */
int cmp_timeval(struct timeval *tv1, struct timeval *tv2)
{
	if (tv1->tv_sec < tv2->tv_sec) return -1;
	if (tv1->tv_sec > tv2->tv_sec) return 1;
	if (tv1->tv_usec < tv2->tv_usec) return -1;
	if (tv1->tv_usec > tv2->tv_usec) return 1;
	return 0;
}


void add_usecs(struct timeval *tv, unsigned int usecs)
{
	tv->tv_usec += usecs;
	tv->tv_sec += tv->tv_usec / 1000000;
	tv->tv_usec = tv->tv_usec % 1000000;
}

void set_usecs(struct timeval *tv, unsigned int const usecs)
{
	assert(tv != NULL);
	tv->tv_sec = (unsigned long int) usecs / 1000000;
	tv->tv_usec = (unsigned long int) usecs % 1000000;
}


#define MAX_REPORT_PAYLOAD 100
void report_discarded_msg(const char *msg, int len, struct sockaddr_in *from,
			  const char *reason)
{
	char payload[MAX_REPORT_PAYLOAD * 3 + 100], *pos;
	int i, max;

	syslog(LOG_INFO, "discarded message from %s:%i, len=%i, reason: %s",
	       inet_ntoa(from->sin_addr), ntohs(from->sin_port), len, reason);
	dynamics_strlcpy(payload, "   payload:", sizeof(payload));
	if (len > MAX_REPORT_PAYLOAD)
		max = MAX_REPORT_PAYLOAD;
	else
		max = len;
	pos = payload;
	while (*pos != '\0') pos++;
	for (i = 0; i < max; i++) {
		snprintf(pos, 4, " %02x", msg[i]);
		pos += 3;
	}
	if (len > max)
		dynamics_strlcpy(pos, " ...",
				 sizeof(payload) - (pos - payload));
	syslog(LOG_INFO, "%s", payload);
}

int get_hex_digit(const char c)
{
        if (c >= '0' && c <= '9') return (c - '0');
        if (c >= 'a' && c <= 'f') return (c - 'a' + 10);
        if (c >= 'A' && c <= 'F') return (c - 'A' + 10);
        return -1;
}

/**
 * dynamics_strlcpy:
 * @dst: destination
 * @src: source
 * @size: maximum size of @dst
 *
 * Copy up to size-1 characters from @src to @dst and always NUL terminate
 * @dst string (as long as @size is not 0).
 *
 * Returns strlen(src); if this is >= @size, truncation occurred
 */
size_t dynamics_strlcpy(char *dst, const char *src, size_t size)
{
	char *d = dst;
	const char *s = src;
	size_t n;

	if (size == 0)
		return strlen(src);

	n = size - 1;
	while (n > 0) {
		if ((*d++ = *s++) == '\0')
			break;
		n--;
	}

	if (n == 0) {
		/* truncate and search the end of src */
		*d = '\0';
		while (*s != '\0')
			s++;
	}

	return (s - src);
}


/**
 * ip_checksum:
 * @data: Data to calculated checksum for
 * @len: Length of data
 * 
 * Internet checksum calculation (based on RFC 1071)
 * 
 * Returns: Checksum
 */
int ip_checksum(unsigned char *data, int len)
{
	/* Compute Internet Checksum for "count" bytes
	 *         beginning at location "addr".
	 */
	register long sum = 0;
	__u16 *addr = (__u16 *) data;
	__u16 res;

	assert(data != NULL && len >= 0);

	/* add the 16-bit words */
        while (len > 1) {
		sum += *addr++;
		len -= 2;
	}

	/* add left-over byte, if any */
	if (len > 0)
		sum += *((unsigned char *) addr);

	/* fold 32-bit sum to 16 bits */
	while (sum >> 16)
		sum = (sum & 0xffff) + (sum >> 16);

	res = ~sum;
	return res;
}


/**
 * udp_checksum:
 * @ip: pointer to the IP header of the packet
 * @udp: pointer to the UDP header of the packet
 *
 * Calculate UDP checksum according to RFC 768.
 * Example use: udp->check = 0; udp->check = udp_checksum(ip, udp);
 *
 * Returns: Calculated checksum in network byte order
 */
__u16 udp_checksum(struct iphdr *ip, struct udphdr *udp)
{
	unsigned long sum = 0;
	__u16 *pos;
	int left;

	/* add pseudo header */
	sum += (ip->saddr >> 16) + (ip->saddr & 0xffff);
	sum += (ip->daddr >> 16) + (ip->daddr & 0xffff);
	sum += htons(ip->protocol);
	sum += udp->len;

	left = ntohs(udp->len);
	pos = (__u16 *) udp;
	while (left > 1) {
		sum += *pos++;
		left -= 2;
	}
	if (left)
		sum += htons(*((__u8 *) pos) << 8);

	/* fold 32-bit sum to 16 bits */
	while (sum >> 16)
		sum = (sum & 0xffff) + (sum >> 16);

	return (~sum & 0xffff);
}
